//                              -*- Mode: C++ -*-
//: EAHypAnalyzer.C
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 2/5/97
// Last Update ........
//

#include "care/EMClusterTree.hh"
#include "care/EMTrackOnEcal.hh"
#include "care/EAHypAnalyzer.hh"
#include "arte/ArteMessage.hh"
#include "care/ECStrategyFactory.hh"
#include "care/ESCorCOGPos.hh"
#include "arte/ArteProto.hh"
#include "care/caredata.hh"
#include <math.h>
#include "care/ECCellID.hh"
#include "auto/RTRA.hh"
#include "auto/RCCL.hh"
#include "auto/RCLX.hh"
#include "care/CmdCareLevel.hh"
#undef WRITE
#undef DEBUG
/* 
CLASS
  EAHypAnalyzer
  Class for finding the hypothetical showers.
  It stores a vector of pointers to the hyp showers.


*/

extern "C" {
  void mmdrop_( int& link );
  void mmbrel_(int &relBank, int *hitC, int &nrccl);

}


//////////////////// EAHypAnalyzer Class Definitions ////////////////

// No functions
EAHypAnalyzer::EAHypAnalyzer(vector<const EMCluster*> & hypSh) :
  hypShowers(&hypSh) {

  ECStrategyFactory  stratf;
  poscog = stratf.PositionStrategy(); // return current way to evaluate pos.
  // return current way to evaluate position
  poscor = new ESCorCOGPos;
  // get the current PID strategy
  pidstra = stratf.PidStrategy();
  // initializa data for relational bank
  nrccl = 0;
}

EAHypAnalyzer::~EAHypAnalyzer(){
  delete poscog;
  delete poscor;
  delete pidstra;
}

//--
//  Base function to be called with the TopCluster List
//
void EAHypAnalyzer::analyzeTopList(const EMClusterTree & topList){
  doAnalysis(&topList);
}

//--
//  Base function to be called with the TopCluster List
//
void EAHypAnalyzer::doAnalysis(const EMClusterTree * tree){
  if(hypShowers->begin()!=hypShowers->end()){
    // erase all data
    hypShowers->erase(hypShowers->begin(),hypShowers->end());
  }
  //--
  //  test if RCCL bank is empty (it must be empty)
  if( ArtePointer<RCCL>::getLink().getRef()!=0){
    int nrow = ArtePointer<RCCL>::nrow();
    // drop the bank
    mmdrop_(ArtePointer<RCCL>::getLink().getRef());
    //
    static int EAHYPWARN = 0xECA0004;
    static int count = 0;
    if( count<=40){
      count++;
      arteMsg<<"EAHypAnalyzer: RCCL bank not void: "
	     <<nrow<<" entries ";
      arteMsg.warn(EAHYPWARN);
    }
    if( count==40){
      count++;
      arteMsg<<"... Last message of this kind ";
      arteMsg.warn();
    }
  }
  nrccl = 0;
  for(EMClusterMember* child = tree->begin(); child!=tree->end(); 
      child = child->next()){
    if( !child->isALeaf()){
      child->analyzeMe(*this); 
    }
  }
}



//--
// Cluster Analyzer
void EAHypAnalyzer::doAnalysis(const EMCluster * clus){
  if(clus->getHypotheses()==1){
    // Found one shower hypotheses: store it
    hypShowers->push_back(clus);
    //--
    //  Evaluate the information for the cluster and fill an entry 
    fillNewEntry(clus);  
  } else {
    // Analyze Cluster that is not a shower one
    // perform operation on all elements of the cluster
    for(EMClusterMember* child = clus->begin(); 
	child!=clus->end();  child = child->next())
      child->analyzeMe(*this);
  }
}

//--
// EMHitCell Analyzer
void EAHypAnalyzer::doAnalysis(const EMHitCell *){}

//--
// EMTrackOnEcal Analyzer
void EAHypAnalyzer::doAnalysis(const EMTrackOnEcal *){}

//--
// EMClusterMember Analyzer
void EAHypAnalyzer::doAnalysis(const EMClusterMember *){}

//--
//  Evaluate the information for the cluster and fill a new entry in the bank
//
void EAHypAnalyzer::fillNewEntry(const EMCluster * cluster){
  // get a table entry
  mmbook_(ArtePointer<RCCL>::getLink().getRef(), 1);
  ArtePointer<RCCL> rccl(ArtePointer<RCCL>::nrow());
  //
  // Now the entry exists and must be updated...
  updateOldEntry(cluster, rccl);
  // now fill the relations
  //rcLinks.fillNewEntry(cluster, rccl);
#ifdef WRITE
  // test links
  //testLinks(cluster, rccl);
#endif
}

//--
//  Evaluate the information for the cluster and 
//  fill an old entry in the bank
//
void EAHypAnalyzer::updateOldEntry(const EMCluster * clus, 
				   ArtePointer<RCCL> rccl){
  //
  // Evaluate all informations to store in a bank
  //
  poscog->evaluatePos(clus); // evaluate center of gravity

  // now writeout on bank
  const ESPosition * pos = & poscog->getData();
  // 
  // COG info
  rccl->xcog = pos->x();
  rccl->ycog = pos->y();
  for(int i=0; i<3; i++){
    rccl->vcog[i] = pos->var(i);
  }
  // 
  // Corrected COG info
  if( CmdCareLevel::level()>2){
    poscor->correct(poscog->getData()); //evaluate correction 
    pos = & poscor->getData();
  } else {
    pos = & poscog->getData();  // No correction for level 2
  }
 
  rccl->x = pos->x();
  rccl->y = pos->y();
  for(int j=0; j<3; j++){
    rccl->cve[j] = pos->var(j);
  }

  rccl->esum = pos->energy();
  rccl->wid = pos->width();
  rccl->flag = pos->cell()->ncl();
 
 // z dependence
   rccl->z = pos->z();
  if( CmdCareLevel::level()>2){
    // now likelyhood
    pidstra->beginAnalysis(clus, rccl, pos->cell());
  }
}

//--
//  Evaluate the information for the cluster and 
//  fill an old entry in the bank
//
#ifndef DEBUG
void EAHypAnalyzer::testLinks(const EMCluster * , ArtePointer<RCCL> ){
  //
#else
void EAHypAnalyzer::testLinks(const EMCluster * clus, ArtePointer<RCCL> rccl){
  //


  cout<<"Testing cluster"<<endl<<endl;
  clus->dump();
  // test the link done
  cout<<"Cluster number:"<<rccl.index()
      <<" at "<<rccl->x<<","<<rccl->y
      <<" with "<<rccl->wid<<" cells and "<<rccl->esum<<" GeV"<<endl;
  // now loop on relations of this RCCL with HITC
  ArtePointer<RCLX> rclx(rccl);
  for( ArteRelationIterator<RCLX_HITC> relIter(rclx->hitc);
       relIter.valid(); relIter.next()){
    ArtePointer<HITC> cell = relIter;
    cout<<"->Cell: ID="<<cell->gcal<<"  Energy = "<<cell->q
	<<"  Pos:("<<cell->x<<", "<<cell->y<<" ) ";
    cout<<endl;
  }
  // now loop on relations of this RCCL with RTRA
  for( ArteRelationIterator<RCLX_RTRA> relIterat(rclx->rtra);
       relIterat.valid(); relIterat.next()){
    ArtePointer<RTRA> rtra(relIterat);
    cout<<"->Track: ID="<<rtra<<"  Momentum = "<< 1./rtra->pf;
    cout<<endl;
  }
#endif
}


