//                              -*- Mode: C++ -*-
//:ECHitCellArray.C
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 7/4/97
// Last Update ........

//#include <iostream.h>


#ifdef HIST
#  include "ntuple.hh"
#endif
#include "care/care.hh"
#include "care/caredata.hh"
#include "arte/ArteMessage.hh"
#include "care/ECHitCellArray.hh"
#include "care/ECCellID.hh"
#include "care/ECMCcalib.hh"
#include "care/CmdCareDglev.hh"
#include "care/dgcal.hh"
#include "arte/ArteProto.hh"
#include "care/ECDcalIO.hh"
#include "care/EFCompareHitCells.hh"
#include "care/CmdCareCalerr.hh"
#include "care/CmdCareLevel.hh"
#include "auto/ECA.hh"
#include "auto/HITC.hh"
#include "auto/MVEX.hh"
#include "auto/MCME.hh"
#include "auto/MTMT.hh"
#ifdef _OLD_NAMES_
#  include <algo.h>   
#else
#  include <algorithm>
#endif

#ifdef PERFORMANCES
#  include "arte/Chrono.hh"
#  include "arte/ChronoManager.hh"
   Chrono*  conversion=0;
   Chrono*  boardtime=0;
   Chrono*  remHitC=0;
#endif


/*
CLASS
  ECHitCellArray
  Store an array of pointers to the Signal cells (HITC bank).
  It provides the basic functions for the first stage of pattern recognition:
  Filling and Sorting.


*/

/////////////////////// ECHitCellArray static data definition


//--
//  Minimum energy to consider: 150 MeV
Energy_t  ECHitCellArray::minEnergy[6]   = {0.0, 0.050, 0.100, 0.0, 0.0, 0.150};


///////////////////   ECHitCellArray CLASS definition   
 
//--
//  Constructor
ECHitCellArray::ECHitCellArray() : 
  arr_pHITC(), entries(0), virtual_end(arr_pHITC.end())  {}

 
//--
//  Destructor
ECHitCellArray::~ECHitCellArray(){}

//--
//  Fill the array
void ECHitCellArray::fill(){

  int lev = CmdCareDglev::dglevel();  // get the ECAL digi level 

  switch(lev){
  case 0:
    // Standar digitalization for arte-01-05  MCAL -> HITC (Q=energy)
    fill01(lev);
    break;
  case 1:
    // Digitalization for arte >01-05  MCAL -> HITC (Q= ADC channel)
    fill01(lev);
    break;
  case 2:
    // Digitalization till the Second Level Buffer  MCAL -> DCAL
    //
    fill2();
    break;
  default:
    // Error: at least one digitization level must be choosen
    arteMsg<<" DGCAL: wrong digitization level !!";
    arteMsg.error();
    fill01(0);
    break;
  }
  //dump();
}

//--
//  Fill the array
void ECHitCellArray::fill01(int lev){
  ECMCcalib calibration;
  CmdCareCalerr prec;
  ArteTableIterator<HITC> hitcIt;         // create a table iterator on HITC
  start();
  // test if the bank is empty
  if(ArtePointer<HITC>::getLink()()==0) return;
  arr_pHITC.reserve(hitcIt.nrow());   // reserve space for all HITC
  ECCellID cellID;
  Energy_t  calib;
  int ncl;
  CellIdx_t cell;

  for( ; hitcIt.valid(); hitcIt.next()){
    cellID = hitcIt->gcal;
    cell = cellID.cellIdx();
    if(  cell.index()==0){
	cout<<" Warning: gcal=0 for cell "<<hitcIt->gcal<<endl;
    } else {	
    hitcIt->gcal = cell.index(); // use gcal to store the index
    if( lev!=1) {
      // old digitization routine
      ncl = cellID.modulePointer()->ncl;
      calib = calibration[ncl]*prec.factor();
      hitcIt->q *= calib; // do calibrate in energy
      hitcIt->r *= calib; // calibrate error on energy
      if( hitcIt->q> minEnergy[ncl]){ // skip cells with energy 
	                              // less than threshold
	push_back(hitcIt.index());
      }
    } else {
      // new digitization routine
      ArtePointer<ECHC> echc(cell.index()); 
      int adc = (int)  hitcIt->q; // convert to int
      hitcIt->q = dgcalche(cell.index(), adc) * prec.factor();
      if( hitcIt->q > echc->thr1 ){   // skip cells with energy 
	                              // less than threshold
	push_back(hitcIt.index());
      }
    }
    }
  }
}


//--
//  dglev=2 Fill the array and create the HITC bank 
void ECHitCellArray::fill2(){
  int i=0, ix= 0, iy= 0, ndim=0;
  ECCellID  cellid(0);
  //  CmdCareCalerr prec;
  Energy_t ener = 0.0;
  // Coord_t dx,dy,rdim;
  unsigned short channel;
  unsigned short dummy = 0, sfed_id;
  // initialize
  static Coord_t rdim[6], rdx[6][6];
  static int init=0;
  static vector<int> threshs;
  if(init==0){
    Coord_t tmpdx;
    for(int dim=1; dim<6; dim++){
      rdim[dim] = 1./float(dim);
      tmpdx = ECalInfo.xMsize*rdim[dim];
      for(int jx=1; jx<=dim; jx++){
	rdx[dim][jx] = (jx-0.5)*tmpdx;
      }
    }
    init = 1;
    threshs.push_back(10000);
    for(ArteTableIterator<ECHC> echc; echc.valid(); echc.next()){
      threshs.push_back(echc->thr1);
    }
  }
# ifdef PERFORMANCES
  if( conversion==0){
      ChronoManager CM;
      conversion = CM.createChrono("test filling");
      boardtime = CM.createChrono("Board time");
      remHitC =  CM.createChrono("Conversion");
  }
  conversion->start();
# endif
  // start the array
  start();
  arr_pHITC.reserve(ArtePointer<ECBO>::nrow()*32);   // reserve space for all HITC

  // book the maximum number of entries in the HITC bank
  mmbook_(ArtePointer<HITC>::getLink().getRef(), 
	  ArtePointer<ECBO>::nrow()*32);
  ArteTableIterator<HITC> hitc;

# ifdef PERFORMANCES
  conversion->stop();
  boardtime->start();
#endif

  // loop on info in DCAL
  CellIdx_t cellidx;
  int thresh;
  ECDcalIO dcal;
  dcal.start(); 
  for(; dcal.valid()>=5+32; ){
    
    int fed_id = 0; // read the Front End Detector IDentifier

    // read the header
    dcal>> sfed_id >> dummy >> dummy >> dummy >> dummy;
    fed_id = sfed_id & 255;
    // cout<<"Fed id:"<< fed_id;
    // find the related ECBO entry given the fed_id
    ArteTableIterator<ECBO> ecbo(fed_id);
    //    if( fed_id!=0){
    //  for(; ecbo.valid(); ecbo.next()){
    //	if(ecbo->fdid == fed_id) break;
    //  }
    //}
    if( ecbo.valid() && ecbo->stat==1 && fed_id!=0){
      // correct and active ecbo table entry
      for(i=0; i<32; i++){
	if( ecbo->chan[i]!=-1){
	  cellidx = ecbo->chan[i];
	  //	  int thresh = ArtePointer<ECHC>(cellidx.index())->thr1;
	  thresh = threshs[cellidx.index()];
	  // readout channel active
	  // conversion   ADC channel -> energy (GeV)
	  dcal >> channel;
	  //	  cout<<" Channel :"<<i<<"  Value:"<<channel
	  //    <<" Cell:"<<ecbo->chan[i]<<" Cell Thres:"<<thresh<<endl;
	  //	  ener = dgcalche(cellidx.index(), channel)*prec.factor();
      	  if( channel > thresh){
	    ener = dgcalche(cellidx.index(), channel);
	    cellid(cellidx->ceid);
	    cellid.getModuleIndex(ix, iy, ndim);
	    if( ener > minEnergy[ndim]){ // skip cells with energy 
	                              // less than threshold
	      hitc->x = cellid.modulePointer()->xmin+rdx[ndim][ix];
	      hitc->y = cellid.modulePointer()->ymin+rdx[ndim][iy];
	      hitc->q = ener;               // energy
	      hitc->r = 0.0;                // error on energy
	      hitc->gcal = cellidx.index();   // cell ID (GCCE numbering);
	      push_back(hitc.index());
	      hitc.next();
	    }
	  }
	} else {
	  dcal >> channel;
	  // cout<<" No energy";
	}
	//cout<<endl;
      }  
      // skip the trailer
      dcal >> dummy >> dummy;
    } else {
      if( !ecbo.valid()){
	static int invfed = 0x0ECA0004;
	arteMsg<<" Invalid FED ID in DCAL table ";
	arteMsg.warn(invfed);
      }
      for(i=0; i<32+2; i++){  // skip all the table
	dcal>>dummy;
      }
    }
  }
# ifdef PERFORMANCES
  boardtime->stop();
  remHitC->start();
# endif
  // remove the unused entries
  if( CmdCareLevel::level()>2){
    mmbook_(ArtePointer<HITC>::getLink().getRef(), 
	    hitc.index()-ArtePointer<ECBO>::nrow()*32-1);
  }
# ifdef PERFORMANCES
  remHitC->stop();
# endif
}



//--
//  Sort the array
void ECHitCellArray::sortCells(){
  // Sort desceding in energy Hit Cell pointer array with quicksort
  stable_sort(arr_pHITC.begin(), end(), EFCompareHitCells());
}


//-- 
//   For debugging a dump operation is provided. It print on cout
//   all information stored.
void  ECHitCellArray::dump() const{

  int n = size();
 
  cout<<"Dump array ECHitCellArray of "<<size()<<" elements:"<<endl;
  // loop on cells
  cout<<"Index \tCell\tModule \tEnergy    \tPosition"<<endl;
  for(int i=0; i<n; i++){
    ArtePointer<HITC> hitc(arr_pHITC[i]);
    cout<<i+1<<"\t"<<ECCellID(CellIdx_t(hitc->gcal)->ceid)()<<"\t"
	<<hitc->gcal<<"\t"<<hitc->q
	<<"\t"<<hitc->x<<"\t"<<hitc->y<<endl;
  }
  cout<<"-----------------------"<<endl<<endl;
}


////////// Class ECHitCellArray static function definitions


//--
//   Set the minimum energy a cell
//   must have in order to consider for clusterization
void ECHitCellArray::setMinimumEnergy(Energy_t innerCal, 
					 Energy_t middleCal,
					 Energy_t outerCal){
  // set the minimum energy passed
  minEnergy[0] = 0.0;
  minEnergy[1] = outerCal;  // Outer CAL has 1 cell per module
  minEnergy[2] = middleCal; // Middle CAL has 2x2 cells per module
  minEnergy[3] = 0.0;
  minEnergy[4] = 0.0;
  minEnergy[5] = innerCal;  // Inner CAL has 5x5 cells per module
}
			    
//--
//   Printout Information for the current MinimumEnergy data stored
void ECHitCellArray::dumpCutOff(){
 
  arteMsg<<" Cut Off energy for inner, middle and outer ECAL (GeV):";
  arteMsg.allp();
  arteMsg<<minEnergy[5]<<"  "<<minEnergy[2]<<"  "<<minEnergy[1];
  arteMsg.allp();
}



