//                              -*- Mode: C++ -*-
//: EMCellFraction.C
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 28/11/97
// Last Update ........
//

#include "care/EMCellFraction.hh"
#include "care/EMCluster.hh"
#include "care/EAClusterAnalyzer.hh"
#include"care/EDClusterModifier.hh"
#include "arte/ArteMessage.hh"
#include "care/ECCellID.hh"

/**
 Concrete derived class for the hierarchical cluster tree used in
  pattern recognition of ECAL. This class is a "leaf": it does not have
  children, it cannot receive Add, Move or Remove message.
  It is used to store a part of a cell energy.
*/

//////////////  Class EMCellFraction definitions //////////////////////////////

//-- Constructors
//   The construction of EMCellFraction is done referring to some
//   parent. The depth level is adjusted accordingly.
EMCellFraction::EMCellFraction(EMHitCell* hitcell, Energy_t fract ) : 
  EMHitCell(), hitc(hitcell), fraction(fract){
    EMHitCell::pointToCell(hitc->whichBank());
}


//-- Default Constructors
EMCellFraction::EMCellFraction() : EMHitCell(), hitc(0), fraction(0.0){}

//-- Destructor
//   A virtual destructor is needed: but not for this derived class: 
//   it is a dummy one
EMCellFraction::~EMCellFraction(){}


// Status access functions
//--


//--
// Return the Energy of the Cell itself. 
Energy_t  EMCellFraction::maxEnergy() const{
  return energy();
}


//--
// Analysis and traversing of the cluster tree is done by EAClusterAnalyzers
// that are object with definite analyzing properties. 
// A direct call to the DoAnalisys() member of EAClusterAnalyzer for this
// "leaf" is done.
void EMCellFraction::analyzeMe(EAClusterAnalyzer & analyzer) const {
  analyzer.doAnalysis(this);
}

void EMCellFraction::modifyMe(EDClusterModifier & modifier){
  modifier.doModify(this);
}


//--
/// Access the energy of a hit cell
Energy_t  EMCellFraction::energy() const{
  return EMHitCell::energy()*fraction;
}


//--
//  Dump for debugging
void EMCellFraction::dump() const {
  // hitc->dump();
  for(int i=0; i<depthLevel(); i++)
    arteMsg<<"|";
  arteMsg<<"->FracCell: IDX="<<hitc->cellIdx()
	 <<" fraction:"<<fraction<<"   Energy = "<<energy();
  //	 <<" Pos:("<<EMHitCell::x()<<", "<<EMHitCell::y()<<" ) ";
  if(parentCluster==0) {
    arteMsg<<" VOID parent.";
    arteMsg.allp();
  } else {
    arteMsg.allp();
  }
}

