//                              -*- Mode: C++ -*-
//: EMClusterTree.C
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 3/4/97
// Last Update ........
//

#include "care/EMClusterTree.hh"
#include "care/EAClusterAnalyzer.hh"
#include"care/EDClusterModifier.hh"
#include "arte/ArteMessage.hh"
#include "care/ELLUTsManager.hh"
#include "care/ELCellGrid.hh"
#ifdef _OLD_NAMES_
#  include <algo.h>   
#else
#  include <algorithm>
#endif



const int MAXCELLS = 6000;
const int MAXTRACKS =500;
const int MAXFRACT = 500;
//////////////  Class EMClusterTree definitions //////////////////////////////

// Default constructor
EMClusterTree::EMClusterTree() : EMCluster(), 
  subCluster(MAXCELLS), clusCell(MAXCELLS), trackOnEcal(MAXTRACKS),
  cellFrac(MAXFRACT){
  reset();
}


//-- Destructor
//   A virtual destructor is needed
EMClusterTree::~EMClusterTree(){
  // start destroying each object in the list
  // then remove the list
}




//--
// Analysis and traversing of the cluster tree is done by EAClusterAnalyzers
// that are object with definite analyzing properties. 
// EMClusterMembers MUST ignore these properties: in this way a loose coupling
// between EMClusterMembers and EAClusterAnalyzers is achieved and both
// can vary independently during program evolution.
// Examples of EAClusterAnalyzers are those that should return
// total energy of a cluster, the Center of Gravity, the Number of Cells.
void EMClusterTree::analyzeMe(EAClusterAnalyzer & analyzer) const {
  //
  analyzer.doAnalysis(this);
}

void EMClusterTree::modifyMe(EDClusterModifier & modifier){
  //
  modifier.doModify(this);
}



//--
//  Dump for debugging
void EMClusterTree::dump() const {
  arteMsg<<"****************************************************";
  arteMsg.allp();
  arteMsg<<"Cluster List Hypotheses: (H:"<< getHypotheses()
      <<", T:"<<getTracks() 
      <<", C:"<<getChargHyp()
      <<", N:"<<getNeuHyp()<<") ";
  arteMsg.allp();
  EMCluster::dump();
  arteMsg<<"****************************************************";
  arteMsg.allp();
}


//-- Reset all info in the cluster
void  EMClusterTree::reset(){
  EMCluster::reset();
  nextClusIdx =0;
  nextCellIdx = 0;
  nextTrackIdx = 0;
  nextFracIdx = 0;
  // Initialize LUT for cluster pointers
  // assume there are 6000 cells
  if(whichHitCellArray.capacity()<MAXCELLS){
    whichHitCellArray.reserve(MAXCELLS);
  }
}

//-- Start preparing to receive "size" cells
void  EMClusterTree::start(unsigned int size){
  int ncells = MAXCELLS;
  reset();
  fill_n(whichHitCellArray.begin(), ncells, (EMHitCell *)0);
  //
  // Initialize SubCluster and ClusCell arrays
  fill_n(subCluster.begin(),size,EMCluster());
  fill_n(clusCell.begin(),size,EMHitCell());
  fill_n(cellFrac.begin(),size,EMCellFraction());

}


//--
// start a new cluster with a given cell.
// the passed cell is actually MOVED
EMCluster * EMClusterTree::newCluster(EMClusterMember * clusmem){
  // cout<<"start..a new cluster "<<nextClusIdx ;
  subCluster[nextClusIdx].start(clusmem);
  // cout<<"  after start"<<endl;
  EMCluster * newClus = &subCluster[nextClusIdx];
  nextClusIdx++;
  return  newClus;
}


//--
/// get a pointer to the next cell to clusterize
EMHitCell*  EMClusterTree::newHitCell(CellIdx_t currIdx){
  EMHitCell* hitcell = &clusCell[nextCellIdx];
  whichHitCellArray[currIdx.index()] = &clusCell[nextCellIdx];
// update whichHitCellArray
  nextCellIdx++;
  return hitcell;
}

//--
// Store parameters of a track  (called somewhere before caresecondstage) 
void EMClusterTree::newTrack(TrackInfo & trackInfo){
  // fill the parameters for the current track
 
  if(nextTrackIdx<trackOnEcal.size()){
    trackOnEcal[nextTrackIdx++].fill(trackInfo);
  } else if( nextTrackIdx<MAXTRACKS ){
    trackOnEcal.push_back(EMTrackOnEcal());
    trackOnEcal[nextTrackIdx++].fill(trackInfo);
  } else {
    static int StoreWarn = 0x0ECA0006;
    arteMsg<<" Store Track: more than "<<MAXTRACKS
	   <<" tracks. Something wrong?";
    arteMsg.warn(StoreWarn);
  }
}


//--
/// get the space for a new cell fraction 
EMCellFraction* EMClusterTree::newCellFraction(EMHitCell* cell, 
					       Energy_t fraction){
  EMCellFraction* where = 0;
  // fill the parameters for the current cell fraction
  if(nextFracIdx<cellFrac.size()){
    where = cellFrac[nextFracIdx++].pointToCell(cell, fraction);
  } else if( nextFracIdx<MAXFRACT ){
    cellFrac.push_back(EMCellFraction(cell, fraction));
    where = &cellFrac.back();
  } else {
    static int FracWarn = 0x0ECA0007;
    arteMsg<<" Store Cell Fraction: more than "<<MAXFRACT
	   <<" fractional cells. Something wrong?";
    arteMsg.warn(FracWarn);
    where = &cellFrac.back();
  }
  return where;
}


//--
// return the base cluster in which the point on ECAL is
EMHitCell* EMClusterTree::whichHitCell(Coord_t x, Coord_t y) const{

  ELLUTsManager LM;
  ELCellGrid::ErrorTypes ierr;
  // get the number of nearest neibouring cells of the current one
  CellIdx_t ncell = LM.cellsGrid().cellIdx(x,y,ierr);  
  EMHitCell* where = 0;
  if( ierr== ELCellGrid::OK){
    where = whichHitCellArray[ncell.index()];
  }
  return where;
}
