//                              -*- Mode: C++ -*-
//: ESCOGPos.C
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 14/4/97
// Last Update ........
//

#include "care/caredata.hh"
#include "care/EMClusterTree.hh"
#include "care/EMHitCell.hh"
#include "care/ESCOGPos.hh"
#include <math.h>
//#include "ESCOGCorrector.hh"
#include "care/ECCellID.hh"
#include "care/ECStrategyFactory.hh"

class  EMHitCell;
class  EMTrackOnEcal;


/* 
CLASS
  ESCOGPos
  Concrete class for cluster analysis: it evaluates total energy and
  the center of gravity of a cluster
  *


*/
ESPosStrategy* ESCOGPos::peakSt = 0;
//////////////////// ESCOGPos Class Definitions ////////////////

ESCOGPos::ESCOGPos(){
  if( peakSt==0){
    // get the analyzer for the peak
    ECStrategyFactory  straFac;
    peakSt = straFac.PositionStrategy(Peak);
  }
}
ESCOGPos::~ESCOGPos(){}

//--
// ClusterMember Analyzer
void ESCOGPos::doAnalysis(const EMClusterMember * ){
  cout<<" Error Do Analysis called for ClusterMember "<<endl;
}

//--
// TrackOnEcal analysis
void ESCOGPos::doAnalysis(const EMTrackOnEcal *){}

//--
// ClusterTree analysis
void ESCOGPos::doAnalysis(const EMClusterTree *tree){
  const EMCluster* cluster = tree;
  doAnalysis(cluster);
}
 



//--
//  Start a cluster analysis
void ESCOGPos::evaluatePos(const EMCluster* clus){

  
  peakSt->evaluatePos(clus);    // get the peak
  posData = peakSt->getData(); // the peak position is now local
 
  float xpeak = posData.x();
  float ypeak = posData.y();

  currClust =  COGdata();
  doAnalysis(clus);
  float rndata = 1./currClust.energySum;

  //  ESPosition cog;
  posData.setEnergy(currClust.energySum);
  posData.setWidth(currClust.ndata);
  posData.setType(COG);

  float x = currClust.xSum*rndata; // warn: this is the displacement from xpeak
  float y = currClust.ySum*rndata; // warn: this is the displacement from ypeak
  // evaluate covariances
  float vxx = currClust.xxSum*rndata-x*x;
  float vxy = currClust.xySum*rndata-x*y;
  float vyy = currClust.yySum*rndata-y*y;
  // test for covariances be greater than 0.02 
  if( vxx<0.02 )
    vxx =  posData.var(0);
  if(vyy<0.02)
    vyy =  posData.var(2);
  // store covariances
  posData.setVar( vxx, vxy, vyy);
  // evaluate the true cog
  x += xpeak;
  y += ypeak;

  // z dependence
  int ndiv = posData.cell()->ncl();

  static float cnst[6] = {0.0, 16.2, 15.8, 0.0, 0.0, 12.9};
  static float logE[6] = {0.0, 2.89, 2.87, 0.0, 0.0, 0.05};

  float lge = log((double) posData.energy());

  float z = ECalInfo.ecalp->z1+cnst[ndiv]+logE[ndiv]*lge;
  //cout<<" ecalp->z1="<< ECalInfo.ecalp->z1<<" logE="<<lge<<"  Z="<<z<<endl;
  // ok now all the variables are filled
  posData.setPoint(x,y,z);
}


//--
// HitCell Analyzer
void ESCOGPos::doAnalysis(const EMHitCell * hitT){

  Coord_t  dx = hitT->x() - posData.x();
  Coord_t  dy = hitT->y() - posData.y();

  // for a cell add it's energy
  Energy_t  ener = hitT->energy();
  currClust.ndata++;
  currClust.energySum += ener;
  currClust.xSum += dx*ener;
  currClust.ySum += dy*ener;
  currClust.xxSum += dx*dx*ener;
  currClust.xySum += dx*dy*ener;
  currClust.yySum += dy*dy*ener;
}


//--
// Cluster Analyzer
void ESCOGPos::doAnalysis(const EMCluster * cluster){

  // loop on cluster members
  
  for(EMClusterMember* child = cluster->begin(); 
      child!=0;  child = child->next() ){
    child->analyzeMe(*this);
  }
}


