//                              -*- Mode: C++ -*-
//:ECHitCellArray.hh
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 7/4/97
// Last Update ........
//
#ifndef ECHITCELLARRAY_HH_
#define ECHITCELLARRAY_HH_

#include "care/ECALtypes.hh"
#ifdef _OLD_NAMES_
#include <vector.h>
#else
#include <vector>
#endif

/**
class HitCellArray.
  Store an array of pointers to the ECAL cells with signals (HITC bank).
  It provides the basic functions for the first stage of pattern recognition:
  Filling and Sorting.

FILLING
  A loop on all hit cells is done to get (and store) its pointer.
  At the moment, there is a cut on energy constant on all cells.
  It is set to 50 MeV at beginning and can be changed with the
  setCalibration function.
  In future this seems a good place to select for clusterization only cells
  with an energy above a given position dependent minimum.
  For Monte Carlo data there is even the calibration of cells.

SORTING
  A Sorting algorithm is applyed to pointers. The array is sorted in a
  decreasing order in energy. At the moment (7/4/97) the sorting algorithm
  choosen is the STL quicksort.

*/
 
class ECHitCellArray {
public:
  //--
  ///  Constructor
  ECHitCellArray();
  //--
  ///  Destructor
  ~ECHitCellArray();
  /**@name Beheavioral methods */
  //@{ 
  //--
  ///  Fill an array of pointers to the ECAL cells with signals (HITC bank).
  void fill();
  //--
  /**  Sort the array of pointers with respect to the energy measured.
    The final array will be in a decreasing order of measured energy. 
    Sorting is implemented as the STL variant of the quicksort algorithm. */
  void sortCells();
  //@}
  
  /**@name Set/Access methods */
  //@{
  //--
  ///  access the cell to clusterize number index.
  inline int operator[](int index) const;

  //-- 
  ///  return the number of cells to clusterize.
  inline int  size() const;

  //--
  ///   Get the end of the array of pointers.
  inline vector<int >::iterator end();

  //--
  ///   Get the beginning of the array of pointers.
  inline vector<int >::const_iterator begin();

  //--
  ///   Put an entry at the end of the array
  inline void push_back(int value);

  //-- 
  ///   Start the array
  inline void start();
  
  //-- 
  ///   For debugging a dump operation is provided. 
  void  dump() const; 

  //--
  ///   Set the calibration parameters.
  static void setCalibration(Energy_t innerCal, Energy_t middleCal,
			     Energy_t outerCal);
  //--
  ///   Printout Information for the current calibration data stored
  static void dumpCalib();

  //--
  ///   Set the minimum energy a cell
  ///   must have in order to consider for clusterization.
  static void setMinimumEnergy(Energy_t innerCal, Energy_t middleCal,
			       Energy_t outerCal);

  //--
  //   Get the current minimum energies
  static void getMinims( Energy_t & inn, Energy_t & midd, Energy_t &out) {
        inn = minEnergy[5]; midd = minEnergy[2]; out = minEnergy[1];
  }

  //--
  ///   Printout Information for the current MinimumEnergy data stored
  static void dumpCutOff();
  //@}

protected:
  //--
  ///   the array of HITC pointers
  vector<int > arr_pHITC;
  //--
  ///   the number of stored entries
  unsigned int entries;
  //--
  ///   the pointer to the last+1 entry
  vector<int >::iterator virtual_end;

private:


  //--
  ///  Fill an array of pointers to the ECAL cells with signals (HITC bank).
  void fill01(int lev);
  void fill2();
  //--
  ///  Minimum energy to consider
  static Energy_t  minEnergy[6];
 
};

#include "care/ECHitCellArray.icc"
#endif  // ECHITCELLARRAY_HH_
