/*
*************************************************************************
**
** Description: "Cebra Memory Package" - replacement for MZ-package
**              in ARTE program.
**              Use C memory allocation.
**		See function description in cmbook.h
**
** Language:	C
**
** Called from: MM package.
**
** Created: 	20 Sep 97
** Updated:  
**
** Author:  	Sergey Moukhine
**          	HERA-B Collaboration
**          	JINR-Dubna
**          	Russia
**
*************************************************************************
*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "cmbook.h"


#ifdef  __cplusplus
extern "C" {
#endif

/*
*************************************************************************
**		Static variables
*************************************************************************
*/

#define		MAX_BANK_SIZE	1000000
#define		MAX_ERR_SIZE	256

static int*		iqAddress = NULL;

static char		errMessage[MAX_ERR_SIZE];
static char		funName[20];

static void (*errFun)(int *, char *) = cmerror_;

/*
*************************************************************************
**		Internal function prototypes
*************************************************************************
*/

static void Error(int *ErrPtr, int ier);
static void ErrorI(int *ErrPtr, int ier, int i);
static void ErrorPrefix(void);
static char* GetErrorMess(int ier);

/*
*************************************************************************
**		Macro implementation
*************************************************************************
*/

#define	GET_IQ_INDEX(ID) \
   (int)(((long)(ID)-(long)(iqAddress))/sizeof(int)+1)

#define BANK_BY_INDEX(J) \
   (BankT *)((J)*sizeof(int)+((long)iqAddress)-sizeof(BankT))

#define	GET_IQ_CONT(IX) \
   (*(iqAddress+IX-1))

#define	SET_IQ_CONT(IX,V) \
   *(iqAddress+IX-1) = V;

#define CHK_BANK_STRUCT(B,ID) \
   (B->len < sizeof(BankT) ||				\
    B->len > MAX_BANK_SIZE || B->nd < 0 || 		\
    B->origin != ID)

/*
*************************************************************************
**		Function implementation
*************************************************************************
*/

/*---------------------------------------------------------------------*/
/*
** Set error function. Must be called from C only.
*/

void cmSetError(void (*funAddr)(int *, char *))
{
errFun = funAddr;
}

/*---------------------------------------------------------------------*/

void cminit_(int *iq)
{
strcpy(funName,"CMINIT");

if (iqAddress != NULL)
   {
   static int errPtr;
   
   Error(&errPtr,3);
   return;
   }
iqAddress = iq;
}

/*---------------------------------------------------------------------*/

void cmbook_(int *id, int *nd)
{
BankT *bankPtr;
int   len;

strcpy(funName,"CMBOOK");

if (iqAddress == NULL)
   {
   static int errPtr;
   
   Error(&errPtr,4);
   return;
   }

/* check ND */

if (*nd < 0 || *nd > MAX_BANK_SIZE)
   {
   static int errPtr;
   
   Error(&errPtr,6);
   return;
   }
   
/* create bank */

len  = sizeof(BankT) + (*nd)*sizeof(int);

bankPtr = (BankT *)malloc(len);
if (bankPtr == NULL)
   {
   static int errPtr;
   
   Error(&errPtr,1);
   return;
   }
bankPtr->len    = len;
bankPtr->origin = GET_IQ_INDEX(id);
bankPtr->nd     = *nd;
bankPtr->status = 0;		/* this word is not used */

/* clear bank */

memset(&bankPtr->status,0,(*nd+1)*sizeof(int));

/* return bank address */

*id = GET_IQ_INDEX(&(bankPtr->status));

}

/*---------------------------------------------------------------------*/

void cmpush_(int *id, int *nd)
{
BankT *bankPtr;
int   len;
int   idIndex;
int   ndNew;
int   ndOld;

strcpy(funName,"CMPUSH");

idIndex = GET_IQ_INDEX(id);

if (*nd == 0)
   return;
if (*id == 0)
   {
   static int errPtr;
   
   ErrorI(&errPtr,7,idIndex);
   return;
   }

if (iqAddress == NULL)
   {
   static int errPtr;
   
   Error(&errPtr,4);
   return;
   }

/* Get bank pointer */

bankPtr = BANK_BY_INDEX(*id);
if (bankPtr == NULL)
   {
   static int errPtr;
   
   ErrorI(&errPtr,7,idIndex);
   return;
   }

/* check bank structure */

if (CHK_BANK_STRUCT(bankPtr,idIndex))
   {
   static int errPtr;
   
   ErrorI(&errPtr,5,idIndex);
   return;
   }
   
/* change bank length */

ndOld = bankPtr->nd;
ndNew = ndOld + *nd;
if (ndNew < 0)
   ndNew = 0;

if (ndNew > MAX_BANK_SIZE)
   {
   static int errPtr;
   
   Error(&errPtr,6);
   return;
   }

len  = sizeof(BankT) + (ndNew)*sizeof(int);
if (bankPtr->len < len)
   {
   /* re-allocate memory */

   bankPtr = (BankT *)realloc(bankPtr,len);
   if (bankPtr == NULL)
      {
      static int errPtr;
   
      Error(&errPtr,1);
      *id = 0;
      return;
      }
   bankPtr->len = len;
   }

bankPtr->nd = ndNew;


/* clear bank */

if (*nd > 0)
   memset((char *)&(bankPtr->status)+(ndOld+1)*sizeof(int),0,
          (*nd)*sizeof(int));

/* return bank address */

*id = GET_IQ_INDEX(&(bankPtr->status));
}

/*---------------------------------------------------------------------*/

void cmfree_(int *id)
{
BankT *bankPtr;
int   len;

strcpy(funName,"CMFREE");

if (*id == 0)
   return;

if (iqAddress == NULL)
   {
   static int errPtr;
   
   Error(&errPtr,4);
   return;
   }

/* Get bank pointer */

bankPtr = BANK_BY_INDEX(*id);
if (bankPtr == NULL)
   return;

/* free unused space in the bank */

len  = sizeof(BankT) + (bankPtr->nd)*sizeof(int);

if (len != bankPtr->len)
   {
   bankPtr = (BankT *)realloc(bankPtr,len);
   if (bankPtr == NULL)
      {
      static int errPtr;
   
      Error(&errPtr,1);
      *id = 0;
      return;
      }
   }
bankPtr->len = len;

/* return bank address */

*id = GET_IQ_INDEX(&(bankPtr->status));

}

/*---------------------------------------------------------------------*/

void cmprint_(int *id)
{
BankT *bank;
int   idIndex;

idIndex = GET_IQ_INDEX(id);

/* print bank */

printf("======= Bank:  ID = %5d  NID = %5d", *id, idIndex);
if (*id == 0)
   {
   printf("\n");
   return;
   }
   
bank = (BankT *)BANK_BY_INDEX(*id);
printf("  ND = %5d  ORIGIN = %5d  LEN = %5d\n",
        bank->nd, bank->origin, bank->len);
}

/*
*************************************************************************
**		Error handling
*************************************************************************
*/

/*---------------------------------------------------------------------*/

static void Error(int *errPtr,
                  int ier	/*  case of error	*/
                 )
{
char *s = errMessage;

ErrorPrefix();
sprintf(strlen(s)+s, GetErrorMess(ier));
*errPtr = ier;
errFun(errPtr, s);
}

/*---------------------------------------------------------------------*/

static void ErrorI(int *errPtr,
		   int ier,	/*  case of error	*/
		   int i	/*  data		*/
                  )
{
char *s = errMessage;

ErrorPrefix();
sprintf(strlen(s)+s, GetErrorMess(ier), i);
*errPtr = ier;
errFun(errPtr, s);
}

/*---------------------------------------------------------------------*/

static void ErrorPrefix()
{
sprintf(errMessage, "Error in %s: ", funName);
}

/*---------------------------------------------------------------------*/

static char* GetErrorMess(
                         int ier	/*  case of error	*/
                         )
{
  static char text[MAX_ERR_SIZE];

  switch (ier)
    {
      case 1:
        strcpy(text, "not enought memory");
        break;
	
      case 2:
        strcpy(text, "CM-package tables are broken?: call to %s");
        break;
        
      case 3:
        strcpy(text, "CMINIT called twice");
        break;
	
      case 4:
        strcpy(text, "operations before initialization");
        break;
	
      case 5:
        strcpy(text, "bank at NID=%d has a wrong structure");
        break;
	
      case 6:
        strcpy(text, "wrong parameters");
        break;
	
      case 7:
        strcpy(text, "bank at NID=%d does not exist");
        break;
	
      default:
        sprintf(text, "unknown CM-package error: %d", ier);
    }
   return(text);
}

/* ///////////////////////// DEBUG /////////////////////////// */

void cmcopy_(void *s1, void *s2)
{
memmove(s1,s2,4);
}

#ifdef  __cplusplus
      }
#endif
