/*
*************************************************************************
**
** Description: class EvdirSqlCompiler:
**	        Compiler Main functions.
**
** Language:	C++
**
** Created: 20 Aug 97
** Updated:  
**
** Author:  Sergey Moukhine
**          HERA-B Collaboration
**          JINR-Dubna
**          Russia
**
*************************************************************************
*/

#include <evdir/EvdTable.h>
#include <evdir/EvdCommon.h>
#include <evdir/EvdCompiler.h>
#include <evdir/EvdPrint.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

static EvdirSqlPrint prt;

/////////////////////////////////////////////////////////////////////////

int EvdirSqlCompiler::SelectCompiler(
                char *body,		// input text		
                int  nFields,		// n of fields in "fields"
                NameAndValueT *fields,	// names and types of fields
                char **error,		// error message	
                char **tcode		// T-code		
                                      )
/* Return value: OK or FAILURE */
{

/*
**   COMPILE
*/

LexInit(LxCompile);
lex.curSymb = body;

*error = CompilerData.errorBuffer;
*tcode = NULL;

if (setjmp(CompilerData.longJmpBuf) != 0)
   {
   /*
   **   if error somewhere in the compiler
   */
   TableDrop(TbAll);
   return(FAILURE);
   }

CommonInitialization(nFields, fields);

if (yyparse())
   Error(5);

#ifdef DEBUG
		TablePrint(TbName);
		TablePrint(TbField);
		TablePrint(TbDate);
		TablePrint(TbTime);
		TablePrint(TbNumber);
		TablePrint(TbPrimary);
		TablePrint(TbExpr);
#endif

/*
**   FILL RUN TABLE
*/

LexInit(LxFillRunTable);
lex.curSymb = body;

*error = CompilerData.errorBuffer;

if (setjmp(CompilerData.longJmpBuf) != 0)
   {
   /*
   **   if error => go through ALL runs
   */
     TableDrop(TbRun);
     CompilerData.errorBuffer[0] = '\0';
     goto fill_tcode;
   }

if (yyparse())
   Error(5);


fill_tcode:

#ifdef DEBUG
		TablePrint(TbRun);
#endif

FillOutputData(tcode);
TableDrop(TbAll);

return(OK);
}

/////////////////////////////////////////////////////////////////////////

/*
**	INITIALIZATION
*/

void EvdirSqlCompiler::CommonInitialization(
                int  nFields,		// n off fields in "fields"
                NameAndValueT *fields	// names and types of fields
				           )
{
CompilerData.errorBuffer[0] = '\0';

tab = &CompilerData.table;
rec = &CompilerData.record;
 
TableInit(TbAll);

for (int i=0; i<nFields; i++)
   {
   NameAndValueT f;
   
   strcpy(f.name,fields[i].name);
   for (char* s=f.name; *s; s++)
      *s = toupper(*s);
   CompilerData.record.field.field = ADD_NAME(f.name);
   CompilerData.record.field.kind = fields[i].kind;
   ADD_RECORD(TbField);
   }

}

/////////////////////////////////////////////////////////////////////////

void EvdirSqlCompiler::FillOutputData(char **tcode)
{
/*
**	T-code will be helpfull if Compiler and Interpreter
**	will run as separate processes. In this case T-code
**	can be moved trough the pipe.
**
**	T-code structure:
**
**	Total_length_in_bytes
**	N_names,	(list of (strlen+1) + 0-term. strings: names);
**	N_numbers,	(list of NumberT);
**	N_dates,	(list of DateT);
**	N_times,	(list of TimeT);
**	N_fields,	(list of FieldT);
**	N_primaries,	(list of PrimaryT);
**	N_exprs,	(list of ExprT);
**	N_runs,		(list of RunPairT);
**	
*/

int      n;
int      i;
int      len;
int      size;
TyRun    *t;

//	Calculate length of T-code

len = (1+8)*sizeof(int);

n = NUSED(TbName);
for (i=0; i<n; i++)
   len += sizeof(int) + strlen(GET_RECORD(TyName,TbName,i)->text) +1;
len += NUSED(TbNumber)*sizeof(TyNumber);
len += NUSED(TbDate)*sizeof(TyDate);
len += NUSED(TbTime)*sizeof(TyTime);
len += NUSED(TbField)*sizeof(TyField);
len += NUSED(TbPrimary)*sizeof(TyPrimary);
len += NUSED(TbExpr)*sizeof(TyExpr);

n = NUSED(TbRun);
if (n != 0)
   {
   t = GET_RECORD(TyRun,TbRun,0);
   memcpy(&rec->runTab,t,sizeof(TyRun));

   len += NUSED(TbRunTab)*sizeof(TyRunTab);
   }

*tcode = new char[len];
if (*tcode == NULL)
   Error(1);

//	Fill T-code

char* b = *tcode;
memcpy(b, (const void*)&len, sizeof(int));
b += sizeof(int);

n = NUSED(TbName);
memcpy(b, (const void*)&n, sizeof(int));
b += sizeof(int);
for (i=0; i<n; i++)
   {
   char *s;
   s = GET_RECORD(TyName,TbName,i)->text;
   size = strlen(s) + 1;
   memcpy(b, (const void*)&size, sizeof(int));
   b += sizeof(int);
   memcpy(b, s, size);
   b += size;
   }

n = NUSED(TbNumber);
memcpy(b, (const void*)&n, sizeof(int));
b += sizeof(int);
for (i=0; i<n; i++)
   {
   size = sizeof(TyNumber);
   memcpy(b, (const void*)GET_RECORD(TyNumber,TbNumber,i), size);
   b += size;
   }

n = NUSED(TbDate);
memcpy(b, (const void*)&n, sizeof(int));
b += sizeof(int);
for (i=0; i<n; i++)
   {
   size = sizeof(TyDate);
   memcpy(b, (const void*)GET_RECORD(TyDate,TbDate,i), size);
   b += size;
   }

n = NUSED(TbTime);
memcpy(b, (const void*)&n, sizeof(int));
b += sizeof(int);
for (i=0; i<n; i++)
   {
   size = sizeof(TyTime);
   memcpy(b, (const void*)GET_RECORD(TyTime,TbTime,i), size);
   b += size;
   }

n = NUSED(TbField);
memcpy(b, (const void*)&n, sizeof(int));
b += sizeof(int);
for (i=0; i<n; i++)
   {
   size = sizeof(TyField);
   memcpy(b, (const void*)GET_RECORD(TyField,TbField,i), size);
   b += size;
   }

n = NUSED(TbPrimary);
memcpy(b, (const void*)&n, sizeof(int));
b += sizeof(int);
for (i=0; i<n; i++)
   {
   size = sizeof(TyPrimary);
   memcpy(b, (const void*)GET_RECORD(TyPrimary,TbPrimary,i), size);
   b += size;
   }

n = NUSED(TbExpr);
memcpy(b, (const void*)&n, sizeof(int));
b += sizeof(int);
for (i=0; i<n; i++)
   {
   size = sizeof(TyExpr);
   memcpy(b, (const void*)GET_RECORD(TyExpr,TbExpr,i), size);
   b += size;
   }

n = 0;
if (NUSED(TbRun) != 0)
   n = NUSED(TbRunTab);
memcpy(b, (const void*)&n, sizeof(int));
b += sizeof(int);
for (i=0; i<n; i++)
   {
   size = sizeof(TyRunTab);
   memcpy(b, (const void*)GET_RECORD(TyRunTab,TbRunTab,i), size);
   b += size;
   }
}

/////////////////////////////////////////////////////////////////////////

int EvdirSqlCompiler::GetInt(int ind)
{
return(GET_RECORD(TyNumber,TbNumber,ind)->value.i);
}

/////////////////////////////////////////////////////////////////////////

double EvdirSqlCompiler::GetDouble(int ind)
{
return(GET_RECORD(TyNumber,TbNumber,ind)->value.d);
}

/////////////////////////////////////////////////////////////////////////

int EvdirSqlCompiler::RunOperation(OperationKind op, int p1, int p2)
{
//  NOT, AND, OR operations with Run table
//  Operation OpPlus means reduce the number of entries if possible

int ind1 = MIN(p1,p2);
int ind2 = MAX(p1,p2);

EvdirSqlTable *t1 = GET_RECORD(TyRun,TbRun,ind1);
EvdirSqlTable *t2 = GET_RECORD(TyRun,TbRun,ind2);
RunPairT      *r1;
RunPairT      *r2;

int i1;
int n1;
int i2;
int n2;

switch (op)
   {
   case OpNot:
   ///////////
   
      TableInit(TbRunTab);
      n1 = t1->Nused();
      r1 = (RunPairT *)t1->Get(0);
      rec->run.beg = FIRST_RUN_NUMBER;
      rec->run.end = FIRST_RUN_NUMBER;
      for (i1=0; i1<n1; i1++)
         {
         rec->run.end = r1[i1].beg - 1;
         if (rec->run.beg <= rec->run.end)
            ADD_RECORD(TbRunTab);
         rec->run.beg = r1[i1].end + 1;
         }
      rec->run.end = LAST_RUN_NUMBER;
      if (rec->run.beg <= LAST_RUN_NUMBER)
         ADD_RECORD(TbRunTab);
      
      t1->Drop();
      tab->run.Replace(&rec->runTab,ind1);
      TableInit(TbRunTab);
      return(ind1);

   case OpAnd:
   ///////////
   
      TableInit(TbRunTab);
      n1 = t1->Nused();
      r1 = (RunPairT *)t1->Get(0);
      n2 = t2->Nused();
      r2 = (RunPairT *)t2->Get(0);
      i1 = 0;
      i2 = 0;
      while (i1<n1 && i2<n2)
         {
         if (r1[i1].end < r2[i2].beg)
            {
            i1++;
            continue;
            }
         if (r2[i2].end < r1[i1].beg)
            {
            i2++;
            continue;
            }
         rec->run.beg = MAX(r1[i1].beg,r2[i2].beg);
         rec->run.end = MIN(r1[i1].end,r2[i2].end);
         ADD_RECORD(TbRunTab);
         if (r1[i1].end <= r2[i2].end)
            {
            i1++;
            continue;
            }
         if (r2[i2].end <= r1[i1].end)
            i2++;
         }
      
      t1->Drop();
      tab->run.Replace(&rec->runTab,ind1);
      TableInit(TbRunTab);
      if (ind2+1 == tab->run.Nused())
         tab->run.Pop();
      return(ind1);
      
   case OpOr:
   //////////
   
      TableInit(TbRunTab);
      n1 = t1->Nused();
      r1 = (RunPairT *)t1->Get(0);
      n2 = t2->Nused();
      r2 = (RunPairT *)t2->Get(0);
      i1 = 0;
      i2 = 0;
      while (i1<n1 && i2<n2)
         {
//       end before start

         if (r1[i1].end < r2[i2].beg)
            {
            rec->run.beg = r1[i1].beg;
            rec->run.end = r1[i1].end;
            ADD_RECORD(TbRunTab);
            i1++;
            continue;
            }
         if (r2[i2].end < r1[i1].beg)
            {
            rec->run.beg = r2[i2].beg;
            rec->run.end = r2[i2].end;
            ADD_RECORD(TbRunTab);
            i2++;
            continue;
            }

//       in the middle

         rec->run.beg = MIN(r1[i1].beg,r2[i2].beg);
         rec->run.end = MAX(r1[i1].end,r2[i2].end);
         while ((i1<n1 && i2<n2) &&
                ((r1[i1].end <= r2[i2].end && r1[i1].end >= r2[i2].beg) ||
                 (r2[i2].end <= r1[i1].end && r2[i2].end >= r1[i1].beg)))
            {
            rec->run.end = MAX(r1[i1].end,r2[i2].end);
            if (r1[i1].end <= r2[i2].end)
               {
               i1++;
               continue;
               }
            if (r2[i2].end <= r1[i1].end)
               {
               i2++;
               continue;
               }
            }

         if (i1 == n1)
            i2++;
         else
         if (i2 == n2)
            i1++;
         else
         if (r1[i1].end < r2[i2].beg)
            i1++;
         else
         if (r2[i2].end < r1[i1].beg)
            i2++;
        
         ADD_RECORD(TbRunTab);
         }

// tail

      if (i1 == n1)
         for (;i2<n2; i2++)
            {
            rec->run.beg = r2[i2].beg;
            rec->run.end = r2[i2].end;
            ADD_RECORD(TbRunTab);            
            }
      else
      if (i2 == n2)
         for (;i1<n1; i1++)
            {
            rec->run.beg = r1[i1].beg;
            rec->run.end = r1[i1].end;
            ADD_RECORD(TbRunTab);            
            }
      
      t1->Drop();
      tab->run.Replace(&rec->runTab,ind1);
      TableInit(TbRunTab);
      if (ind2+1 == tab->run.Nused())
         tab->run.Pop();
      return(ind1);

   case OpPlus:
   ////////////
   
   // Operation "OpPlus" means reduce the number of entries if possible

      TableInit(TbRunTab);
      n1 = t1->Nused();
      if (n1 <= 1)
         return(ind1);
      r1 = (RunPairT *)t1->Get(0);

      rec->run.beg = r1[0].beg;
      rec->run.end = r1[0].end;
      for (i1=1; i1<n1; i1++)
         {
         if (rec->run.end == r1[i1].beg-1)
            {
            rec->run.end = r1[i1].end;
            }
         else
            {
            ADD_RECORD(TbRunTab);
            rec->run.beg = r1[i1].beg;
            rec->run.end = r1[i1].end;
            }
         }
      ADD_RECORD(TbRunTab);

      t1->Drop();
      tab->run.Replace(&rec->runTab,ind1);
      TableInit(TbRunTab);
      return(ind1);

   default:
   ////////
   
      ErrorS(7,"TypeOfOperation");
      return(FAILURE);  // this code will never reached
   }
}

/////////////////////////////////////////////////////////////////////////

ValueKind EvdirSqlCompiler::TypeOfPrimary(int ind)
{
PrimaryT *pr;
NumberT  *nu;
PrimaryKind kind;

pr = GET_RECORD(TyPrimary,TbPrimary,ind);
kind = pr->kind;
switch (kind)
   {
   case PrNumber:
      nu = GET_RECORD(TyNumber,TbNumber,pr->link);
      if (nu->kind == NuInt)
         return(ValInt);
      else
         return(ValDouble);
      
   case PrDate:
      return(ValDate);
      
   case PrTime:
      return(ValTime);
      
   case PrField:
      return(GET_RECORD(TyField,TbField,pr->link)->kind);
      
   default:
      return(ValUnknown);
   }
}

/////////////////////////////////////////////////////////////////////////

ValueKind EvdirSqlCompiler::TypeOfOperation(OperationKind op,
				 	    int	          p1,
				 	    int           p2
					   )
{
ValueKind v1 = (ValueKind)p1;
ValueKind v2 = (ValueKind)p2;

switch(op)
   {
   case OpUminus:
      switch (v1)
         {
         case ValInt:
         case ValDouble:
            return(v1);
         default:
            ErrorSS(15,"\"-\"",prt.NameOfValueKind(v1));
         }

   case OpUplus:
      return(v1);
      
   case OpNot:
      switch (v1)
         {
         case ValBool:
            return(v1);
         default:
            ErrorSS(15,"NOT",prt.NameOfValueKind(v1));
         }

   case OpBit:
      switch (v1)
         {
         case ValInt:
         case ValDouble:
            return(ValBit);
         default:
            ErrorSS(15,"\"#\"",prt.NameOfValueKind(v1));
         }

   case OpPlus:
   case OpMinus:
   case OpMult:
   case OpDiv:
      switch (v1)
         {
         case ValInt:
            switch (v2)
               {
               case ValInt:
                  return(ValInt);
               case ValDouble:
                   return(ValDouble);
               default:
                  ErrorS(14,prt.NameOfValueKind(v2));
               }
         case ValDouble:
            switch (v2)
               {
               case ValInt:
               case ValDouble:
                  return(ValDouble);
               default:
                  ErrorS(14,prt.NameOfValueKind(v2));
               }
         default:
            ErrorS(14,prt.NameOfValueKind(v1));
         }

   case OpLt:
   case OpLe:
   case OpGt:
   case OpGe:
   case OpEq:
   case OpNe:
      switch (v1)
         {
         case ValInt:
         case ValDouble:
            switch (v2)
               {
               case ValInt:
               case ValDouble:
                  return(ValBool);
               case ValBit:
                  if (v1 == ValInt)
                     return(ValBool);
                  else
                     ErrorS(16,prt.NameOfValueKind(v1));
               default:
                  ErrorSS(15,"LT,LE,GT,GE,EQ,NE",
                                      prt.NameOfValueKind(v2));
               }
         case ValTime:
            if (v2 == ValTime)
               return(ValBool);
            else
               ErrorSS(15,"LT,LE,GT,GE,EQ,NE",
                                   prt.NameOfValueKind(v2));
            
         case ValDate:
            if (v2 == ValDate)
               return(ValBool);
            else
               ErrorSS(15,"LT,LE,GT,GE,EQ,NE",
                                   prt.NameOfValueKind(v2));
            
         default:
            ErrorSS(15,"LT,LE,GT,GE,EQ,NE",
                                prt.NameOfValueKind(v1));
         }

   case OpAnd:
      if (v1 == ValBool && v2 == ValBool)
         return(v1);
      if (v1 != ValBool)
         ErrorSS(15,"AND", prt.NameOfValueKind(v1));
      if (v2 != ValBool)
         ErrorSS(15,"AND", prt.NameOfValueKind(v2));

   case OpOr:
      if (v1 == ValBool && v2 == ValBool)
         return(v1);
      if (v1 != ValBool)
         ErrorSS(15,"OR", prt.NameOfValueKind(v1));
      if (v2 != ValBool)
         ErrorSS(15,"OR", prt.NameOfValueKind(v2));

   case OpIn:
   case OpNotIn:
      switch (v1)
         {
         case ValInt:
         case ValDouble:
            switch (v2)
               {
               case ValInt:
               case ValDouble:
                  return(ValBool);
               case ValBit:
                  if (v1 == ValInt)
                     return(ValBool);
                  else
                     ErrorS(16,prt.NameOfValueKind(v1));
               default:
                  ErrorSS(15,"IN",prt.NameOfValueKind(v2));
               }
         case ValTime:
            if (v2 == ValTime)
               return(ValBool);
            else
               ErrorSS(15,"IN",prt.NameOfValueKind(v2));
            
         case ValDate:
            if (v2 == ValDate)
               return(ValBool);
            else
               ErrorSS(15,"IN",prt.NameOfValueKind(v2));
            
         default:
            ErrorSS(15,"IN",prt.NameOfValueKind(v1));
         }

   default:
     ErrorS(7,"TypeOfOperation");
   
   }

return(ValUnknown); // This operator never reached; used to avoid warnings 
}

/////////////////////////////////////////////////////////////////////////

ValueKind EvdirSqlCompiler::TypeOfOpBetw(
				 	 int           p1,
				 	 int           p2,
					 int           p3
				        )
{
ValueKind v1 = (ValueKind)p1;
ValueKind v2 = (ValueKind)p2;
ValueKind v3 = (ValueKind)p3;

if (!((v2 == ValInt || v2 == ValDouble) &&
      (v3 == ValInt || v3 == ValDouble))
    && (v2 != v3))
   Error(17);
switch (v1)
   {
   case ValInt:
   case ValDouble:
      switch (v2)
         {
         case ValInt:
         case ValDouble:
            return(ValBool);
         case ValBit:
            if (v1 == ValInt)
               return(ValBool);
            else
               ErrorS(16,prt.NameOfValueKind(v1));
         default:
            ErrorSS(15,"BETWEEN",prt.NameOfValueKind(v2));
         }
   case ValTime:
      if (v2 == ValTime)
         return(ValBool);
      else
         ErrorSS(15,"BETWEEN",prt.NameOfValueKind(v2));
            
   case ValDate:
      if (v2 == ValDate)
         return(ValBool);
      else
         ErrorSS(15,"BETWEEN",prt.NameOfValueKind(v2));
            
   default:
      ErrorSS(15,"BETWEEN",prt.NameOfValueKind(v1));
   }
return(ValBool);
}

