/*
*************************************************************************
**
** Description: header file for class EvdirSqlCompiler
**
** Created: 20 Aug 97
** Updated:  
**
** Author:  Sergey Moukhine
**          HERA-B Collaboration
**          JINR-Dubna
**          Russia
**
*************************************************************************
*/
#ifndef		EVDIR_SQL_COMPILER_H
#define		EVDIR_SQL_COMPILER_H

#include <setjmp.h>

#define FIRST_RUN_NUMBER	1
#define LAST_RUN_NUMBER		99999999

/*--------------------------------------------------------------*/
/*	Lexic types definitions					*/
/*--------------------------------------------------------------*/

#define  LEX_TEXT_SIZE   1023
#define  LEX_NUMBER_LEN  18
#define  LEX_STRING_LEN  LEX_TEXT_SIZE
#define  LEX_NAME_LEN    MAX_LEN_OF_NAME

#define  TOKEN_EOF	-1

typedef enum
{
  LxCompile,
  LxFillRunTable
} LexStartKind;

typedef struct {
   char text[LEX_TEXT_SIZE + 1];/*  input line text		*/
   int  len;			/*  len of text (-1 = EOF)	*/
   int  first;			/*  1-st char (index)		*/
   int  last;			/*  last char (index)		*/
   int  init;			/*  init flag			*/
   LexStartKind kind;		/*  kind of job to do		*/
   int  token;                  /*  token			*/
   char name[LEX_NAME_LEN + 1];	/*  current name		*/
   int  value;                  /*  current value		*/
   char *curSymb;               /*  cur symb in input buffer	*/
} LexDataT;

/*--------------------------------------------------------------*/
/*	Translator types definitions				*/
/*--------------------------------------------------------------*/

typedef struct {
EvdirSqlTable			number;
EvdirSqlTable			name;
EvdirSqlTable			date;
EvdirSqlTable			time;
EvdirSqlTable			field;
EvdirSqlTable			primary;
EvdirSqlTable			expr;
EvdirSqlTable			run;
} TranTablesT;

typedef struct {
NumberT			number;
NameT			name;
DateT			date;
TimeT			time;
FieldT			field;
PrimaryT		primary;
ExprT			expr;
RunPairT		run;
EvdirSqlTable		runTab;
} TranRecordsT;

typedef struct {
TranTablesT		table;
TranRecordsT		record;

char			errorBuffer[LEX_TEXT_SIZE*2];
jmp_buf			longJmpBuf;
} TranDataT;


/*--------------------------------------------------------------*/
/*	Storage types						*/
/*--------------------------------------------------------------*/

typedef enum {

TbNumber,
TbName,
TbPrimary,
TbExpr,
TbDate,
TbTime,
TbField,
TbRun,		/* Table of tables of run pairs	*/
TbRunTab,
TbAll
} TranTablesKind;

/*--------------------------------------------------------------*/
/*	Translator data						*/
/*--------------------------------------------------------------*/

class EvdirSqlCompiler {

private:
////////

TranDataT		CompilerData;
LexDataT		lex;

TranTablesT*		tab;
TranRecordsT*		rec;

//-------------------------------------

void  TranError(void);
char* GetErrorMess(int ind);
void  EmergencyExit(void);

void LexCheckCharSet();
int  LexRead(char *buf, int size);
void LexSkipBlanks();
int  LexNext(int* yylval);

void* MyAlloc(int size);
int   IndName(char *name);

int GetInt(int ind);
double GetDouble(int ind);

int RunOperation(OperationKind op, int p1, int p2);

ValueKind TypeOfPrimary(int ind);
ValueKind TypeOfOperation(OperationKind op,
			  int	        p1,
			  int           p2
			 );
ValueKind TypeOfOpBetw   (
			  int           p1,
			  int           p2,
			  int           p3
			  );

void CommonInitialization(int, NameAndValueT*);
void FillOutputData(char **tcode);


public:
////////

/*------------------------------------------------------*/
/*	Storage functions				*/
/*------------------------------------------------------*/

void TableInit(TranTablesKind tab);
void TableDrop(TranTablesKind tab);
int  TableAdd(TranTablesKind tab, void *d);
int  TableN(TranTablesKind tab);
void *TableGet(TranTablesKind tab, int i);
char *TableName(TranTablesKind tab, int i);
int  TableInd(TranTablesKind tab, int v);
int  TableFind(TranTablesKind tab);
int  GetIndByName(TranTablesKind tab, char *n);
void TablePrint(TranTablesKind tabKind);

/*
**--------------------------------------------------------------
**	Error Functions 					
**--------------------------------------------------------------
*/

void Error(int ind);
void ErrorI(int ind, int i);
void ErrorS(int ind, const char *s);
void ErrorIS(int ind, int i, const char *s);
void ErrorSS(int ind, const char *s1, const char *s2);


/*-------------------------------------------------------------   
**	Compiler Lexic Analyser
**-------------------------------------------------------------
*/

void LexInit(LexStartKind);
int  yylex(int*);
int  yyparse(void);
void yyerror(const char *err);

/*
**--------------------------------------------------------------
**		Compiler functions
**--------------------------------------------------------------
*/

int SelectCompiler(
                char *body,		// input text		
                int  nFields,		// n of fields in "fields"
                NameAndValueT *fields,	// names and types of fields
                char **error,		// error message	
                char **tcode		// T-code		
                  );
/*
** Return value: OK or FAILURE
**
** Output:
**      error - points to internal static buffer
**      tcode - NULL or must be freed in calling function (?)
*/


};

/*------------------------------------------------------*/
/*	Storage records types				*/
/*------------------------------------------------------*/

#define TyNumber		NumberT
#define TyName			NameT
#define TyPrimary		PrimaryT
#define TyDate			DateT
#define TyTime			TimeT
#define TyField			FieldT
#define TyExpr			ExprT
#define TyRun			EvdirSqlTable
#define TyRunTab		RunPairT

/*------------------------------------------------------*/
/*	Storage macros					*/
/*------------------------------------------------------*/

#define ADD_RECORD(tab) TableAdd(tab, NULL) 
#define GET_RECORD(type,tab,i) ((type *)TableGet(tab,(i)))
#define INIT_TABLE(tab) TableInit(tab)
#define NUSED(tab) TableN(tab)

#define ADD_NAME(name) TableAdd(TbName,name) 
#define GET_NAME(tab,i) TableName(tab,(i)) 

#endif
