/*
*************************************************************************
**
** Description: class EvdirSqlCompiler:
**	        Lexic Analyser
**
** Language:	C++
**
** Created: 20 Aug 97
** Updated:  
**
** Author:  Sergey Moukhine
**          HERA-B Collaboration
**          JINR-Dubna
**          Russia
**
*************************************************************************
*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>

#include <evdir/EvdTable.h>
#include <evdir/EvdCommon.h>
#include <evdir/EvdCompiler.h>
#include "lang.tab.h"

static char *NameOfToken(int token);

#define MAX_EXP_VALUE  15

/*
**	LexCheckCharSet
**	---------------
*/

void EvdirSqlCompiler::LexCheckCharSet()

{
char *s;
char *s1;

for (s=lex.text; *s; s++)
   {
   if ((unsigned char)*s >= 33 && (unsigned char)*s <= 126)
      continue;

   if (isspace(*s))
      continue;
      
   if ((unsigned char)*s >= 128 && (unsigned char)*s <= 255)
      continue;

   /* if error */
   
   lex.last = s - lex.text;
   for (s1=lex.text; s1 < s; s1++)
      if (isspace(*s1))
         *s1 = ' ';
   
   ErrorI(4, *s);
   }
}

/*
**	LexInit
**	-------
*/

void EvdirSqlCompiler::LexInit(LexStartKind kind)

{
lex.len   = 0;
lex.first = 0;
lex.last  = 0;
lex.init  = 0;
lex.kind = kind;
switch(kind)
   {
   case LxCompile:
      lex.token = COMPILE;
      break;
   case LxFillRunTable:
      lex.token = FILL_RUN_TABLE;
      break;
   }
} 

/*
** FUNCTION
**	LexRead (read new line)
**	-------
**
** PARAMETERS
**	buf - output buffer
**	size - buffer size
**
** RETURNS
**	length of string read ( -1 if EOF)
**
*/

int EvdirSqlCompiler::LexRead(

char *buf,	/* buffer to read to	*/
int  size	/* buffer size		*/

)

{
char *end;
int  n;

if (!(*lex.curSymb))
   return(-1);

end = strchr(lex.curSymb, '\n');
if (!end)
   {
   n = MIN(size, (int)strlen(lex.curSymb));
   strncpy(buf, lex.curSymb, n);
   buf[n] = '\0';
   lex.curSymb += n;
   return(n);
   }

n = end - lex.curSymb;
if (n <= size)
   {
   strncpy(buf, lex.curSymb, n);
   buf[n] = '\0';
   lex.curSymb += n+1;
   return(n);
   }
else
   {
   n = MIN(size, n);
   strncpy(buf, lex.curSymb, n);
   buf[n] = '\0';
   lex.curSymb += n;
   return(n);
   }
}

/*
**	LexSkipBlanks
**	-------------
**/

void EvdirSqlCompiler::LexSkipBlanks()

{
char c;

while (1)			/*  Skip blank lines	*/
   {
   if (lex.first >= lex.len)	/*  End of line	*/
      {
      lex.len = LexRead(lex.text, LEX_TEXT_SIZE);
      lex.first = 0;
      lex.last  = 0;
      if (lex.len < 0)
         return;

      LexCheckCharSet();
      }
   while (lex.first < lex.len)	/*  at one line		*/
      {
      c = lex.text[lex.first];
      if (isspace(c))
         {
         lex.text[lex.first] = ' ';
         lex.first++;
         continue;
         }
      lex.last = lex.first;	/*  not blank found	*/
      return;
      }
   }
}

/*
**	LexNext
**	-------
**
** RETURNS
**	next token
*/


int EvdirSqlCompiler::LexNext(int *yylval)

{
char c;
char *s;
int  n;
double d, exp;
int  sign = 0;
TranRecordsT *rec = &CompilerData.record;

if (lex.init == 0)
   {
   lex.init = 1;
   return(lex.token);
   }
   
if (lex.len < 0)
   {
   lex.token = TOKEN_EOF;
   return(lex.token);
   }

LexSkipBlanks();
if (lex.len < 0)
   {
   lex.token = TOKEN_EOF;
   return(lex.token);
   }

lex.last = lex.first;
c = lex.text[lex.first];

/*	Check  INT & DOUBLE	*/

if (isdigit(c))
   {
   lex.token = INT;
   lex.value = c -'0';
   for (lex.last++; lex.last < lex.len; lex.last++)
      {
      c = lex.text[lex.last];
      if (isdigit(c))
         {
         if (lex.last - lex.first +1 > LEX_NUMBER_LEN)
            Error(3);
         lex.value = (lex.value)*10 + (c - '0');
         }
      else
         break;	/*  end of number  */
      }
   lex.first = lex.last;
   if (c != '.')
      {
      rec->number.kind = NuInt;
      rec->number.value.i = lex.value;
      *yylval = ADD_RECORD(TbNumber);
      return(lex.token);
      }

   lex.token = DOUBLE;
   d = lex.value;
   exp = 0.1;
   for (lex.last++; lex.last < lex.len; lex.last++)
      {
      c = lex.text[lex.last];
      if (isdigit(c))
         {
         if (lex.last - lex.first +1 > LEX_NUMBER_LEN)
            Error(3);
         d += (c - '0')*exp;
         exp *= 0.1;
         }
      else
         break;	/*  end of mantissa  */
      }
   lex.first = lex.last;
   if (!(c == 'E' || c == 'e'))
      {
      rec->number.kind = NuDouble;
      rec->number.value.d = d;
      *yylval = ADD_RECORD(TbNumber);
      return(lex.token);
      }

   /* Exponent */
   lex.last++;
   c = lex.text[lex.last];
   if (c == ' ' || c == '+' || isdigit(c))
      sign = 1;
   else
   if (c == '-')
      sign = -1;
   else
      Error(10);
   
   if (!isdigit(c))
      lex.last++;
   c = lex.text[lex.last];
   if (!isdigit(c))
      Error(10);
   lex.value = (c - '0');
   for (lex.last++; lex.last < lex.len; lex.last++)
      {
      c = lex.text[lex.last];
      if (isdigit(c))
         {
         if (lex.last - lex.first +1 > 2) /* max exponent length = 2 */
            Error(10);
         lex.value = (lex.value)*10 + (c - '0');
         }
      else
         break;	/*  end of number  */
      }
   lex.first = lex.last;
   if (lex.value > MAX_EXP_VALUE)
      Error(10);
   d *= pow(10.,(double)(sign*lex.value));
   rec->number.kind = NuDouble;
   rec->number.value.d = d;
   *yylval = ADD_RECORD(TbNumber);
   return(lex.token);
   }

/*	Check NAME and KEYWORDs		*/

if ( isalpha(c))
   {
   lex.token = NAME;
   for (lex.last++; lex.last < lex.len; lex.last++)
      {
      c = lex.text[lex.last];
      if ( !(isdigit(c) || isalpha(c) || c == '_'  ))
         break;
      }
   n = lex.last - lex.first;
   if (n > LEX_NAME_LEN)
      Error(6);
   memmove(lex.name, lex.text + lex.first, n);
   lex.name[n] = '\0';
   lex.first = lex.last;

   /*	Check KEYWORDs		*/

   for(s=lex.name; *s; s++)
      *s = toupper(*s);
      
   if (strcmp(lex.name, "JAN") == 0)
      {
      *yylval = 1;
      lex.token = MONTH;
      }
   else
   if (strcmp(lex.name, "FEB") == 0)
      {
      *yylval = 2;
      lex.token = MONTH;
      }
   else
   if (strcmp(lex.name, "MAR") == 0)
      {
      *yylval = 3;
      lex.token = MONTH;
      }
   else
   if (strcmp(lex.name, "APR") == 0)
      {
      *yylval = 4;
      lex.token = MONTH;
      }
   else
   if (strcmp(lex.name, "MAY") == 0)
      {
      *yylval = 5;
      lex.token = MONTH;
      }
   else
   if (strcmp(lex.name, "JUN") == 0)
      {
      *yylval = 6;
      lex.token = MONTH;
      }
   else
   if (strcmp(lex.name, "JUL") == 0)
      {
      *yylval = 7;
      lex.token = MONTH;
      }
   else
   if (strcmp(lex.name, "AUG") == 0)
      {
      *yylval = 8;
      lex.token = MONTH;
      }
   else
   if (strcmp(lex.name, "SEP") == 0)
      {
      *yylval = 9;
      lex.token = MONTH;
      }
   else
   if (strcmp(lex.name, "OCT") == 0)
      {
      *yylval = 10;
      lex.token = MONTH;
      }
   else
   if (strcmp(lex.name, "NOV") == 0)
      {
      *yylval = 11;
      lex.token = MONTH;
      }
   else
   if (strcmp(lex.name, "DEC") == 0)
      {
      *yylval = 12;
      lex.token = MONTH;
      }
   else
   if (strcmp(lex.name, "AND") == 0)
      lex.token = AND;
   else
   if (strcmp(lex.name, "BETWEEN") == 0)
      lex.token = BETWEEN;
   else
   if (strcmp(lex.name, "NOT") == 0)
      lex.token = NOT;
   else
   if (strcmp(lex.name, "IN") == 0)
      lex.token = IN;
   else
   if (strcmp(lex.name, "OR") == 0)
      lex.token = OR;
   else
   if (strcmp(lex.name, "RUN") == 0 && lex.kind == LxFillRunTable)
      lex.token = RUN;

   if (lex.token == NAME)
      *yylval = ADD_NAME(lex.name);

   return(lex.token);
   }

/*	Check Simbols	*/

lex.token = c;
switch (c)
   {
   case '=':
      lex.first++;
      lex.token = EQ;
      return(lex.token);

   case '!':
      if (lex.text[lex.first+1] == '=')
         {
         lex.token = NE;
         lex.first += 2;
         return(lex.token);
         }
      return(lex.token);

   case '>':
      if (lex.text[lex.first+1] == '=')
         {
         lex.token = GE;
         lex.first++;
         }
      else
         lex.token = GT;
      lex.first++;
      return(lex.token);

   case '<':
      if (lex.text[lex.first+1] == '=')
         {
         lex.token = LE;
         lex.first++;
         }
      else
         lex.token = LT;
      lex.first++;
      return(lex.token);

   default:
      lex.first++;
      return(lex.token);
   }
// return(0); // never reached; used to avoid compiler warnings
}

static char *NameOfToken(int token)
{
static char name[20];
switch (token)
   {
   case TOKEN_EOF:	strcpy(name,"TOKEN_EOF"); break;
   case EQ:		strcpy(name,"EQ"); break;
   case NE:		strcpy(name,"NE"); break;
   case LT:		strcpy(name,"LT"); break;
   case LE:		strcpy(name,"LE"); break;
   case GT:		strcpy(name,"GT"); break;
   case GE:		strcpy(name,"GE"); break;
   case NAME:		strcpy(name,"NAME"); break;
   case	INT:		strcpy(name,"INT"); break;
   case MONTH:		strcpy(name,"MONTH"); break;
   case AND:		strcpy(name,"AND"); break;
   case OR:		strcpy(name,"OR"); break;
   case NOT:		strcpy(name,"NOT"); break;
   case IN:		strcpy(name,"IN"); break;
   case RUN:		strcpy(name,"RUN"); break;
   case BETWEEN:	strcpy(name,"BETWEEN"); break;
   case COMPILE:	strcpy(name,"COMPILE"); break;
   case FILL_RUN_TABLE:	strcpy(name,"FILL_RUN_TABLE"); break;
   default:
   if (token >255)
      strcpy(name,"Unknown");
   else
      {   		
      name[0]=token; name[1]='\0';
      }
   break;
   }
return(name);
}


/*
** FUNCTION
**	yylex (used by Yacc)
**	--------------------
** RETURNS
**	token
*/

int EvdirSqlCompiler::yylex(int *yylval)

{
int token;

token = LexNext(yylval);

// /*DEBUG*/ printf("TOKEN: %s\n", NameOfToken(token));

return(token);
}


/*
** FUNCTION
**	yyerror (used by Yacc)
**	--------------------
*/

void EvdirSqlCompiler::yyerror(const char *err)

{
Error(5);
// Next statement is never reached; used to avoid compiler warnings
if (0) printf("%s %s\n",err,NameOfToken(0));
}
