#include "gpack/gpack.h"
#include "gpack/gpackproto.h"
#include "gpack/datastream.hh"
#include "gpack/evdstream.hh"

static void*      streams[100];
static StreamType streamTypes[100];

int gp_setstream(StreamType type)
{
  /* ============= */
  /* C interface   */
  /* ============= */
  int           stream;
  void          *p;

  for(stream = 0; stream < 100 && streams[stream]; stream++ ); 
  if( stream == 100 ) return GP_BUFFER_OVERFLOW; 
  switch( type )
    {
    case FileStream:
      p = (void*)new File_Stream;
      break;

    case ShmStream:
      p = (void*)new Shared_Memory;
      break;

    case PublicStream:
      p = (void*)new PublicED;
      break;

    case PrivateStream:
      p = (void*)new PrivateED;
      break;

    default:
      return GP_INVALID_DATA;
    }

  if( !p ) return GP_OUT_OF_MEMORY;
  streams[stream]     = p;
  streamTypes[stream] = type;
  return stream;
}

int gp_settarget(int stream, TargetType target)
{
  /* ============= */
  /* C interface   */
  /* ============= */
  Shared_Memory *s;
  File_Stream   *f;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;

  switch( streamTypes[stream] )
    {
    case ShmStream:
      s = (Shared_Memory*)streams[stream];
      return s->settarget(target);

    case FileStream:
      f = (File_Stream*)streams[stream];
      return f->settarget(target);

    case PublicStream:
    case PrivateStream:
      return GP_NOT_IMPLEMENTED;
      
    default:
      break;
    }
  return GP_INVALID_DATA;
}

int gp_setname(int stream, char* filename, int openmode)
{
  /* ============= */
  /* C interface   */
  /* ============= */
  Shared_Memory *s;
  File_Stream *f;
  PrivateED   *v;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;
  
  switch( streamTypes[stream] )
    {
    case FileStream:
      f = (File_Stream*)streams[stream];
      return f->setname(filename, openmode);

    case PrivateStream:
      v = (PrivateED*)streams[stream];
      return v->setname(filename, openmode);

    case ShmStream:
      s = (Shared_Memory*)streams[stream];
      return s->setname(filename, openmode);

    default:
      break;
    }
  return GP_NOT_IMPLEMENTED;
}

int gp_setkey(int stream, int key)
{
  /* ============= */
  /* C interface   */
  /* ============= */
  Shared_Memory *s;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;
  if( streamTypes[stream] != ShmStream )  return GP_NOT_IMPLEMENTED;
  s = (Shared_Memory*)streams[stream];
  s->setkey((key_t) key);
  return GP_SUCCESS; 
}

int gp_setbuf(int stream, int size)
{
  /* ============= */
  /* C interface   */
  /* ============= */
  Shared_Memory *s;
  File_Stream   *f;
  PublicED      *b;
  PrivateED     *v;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;

  switch( streamTypes[stream] )
    {
    case ShmStream:
      s = (Shared_Memory*)streams[stream];
      s->setbuf(size);
      return GP_SUCCESS;

    case FileStream:
      f = (File_Stream*)streams[stream];
      f->setbuf(size);
      return GP_SUCCESS;
      
    case PublicStream:
      b = (PublicED*)streams[stream];
      b->setbuf(size);
      return GP_SUCCESS;
      
    case PrivateStream:
      v = (PrivateED*)streams[stream];
      v->setbuf(size);
      return GP_SUCCESS;
      
    default:
      break;
    }
  return GP_INVALID_DATA;
}

int gp_create_ring(int stream)
{
  /* ============= */
  /* C interface   */
  /* ============= */
  Shared_Memory *s;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;
  if( streamTypes[stream] != ShmStream )  return GP_NOT_IMPLEMENTED;
  s = (Shared_Memory*)streams[stream];

  return s->create();
}

int gp_destroy_ring(int stream)
{
  /* ============= */
  /* C interface   */
  /* ============= */
  Shared_Memory *s;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;
  if( streamTypes[stream] != ShmStream )  return GP_NOT_IMPLEMENTED;
  s = (Shared_Memory*)streams[stream];

  return s->destroy();
}

int gp_open(int stream)
{
  /* ============= */
  /* C interface   */
  /* ============= */
  Shared_Memory *s;
  File_Stream   *f;
  PublicED      *b;
  PrivateED     *v;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;

  switch( streamTypes[stream] )
    {
    case ShmStream:
      s = (Shared_Memory*)streams[stream];
      return s->open();

    case FileStream:
      f = (File_Stream*)streams[stream];
      return f->open();

    case PublicStream:  
      b = (PublicED*)streams[stream];
      return b->open();

    case PrivateStream:
      v = (PrivateED*)streams[stream];
      return v->open();

    default:
      break;
    }
  return GP_INVALID_DATA;
}

int gp_close(int stream)
{
  /* ============= */
  /* C interface   */
  /* ============= */
  Shared_Memory *s;
  File_Stream   *f;
  PublicED      *b;
  PrivateED     *v;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;

  switch( streamTypes[stream] )
    {
    case ShmStream:
      s = (Shared_Memory*)streams[stream];
      return s->close();

    case FileStream:
      f = (File_Stream*)streams[stream];
      return f->close();
      
    case PublicStream:  
      b = (PublicED*)streams[stream];
      return b->close();

    case PrivateStream:
      v = (PrivateED*)streams[stream];
      return v->close();

    default:
      break;
    }
  return GP_INVALID_DATA;
}

int gp_getevt(int  stream, EventHdr* h)
{
  /* ============= */
  /* C interface   */
  /* ============= */
  Shared_Memory *s;
  File_Stream   *f;
  PublicED      *b;
  PrivateED     *v;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;

  switch( streamTypes[stream] )
    {
    case ShmStream:
      s = (Shared_Memory*)streams[stream];
      return s->getevt(*h);

    case FileStream:
      f = (File_Stream*)streams[stream];
      return f->getevt(*h);
      
    case PublicStream:  
      b = (PublicED*)streams[stream];
      return b->getevt(*h);

    case PrivateStream:
      v = (PrivateED*)streams[stream];
      return v->getevt(*h);

    default:
      break;
    }
  return GP_INVALID_DATA;
}

int gp_putevt(int  stream, EventHdr* h)
{
  /* ============= */
  /* C interface   */
  /* ============= */
  Shared_Memory *s;
  File_Stream   *f;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;

  switch( streamTypes[stream] )
    {
    case ShmStream:
      s = (Shared_Memory*)streams[stream];
      return s->putevt(*h);

    case FileStream:
      f = (File_Stream*)streams[stream];
      return f->putevt(*h);
      
    case PublicStream:
    case PrivateStream:
      return GP_NOT_IMPLEMENTED;

    default:
      break;
    }
  return GP_INVALID_DATA;
}

int gp_gettabinfo(int   stream, char* name, int* ncols, int* nrows)
{
  /* =========== */
  /* C interface */
  /* =========== */
  Shared_Memory *s;
  File_Stream   *f;
  PublicED      *b;
  PrivateED     *v;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;

  switch( streamTypes[stream] )
    {
    case ShmStream:
      s = (Shared_Memory*)streams[stream];
      return s->gettabinfo(name, *ncols, *nrows);

    case FileStream:
      f = (File_Stream*)streams[stream];
      return f->gettabinfo(name, *ncols, *nrows);
      
    case PublicStream:  
      b = (PublicED*)streams[stream];
      return b->gettabinfo(name, *ncols, *nrows);

    case PrivateStream:
      v = (PrivateED*)streams[stream];
      return v->gettabinfo(name, *ncols, *nrows);

    default:
      break;
    }
  return GP_INVALID_DATA;
}

int gp_skiptab(int   stream)
{
  /* =========== */
  /* C interface */
  /* =========== */
  Shared_Memory *s;
  File_Stream   *f;
  PublicED      *b;
  PrivateED     *v;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;

  switch( streamTypes[stream] )
    {
    case ShmStream:
      s = (Shared_Memory*)streams[stream];
      return s->skiptab();

    case FileStream:
      f = (File_Stream*)streams[stream];
      return f->skiptab();
      
    case PublicStream:  
      b = (PublicED*)streams[stream];
      return b->skiptab();

    case PrivateStream:
      v = (PrivateED*)streams[stream];
      return v->skiptab();

    default:
      break;
    }
  return GP_INVALID_DATA;
}

int gp_gettab(int   stream,
	      char* name,
	      int   ncols,
	      int   nrows,
	      char* description,
	      void* data)
{
  /* =========== */
  /* C interface */
  /* =========== */
  Shared_Memory *s;
  File_Stream   *f;
  PublicED      *b;
  PrivateED     *v;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;

  switch( streamTypes[stream] )
    {
    case ShmStream:
      s = (Shared_Memory*)streams[stream];
      return s->gettab(name, ncols, nrows, description, data);

    case FileStream:
      f = (File_Stream*)streams[stream];
      return f->gettab(name, ncols, nrows, description, data);
      
    case PublicStream:  
      b = (PublicED*)streams[stream];
      return b->gettab(name, ncols, nrows, description, data);

    case PrivateStream:
      v = (PrivateED*)streams[stream];
      return v->gettab(name, ncols, nrows, description, data);

    default:
      break;
    }
  return GP_INVALID_DATA;
}

int gp_puttab(int   stream,
	      char* name,
	      int   ncols,
	      int   nrows,
	      char* description,
	      void* data)
{
  /* =========== */
  /* C interface */
  /* =========== */
  Shared_Memory *s;
  File_Stream   *f;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;

  switch( streamTypes[stream] )
    {
    case ShmStream:
      s = (Shared_Memory*)streams[stream];
      return s->puttab(name, ncols, nrows, description, data);

    case FileStream:
      f = (File_Stream*)streams[stream];
      return f->puttab(name, ncols, nrows, description, data);
      
    case PublicStream:
    case PrivateStream:
      return GP_NOT_IMPLEMENTED;

    default:
      break;
    }
  return GP_INVALID_DATA;
}

int gp_flush(int   stream)
{
  /* =========== */
  /* C interface */
  /* =========== */
  Shared_Memory *s;
  File_Stream   *f;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;

  switch( streamTypes[stream] )
    {
    case ShmStream:
      s = (Shared_Memory*)streams[stream];
      return s->flushevt();

    case FileStream:
      f = (File_Stream*)streams[stream];
      return f->flushevt();
      
    case PublicStream:
    case PrivateStream:
      return GP_NOT_IMPLEMENTED;

    default:
      break;
    }
  return GP_INVALID_DATA;
}

int gp_setevtype(int stream, EventType type)
{
  /* =========== */
  /* C interface */
  /* =========== */
  Shared_Memory *s;
  File_Stream   *f;
  PublicED      *b;
  PrivateED     *v;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;

  switch( streamTypes[stream] )
    {
    case ShmStream:
      s = (Shared_Memory*)streams[stream];
      s->setevtype(type);
      return GP_SUCCESS;

    case FileStream:
      f = (File_Stream*)streams[stream];
      f->setevtype(type);
      return GP_SUCCESS;
      
    case PublicStream:
      b = (PublicED*)streams[stream];
      b->SetEventType(type);
      return GP_SUCCESS;
      
    case PrivateStream:
      v = (PrivateED*)streams[stream];
      v->SetEventType(type);
      return GP_SUCCESS;
      
    default:
      break;
    }
  return GP_INVALID_DATA;
}

int gp_getevtype(int stream, EventType* type)
{
  /* =========== */
  /* C interface */
  /* =========== */
  Shared_Memory *s;
  File_Stream   *f;
  PublicED      *b;
  PrivateED     *v;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;

  switch( streamTypes[stream] )
    {
    case ShmStream:
      s = (Shared_Memory*)streams[stream];
      *type = s->getevtype();
      return GP_SUCCESS;

    case FileStream:
      f = (File_Stream*)streams[stream];
      *type = f->getevtype();
      return GP_SUCCESS;
      
    case PublicStream:
      b = (PublicED*)streams[stream];
      *type = b->GetEventType();
      return GP_SUCCESS;
      
    case PrivateStream:
      v = (PrivateED*)streams[stream];
      *type = v->GetEventType();
      return GP_SUCCESS;
      
    default:
      break;
    }
  return GP_INVALID_DATA;
}

int gp_getevtpos(int stream)
{
  /* =========== */
  /* C interface */
  /* =========== */
  Shared_Memory *s;
  File_Stream   *f;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;

  switch( streamTypes[stream] )
    {
    case ShmStream:
      s = (Shared_Memory*)streams[stream];
      return s->getevtpos();

    case FileStream:
      f = (File_Stream*)streams[stream];
      return f->getevtpos();			/* in bytes */
      
    default:
      break;
    }
  return GP_INVALID_DATA;
}

char* gp_errmsg(int stream)
{
  /* ============= */
  /* C interface   */
  /* ============= */
  Shared_Memory *s;
  File_Stream   *f;
  PublicED      *b;
  PrivateED     *v;
  char *defaultMsg = "Stream has invalid stream type!";
  char *outOfRange = "Stream number is aout of range (0 : 99)";
  char *notInitial = "Stream is not initialized";

  if( stream < 0 && stream >= 100 )       return outOfRange;
  if( !streams[stream] )                  return notInitial;

  switch( streamTypes[stream] )
    {
    case ShmStream:
      s = (Shared_Memory*)streams[stream];
      return s->getErrorMsg();

    case FileStream:
      f = (File_Stream*)streams[stream];
      return f->getErrorMsg();
      
    case PublicStream:
      b = (PublicED*)streams[stream];
      return b->getErrorMsg();
      
    case PrivateStream:
      v = (PrivateED*)streams[stream];
      return v->getErrorMsg();
      
    default:
      break;
    }
  return defaultMsg;
}

int gp_select(int stream, char* criteria)
{
  /* ============= */
  /* C interface   */
  /* ============= */
  PublicED      *b;
  PrivateED     *v;

  if( stream < 0 && stream >= 100 )       return GP_INVALID_DATA;
  if( !streams[stream] )                  return GP_INVALID_DATA;

  switch( streamTypes[stream] )
    {
    case ShmStream:
    case FileStream:
      return GP_NOT_IMPLEMENTED;
      
    case PublicStream:
      b = (PublicED*)streams[stream];
      return b->select( criteria );
      
    case PrivateStream:
      v = (PrivateED*)streams[stream];
      return v->select( criteria );
      
    default:
      break;
    }
  return GP_INVALID_DATA;
}

