#include "gpack/filestream.hh"
#include "gpack/datastream.hh"
#include "evdir/EvdSql.h"

//-----------------------------------------------------------------------
// Event Directory (EVD) is a file or a group of files which consist of 
// entries describing:
// 
//  o event ID     -- info identifying event: run number & event number
//  o event address -- info defining which data file the event belogns to
//                     and what is the position of the event inside that file
//  o event properties -- info defining event characteristics (for selection
//                        purposes)
//
// The events are kept in 2 types of files:  DST -- "full events" 
// and MINI -- "compact" info about event (without raw data).  Therefore,
// the EVD must support the both types.
//
// EVD can be "public" or "private".  For "private" EVD we keep full path
// name of data file.  
// Public EVD is based on the fact that it is used well defined naming 
// convention for data files.   For instance, for DST files:
//
// <Distribution Root>/<year>/exp<exp>/dst/run<run>/run<run>_<file>.dst
// 
// Thus the data file name may be encoded and we can keep only:
//    year    -  year of data taking
//    exp     - experiment number
//    run     - run number
//    file    - file number
// in order to restore full path name of a data file ( Distribution Root
// is a permanent part of the filename).  Note that because of Root is not
// kept inside of event directory,  EVD does not depend on the mount point
// of the "event tree".  So the layout of EVD element can look like:
//
//    year, exp, run, file, event, dstAddr, miniAddr, datime, classmask
//
//  Actually, the first 4 numbers are permanent for thousands of events,
//  therefore, we should divide  EVD element in two parts:
//      -- encoded path

typedef struct EPath
{
  int year;
  int exp;
  int run;
  int file;
} EPath;

//      -- event tag 

typedef struct Tag
{
  int event;
  int dst;
  int mini;
  int datime;
  int classmask;
} Tag;

// Event number cannot be equal 0, so zero marks the end of group of tags.
// The layout of EVD file we use is the following:
// 0
// "marked" year, 
//     "marked" exp, 
//          "marked" run, 
//              "marked" file, 
//                  tag, tag, tag, ..., tag, 0,
//              "marked" file, 
//                  tag, ..., tag, 0, 
//              ....,
//          "marked" run, 
//              "marked" file, 
//                  tag, ..., tag, 0,
//       ....
// The marker serves for recognition of which element of encoded path 
// is changed
//------------------------------------------------------------------------

enum EvdMarker 
{
  FileMask = 0x04000000,
  RunMask  = 0x08000000,
  ExpMask  = 0x10000000,
  YearMask = 0x20000000,
  NameMask = 0x40000000,
  IsMarker = 0x80000000,
  GetMark  = 0x7c000000,
  GetValue = 0x03ffffff
};


#define NPARAMS 6
typedef enum Params
{
  TClass = 0,
  TDate,
  TTime,
  TExp,
  TRun,
  TEvent
} Params;

class evdstream: public filestream
{
private:
  int            openmode;
  EventType      eventType;
  char           dataname[MAXPATHNAME];
  int            hYear;
  int            hExp;
  int            hRun;
  int            hFile;
  int            flags;
  int            event;
  int            dst;
  int            mini;
  int            datime;
  int            classmask;

protected:

  EvdirSql*      sql;
  NameAndValueT* parm;
  File_Stream    *dataStream;
  char           errmsg[256];

public:
            evdstream();
           ~evdstream();

  char*     getErrorMsg();
  void      setMode(FileModes m);
  int       getMode();
  void      SetEventType(EventType t);
  EventType GetEventType();
  int       setYER(int year, int exp, int run); 
  void      getYER(int& year, int& exp, int& run);
  int       setFileNo(int file);
  int       getFileNo();
  int       setDataFname(char* filename);
  int       putTag(int evt, int dt, int clmsk, int dst, int mini);
  int       getTag(int& evt, int& dt, int& clmsk, int& dst, int& mini);
  int       getevt(EventHdr& h);
  int       gettabinfo(char *name, int& ncols, int& nrows);
  int       skiptab();
  int       gettab(char* name, int ncols, int nrows, char* desc, void* data);
  int       filter(EventProp p);
  int       select(char* criteria);
};

class PublicED: public evdstream
{
private:

  char* root;             // contains HB_ROOT_PATH env. var.

  // Run Catalog related variables (for ReadOnly mode):
  int   rcfd;             // Run Catalog File Descriptor
  char* buffer;           // Internal buffer (cache) of Run Catalog
  int   bufsize;
  int   used;
  int   cursor;           // position in internal buffer
 
  // Event Directory:
  int   findYER(int& year, int& exp, int& run);
  int   nextYER(int& year, int& exp, int& run);

public:
        PublicED();
       ~PublicED();

  int   open();
  int   close();
  int   getTag(int& e, int& s, int& c, int& d, int& m);
  int   getevt(EventHdr& h);
};

typedef evdstream PrivateED;

#include "gpack/evdstream.icc"
