#ifndef _RCMATCHCONST_H_
#define _RCMATCHCONST_H_
//------------------------------------------------------------------------
// Created: 23.08.97 by   Olya Igonkina     for MARPLE

//------------------------------------------------------------------------
#include <math.h>

#include "marple/NameRcMatchConst.h"

#include "marple/RcPoint.hh"
#include "marple/RcLimit.h"

#include <CLHEP/Utilities/fortran.h>
extern "C" 
{ 
  extern void FORTRAN_ROUTINE( kugetr   )(float *); 
  extern void FORTRAN_ROUTINE( kugeti   )(int *  ); 
}

/** {\bf MARPLE structure of constants for matching prosedure    }  \\

  The parameters used by matching are defined by the {\bf flag} : \\

 \begin{tabular}{lccccc}
   flag = & 1  & 1  & 1  & 1  & 1  \\
          & |  & |  & |  & |  & pe \\
          & |  & |  & |  & ty &    \\
          & |  & |  & tx &    &    \\
          & |  & y  &    &    &    \\
          & x  &    &    &    &    \\
 \end{tabular}

 set 1 at the corresponding position then       
 parameter will be taken into account       \\         
 otherwise set 0,                           \\       
 usual value of the flag is 11110 which means   
 that only { x,y,tx,ty } are taken into account  

 {\bf BE SURE that flag is defined !!! }

 {\bf possible cuts for matching : }                                       \\
 1. set {\bf $ \chi^2 $ } to restrict the $\chi^2 $ value of matching      \\
 2. use delta cut to restrict the distance between two objects to be match \\
 3. use pair cut to restrict the number of combination with given objects to be match

 @memo MARPLE structure of constants for matching prosedure 
*/

struct RcMatchConst
{
  ///           flag defined the parameters used by match procedure ( integer value)
  int flag;
  ///           z coordinate of the plane of matching
  HepDouble z_plane;
  
  ///           smallest Z coordination of point to be taken into account 
  HepDouble z_min;
  ///           bigest   Z coordination of point to be taken into account 
  HepDouble z_max;
  
  ///           = 0 if pair cut is not used during match
  int use_pair_cut;
  ///the number of combinations with left objects to be match if pair cut is used 
  int n_pair_left;
  ///the number of combinations with right objects to be match if pair cut is used 
  int n_pair_right;

  ///              = 0 if delta cut is not used during match
  int use_delta_cut;
  ///                 $ | X_left - X_right | < \delta_x $ if delta cut is used
  HepDouble delta_x;
  ///                 $ | Y_left - Y_right | < \delta_y $ if delta cut is used
  HepDouble delta_y;
  ///                 $ | TX_left - TX_right | < \delta_tx $ if delta cut is used
  HepDouble delta_tx;
  ///                 $ | TY_left - TY_right | < \delta_ty $ if delta cut is used
  HepDouble delta_ty;
  ///            $ | PE_left - PE_right | < \delta_pe $ pe = Q/p if delta cut is used
  HepDouble delta_pe;
  ///                 the upper limit on $ \chi^2 $ of matching
  HepDouble chi2;

  // functions:
  //    iflag       should have the form described above           
  // if iflag  = -1 then the member flag is used for parameters check 

  ///              check that zposition lies inside the  [ z\_min, z\_max]
  inline int geomAccept( const double zposition) const;
  /// check that | rp1.par - rp2.par | <= delta.par where par={x,y,tx,ty,pe}, return 0 if false
  int delta_cut_ok( const RcPoint & rp1,
		    const RcPoint & rp2,
		    int iflag = -1 ) const;
  ///                      check that | rp1.par - rlimit.par | <= delta.par
  int delta_cut_ok( const RcPoint & rp1, 
		    const RcLimit & rlimit,
		    int iflag = -1 ) const;

  ///                    check that RcPoint still could be used by matching
  inline int pair_cut_ok( const RcPoint & rp) const;
  ///                    check that both RcPoints still could be used by matching
  inline int pair_cut_ok( const RcPoint & rp1, 
			  const RcPoint & rp2 ) const;


//#ifdef USEKUIP
  //Read a command from external
  inline void  read( NameRcMatchConst name = Eall);

protected:

  //Read a float
  inline float readFloat();

  //Read an integer
  inline int   readInt();
//#endif

};

ostream & operator << (ostream &, const RcMatchConst &);

#include "marple/RcMatchConst.icc"

#endif






