#ifndef __FED__H_
#define __FED__H_
#include <stdlib.h>
#include <stdio.h>
#include <string.h>


inline          short RICHswap(         short v) { 
  return (((v&0xFF)<<8)|((v&0xFF00)>>8)); 
  }
inline unsigned short RICHswap(unsigned short v) {
  return (((v&0xFF)<<8)|((v&0xFF00)>>8)); 
  }
inline          long  RICHswap(long  v) { 
  return (((v&0x000000FF)<<24)|((v&0xFF000000)>>24)|
          ((v&0x0000FF00)<< 8)|((v&0x00FF0000)>>8)); 
  }
inline unsigned long  RICHswap(unsigned long  v) { 
  return (((v&0x000000FF)<<24)|((v&0xFF000000)>>24)|
          ((v&0x0000FF00)<< 8)|((v&0x00FF0000)>>8)); 
  }

inline void RICHswap(unsigned short *v, int n) {
  int i;
  for (i=0;i<n/sizeof(unsigned short);i++) v[i]=RICHswap(v[i]);
  }

// ---------------------------------------------------------------------
/// information about the FED set up
/**
  The class ChannelInfo is used to defined the FED setup.
  It stores the general information about the number of
  daughters fro each mother, and the number of connectors (ASD8's)
  for each daughter. It provide function to find the data
  from one ASD as function of the daughter-id and connector
  number (being zero for the upper connector) in data as well
  as the index in the bit-stream as function of the daughter-ID,
  connector-OD and ASD8-channel number. Invers functions are also
  supported. This information is assumed to be the same for all used
  FED-systems. But which channels don't work or which connector
  have no ASD plugged in, may be different for different FED system.
  Therefore the function ASDWorks and ChannelWorks are virtual function,
  which have to be supplied for the special cases.
 */

class ChannelInfo {
public:
///
  inline int NumConnector() const { return 64; }
/// return the number of data channels
  inline int NumChannel() const { return NumConnector()*16; }
/// returns index if the FED-data is interpreted as short's
  inline int index(int daughter,int connector) const { 
    return (daughter*8+connector);
    }
/// returns index in the bit-stream
  inline int index(int daughter,int connector,int channel) const { 
    return index(daughter,connector)*16+((channel&0x8)|(7-channel&0x7));
    }
/// returns Daughter-ID as function of the bit number
  inline int Daughter (int bit) const { return (bit/16)/8; }
/// returns Connector-ID as function of the bit number
  inline int Connector(int bit) const { return (bit/16)%8; }
/// returns ASD-Channel as function of the bit number
  inline int Channel  (int bit) const {
    int i=bit%16;
    return (i&0x8)|(7-(i&0x7)); 
    }
/// return 1 if complete ASD works/is connected at all else 0     
  virtual int ASDWorks(int i) const     { return 0; }
/// return 1 if a certain channel should be used else 0.
  virtual int ChannelWorks(int i) const { return 0; }
  };
  
// ---------------------------------------------------------------------
/// handling of the binary data from the FED
/**
 The class FEDBuffer provides a buffer for the data coming
 from the FED-system. It is a virtual class to allow the use
 independent of the input device: This means accessing or viewing
 of the data is same, independent if it is read from the SHARC
 or from a file. Therefore there exist two derived classes, the
 OnlineFED (reading from SHARC) and the OfflineFED (reading from a file).
 The code is based on OTR-code.
  */
class FEDBuffer {

protected:
// structure of the data coming from the FED-crate
  struct {
    unsigned short int WC;
    unsigned short int R_ID;
    unsigned short int FLT_EV;
    unsigned short int BX;
    unsigned short int ComponentID;
    unsigned short int B_ID;
    unsigned short int data[64];
    unsigned short int Spare;
    unsigned short int ConnectedDaughters;
    } buffer;

unsigned long fltcount;

public:
  inline FEDBuffer() { fltcount=0; }
/// returns the buffer size in 2 byte words
  inline int WordCount() const { 
    return buffer.WC; 
    }
/// returns the RecordID
  inline int RecordID() const { 
    return buffer.R_ID; 
    }
/// returns the first level trigger event number
  inline int FLTEventNo() const { 
    return buffer.FLT_EV; 
    }
/// returns the first level trigger event number
  inline int NumTrigger() const { 
    return fltcount+1; 
    }
/// returns the bunch clock
  inline int BX() const { 
    return buffer.BX; 
    }
///  returns the mother (or component)-id.
  inline int MotherID() const { 
    return buffer.ComponentID; 
    }
///  returns the Board-id.
  inline int BoardID() const { 
    return buffer.B_ID; 
    }
///  returns the spare word.
  inline int Spare() const { 
    return buffer.Spare; 
    }
///  returns the connected daugther word.
  inline int ConnectedDaughters() const { 
    return buffer.ConnectedDaughters; 
    }

  inline int WrongWordCount() {
    return (buffer.WC != 0x0090);
    }
  inline int WrongRecordID() {
    return (buffer.R_ID != 0x0002);
    }
  
  inline int NumDaughter() const  { return 8; }
  inline int NumConnectorPerDaughter() const  { return 8; }
  inline int DaughterIsConnected(int i) {
    return ((buffer.ConnectedDaughters&(1<<i))!=0);
    }

/// return the i'th 16 bit data word.
  inline unsigned short int Data(int i) const { return buffer.data[i]; }

/// binary write to the datafile
  int write(FILE* fp);
/// hex-dump to a formatted file, like stdout
  void dump(FILE *fp=stdout);
/// hex-dump to a vt100 terminal (supplied by S.Korpar)
  void vtdump();
/// check if data is zero
  int ZeroData();
  
  friend int operator!=(const FEDBuffer &x,const FEDBuffer &y);
  friend int operator==(const FEDBuffer &x,const FEDBuffer &y);
  void operator=(const FEDBuffer &x);
  FEDBuffer& operator|=(const FEDBuffer &x);

  };

// ---------------------------------------------------------------------
/// reading FED-data from file

class OfflineFEDBuffer : public FEDBuffer {
  int byteswap;  
public:
/// constructor
   OfflineFEDBuffer();
/// set byteswap
  inline void ByteSwap() { byteswap=1;}
/// binary read from the file
  int read(FILE *fp);
/// move events forwards backward
  int wind(FILE *fp,int numevent,unsigned long datastart);
  };

// ---------------------------------------------------------------------

/// hit search in FED-Buffer
/**
  To find the valid hits the class needs the data of an FEDBuffer
  and the ChannelInfo, to check which ASD's or hits have to be ignored.
  An example can be found in the documentino of the class RICHFED.
 */
class FEDHits {
  FEDBuffer* fed;
  ChannelInfo* c;

// hit search
  int bit,oldbit,bitmask;

public:
/// constructor combining the data with the channel information
  FEDHits(FEDBuffer* f,ChannelInfo& c);
/// shortcut to access the ChannelInfo::ASDWorks
  inline int ASDWorks    (int bit) { return c->ASDWorks(bit);}
/// shortcut to access the ChannelInfo::ChannelInfo
  inline int ChannelWorks(int bit) { return c->ChannelWorks(bit);}
/// shortcut to access the ChannelInfo::NummConnector
  inline int NumConnector()       { return c->NumConnector();}
/// search for hits as long as 1 is returned.
  int  Search();
/// bit position for last hit  
  inline int  Bit() const { return bit; };
  };


// -----------------------------------------------------------------------

class RunHeader {
  int byteswap;
protected:

// run header
  struct {
    unsigned short wc;
    unsigned short rid;
    unsigned long  RunNo;
    unsigned short RunFlag;
    unsigned short year,month,mday,wday,hour,minutes,seconds,millisec,align1;
    unsigned long  duration;
    unsigned long  numtrigger;
    unsigned long  numeventinfile;
    char           filename[1024];
    char           username[1024];
    char           comment[1024];    
    long           spare[7];
    unsigned short morespare[512-7*2]; 
    unsigned short wcend,align2;
    } runheader;
    
public:
/// set byteswap
  inline void ByteSwap() { byteswap=1;}

  inline unsigned short RunNo()       const { return runheader.RunNo;}
  inline unsigned short RunFlag()     const { return runheader.RunFlag;}
  inline unsigned long  RunDuration() const { return runheader.duration;}
  inline unsigned long  NumTrigger()  const { return runheader.numtrigger; }
  inline unsigned long  EventsInFile()const { return runheader.numeventinfile;}
  inline const char*    UserName()    const { return &runheader.username[0];}
  inline const char*    FileName()    const { return &runheader.filename[0];}
  inline const char*    Comment()     const { return &runheader.comment[0];}
  inline unsigned short Year()        const { return runheader.year; }
  inline unsigned short Month()       const { return runheader.month; }
  inline unsigned short Day()         const { return runheader.mday; }
  inline unsigned short WeekDay()     const { return runheader.wday; }
  inline unsigned short Hour()        const { return runheader.hour; }
  inline unsigned short Minutes()     const { return runheader.minutes; }
  inline unsigned short Seconds()     const { return runheader.seconds; }
  inline unsigned short MilliSeconds() const { return runheader.millisec; }

  inline void    Comment(char *s)   { 
    int l=(strlen(s)<1023) ? strlen(s) : 1023;
    memcpy(runheader.comment,s,l+1);
    }
  inline void    FileName(char *s)   { 
    int l=(strlen(s)<1023) ? strlen(s) : 1023;
    memcpy(runheader.filename,s,l+1);
    }
  inline void    UserName(char *s)   { 
    int l=(strlen(s)<1023) ? strlen(s) : 1023;
    memcpy(runheader.username,s,l+1);
    }

// units of voltages: mV  !!!!

  inline unsigned long& RunNo()        { return runheader.RunNo;}
  inline unsigned short& RunFlag()      { return runheader.RunFlag;}
  inline unsigned long&  RunDuration()  { return runheader.duration;}
  inline unsigned long&  EventsInFile() { return runheader.numeventinfile;}
  inline long& HV_M16a()   { return runheader.spare[0]; }
  inline long& HV_M16b()   { return runheader.spare[1]; }
  inline long& HV_M4a()    { return runheader.spare[2]; }
  inline long& HV_M4b()    { return runheader.spare[3]; }
  inline long& LV_plus()   { return runheader.spare[4]; }
  inline long& LV_minus()  { return runheader.spare[5]; }
  inline long& Threshold() { return runheader.spare[6]; }

  inline long HV_M16a()   const { return runheader.spare[0]; }
  inline long HV_M16b()   const { return runheader.spare[1]; }
  inline long HV_M4a()    const { return runheader.spare[2]; }
  inline long HV_M4b()    const { return runheader.spare[3]; }
  inline long LV_plus()    const { return runheader.spare[4]; }
  inline long LV_minus()   const { return runheader.spare[5]; }
  inline long Threshold() const { return runheader.spare[6]; }


  inline RunHeader() { 
    memset(&runheader,0,sizeof(runheader));
    memset(&runheader.comment[0],' ',sizeof(runheader.comment));
    runheader.comment[0]='\0';
    memset(&runheader.filename[0],' ',sizeof(runheader.filename));
    runheader.filename[0]='\0';
    memset(&runheader.username[0],' ',sizeof(runheader.username));
    runheader.username[0]='\0';
  
    byteswap=0;
    runheader.wc=runheader.wcend=sizeof(runheader)/2;
    runheader.rid=0x501;    
    }
  int write(FILE *fp);
  int read(FILE *fp);
  int printheader(FILE *fp=stdout);
  }; 
  
// -----------------------------------------------------------------------

class RunInfo  {
  int byteswap;
protected:
   
// time info during the run
  struct {
    unsigned short wc;
    unsigned short rid;
    unsigned long  numevent;
    unsigned long  EventNo;
    unsigned long  duration;    // in millisec.
    } runinfo;

public:
/// set byteswap
  inline void ByteSwap() { byteswap=1;}

  inline unsigned long  Duration() const { return runinfo.duration; }
  inline unsigned long  EventNo()  const { return runinfo.EventNo; }
  inline unsigned long  NumEvent() const { return runinfo.numevent; }

  inline RunInfo() { 
    memset(&runinfo  ,0,sizeof(runinfo));
    byteswap=0;
    runinfo.wc=sizeof(runinfo)/2;
    runinfo.rid=0x502;    
    }
  int write(FILE *fp);
  int read(FILE *fp);
  }; 
#endif






