#ifndef _PHOTONDETECTOR_H__
#define _PHOTONDETECTOR_H__

#include "rich/ricutil.h"


class RICHLosses {
int loss[10];
public:
   inline void reset() { memset(&loss[0],0,sizeof(loss)); }
   inline RICHLosses() { reset(); }
   
   inline int& OutOfSuperModules()       { return loss[ 0]; }
   inline int& CellEmpty()               { return loss[ 1]; }
   inline int& OutOfPMT()                { return loss[ 2]; }
   inline int& PMTEfficiency()           { return loss[ 3]; }
   inline int& LensAngle()               { return loss[ 4]; }
   inline int& LensWaveLength()          { return loss[ 5]; }
   
   inline int  OutOfSuperModules() const { return loss[ 0]; }
   inline int  CellEmpty()         const { return loss[ 1]; }
   inline int  OutOfPMT()          const { return loss[ 2]; }
   inline int  PMTEfficiency()     const { return loss[ 3]; }
   inline int  LensAngle()         const { return loss[ 4]; }
   inline int  LensWaveLength()    const { return loss[ 5]; }
   
   void print(double numevent=1,FILE *fp=stdout);
   };

extern RICHLosses RICH_Loss;

// ------------------------------------------------------------------------
/// PMT--type specific information
/**
  The class PMType is used for the description of the common 
  properties of all M16 respectivily M4 pmt's. The geometrical
  properties of the channel grid and the common wavelength
  dependency (assumption!) of all M16 or M4 pmt's can be found
  here. 
  The channel numbers are defined in the following way. If you are
  looking from the back of one M16-pmt:\\[4mm]
  \begin{tabular}{lll}
  lowerleft  corner:&index=0 &(row=0,column=0)\\
  lowerright corner:&index=3 &(row=0,column=3)\\
  upperleft  corner:&index=12&(row=3,colunm=0)\\
  upperleft  corner:&index=15&(row=3,colunm=3)
  \end{tabular}\\[4mm]
  Similar in the M4 case. 
 */
class PMType {
protected:
  char *nname;
  ArtePointer<CPMT> a;
  int  numchan,nnx,nny;
  double xxmin,yymin,xxdiv,yydiv;
  Table eff;

  inline PMType() { nname=strdup("undefined");};
  inline void name(char *n) {
    if (nname) free(nname);
    nname=strdup(n);
    }
  
public:
/// return type-name 
  inline const char*  name() const { return nname; }

/// returns the number of channels
  inline int    numchannel() const { return numchan; }

/// returns the number of columns (number of channels in x-direction) 
  inline int    numcol() const { return nnx; }

/// returns the number of rows    (number of channels in y-direction) 
  inline int    numrow() const { return nny; }

/// returns the pad size of a column (padsize in x-direction) 
  inline double    colsize() const { return xxdiv; }

/// returns the pad size of a row (padsize in y-direction) 
  inline double  rowsize() const { return yydiv; }


/// returns pmt-channelindex: mapping of row and column to linear index
  inline int index(int i,int j) const { 
    int k=i*nnx+j;
    return (k>=numchannel()) ? k=-1 : k;
    }
    
/// evaluate the row from the pmt-channelindex
  inline int row(int index) const { 
    return index/nnx;
    }

/// evaluate the column from the pmt-channelindex
  inline int col(int index) const { 
    return index%nnx;
    }
   
/// returns pmt-channelindex: depending on position where photon hits the pmt
  inline int index(double x,double y) const { 
    return index((int)((y-yymin)/yydiv),(int)((x-xxmin)/xxdiv));
    }
    
/// return x-position as function of the pmt-channelindex
  inline double x(int channel) const {
    return (channel%nnx+0.5)*xxdiv+xxmin;
    }

/// return y-position as function of the pmt-channelindex
  inline double y(int channel) const {
    return (channel/nnx+0.5)*yydiv+yymin;
    }

/// return quantum-efficiency as function of the wavelength
  inline float efficiency(double l) const { return eff.f(l); }

///  print the contents of the class to an ASCII-File 
  void print(FILE *fp=stdout);

///  constructor reading information from an ASCII-File produced using 'print'
  PMType(ScriptFile& f);

/// constructor using a record in the CPMT-ARTE-Table
  PMType(const ArtePointer<CPMT>& a);
  
///  destructor
  ~PMType();
  };
   

// -------------------------------------------------------------------------
/// description of  the lense system

class LensSystem {
Table AngEff,WaveEff;
int id;
public:
  LensSystem(ScriptFile& f);
  void print(FILE* fp);

  inline double AngularEfficiency   (double a) const { return AngEff.f(a); }
  inline double WaveLengthEfficiency(double l) const { return WaveEff.f(l); }
  inline int type() const { return id; }
/// trace/transform photon through lense system
  inline int trace(const Photon& p,double *x,double *y) const {
//
//  !!!!!!!!!  sign change
//

    *x=-p.x().x()/2;
    *y=-p.x().y()/2;
//    *x=p.x().x()/2;
//    *y=p.x().y()/2;
    int ret=1;
    if (rnd_normal()>WaveLengthEfficiency(p.lambda())) {
      RICH_Loss.LensWaveLength()++;
      ret=0;
      }
    else if (rnd_normal()>AngularEfficiency(p.n().z())) {
      RICH_Loss.LensAngle()++;
      ret=0;
      }
    return ret;
    }

/// inverse tracing/transformation through the lense system
  inline Vec3 inversetrace(double x,double y,double theta=0,double phi=0) const {
//
//  !!!!!!!!!  introduce factor -2
//
    return Vec3(-2*x,-2*y,0);
//    return Vec3(x,y,0);
    }
  
  };

// ---------------------------------------------------------------------------
///  PM specific information
/**
 The class PM describe a pmt. It stores the type information of the pmt
 and a gain variation factor, which can be different for each channel of
 the pmt.     
 */

class PM {

protected:
  ArtePointer<CPMD> a;
  LensSystem* ls;
  PMType* pmtype;
  char seno[21];
  inline PM() { 
    CPMD d;
    a=ArtePointer<CPMD>::append(d);
    a->efv0=-1;
    pmtype=0;
    strcpy(seno,"undefined");
    };
  
public:
/// detection of a photon, which hits the pmt-channel number k.  
  int dodetect(int k,const Photon& p);
/**
 The function "dodetect" describes the detection of a photon with a pmt. 
 It determine which channelis hit. Using the channel dependent gain variation 
 factor and the wavelength dependent quantum efficiency, which is assumed 
 to be the same for pmt's of the same type, the total detection probability
 is calculated. Using a random generator the decision is taken, if the photon
 is detected.
 The return value indicated, if the photon is detected (return value=1)
 or not (return value=0). In the first case the channelid is return
 with the last argument.
 */
 
/// returns a reference to the pmt-type information
  inline const PMType&  type() const  {  return *pmtype; }
/// returns a pointer to the LensSystem description
  inline const LensSystem*  Collector() const  {  return ls; }

/// returns the serialno of the pmt
  inline char*  serialno() {  return seno; }

/// returns the index of the supermodule index, where the pmt is plugged into
  inline int    SuperModule() const { return a->isup; }

/// returns the row, where the pmt is plugged in.
  inline int    Row() const { return a->irow; }

/// returns the column, where the pmt is plugged in
  inline int    Col() const { return a->icol; }

/// return the channel dependend efficiency (gain) variation factor
  inline double efficiency(int i) const { 
    return ArtePointer<CEFF> (a->efv0+i)->eff;
    }
/// set the channel dependend efficiency (gain) variation factor
  inline float& efficiency(int i) ;

/// shortcut for type().row(i)   (see class PMType for an explanation) 
  inline int   PMrow(int i) const  { return pmtype->row(i); }

/// shortcut for type().col(i)   (see class PMType for an explanation) 
  inline int   PMcol(int i) const  { return pmtype->col(i); }

/// shortcut for type().numchannel() (see class PMType for an explanation)
  inline int   numchannel()    const  { return pmtype->numchannel(); }

/// shortcut for type().numcol() (see class PMType for an explanation) 
  inline int   numcol()        const  { return pmtype->numcol(); }

/// shortcut for type().numrow() (see class PMType for an explanation) 
  inline int   numrol()        const  { return pmtype->numrow(); }

/// shortcut for type().index(i,j) (see class PMType for an explanation) 
  inline int   channelindex(int i,   int j) const  { 
    return pmtype->index(i,j); }

/// shortcut for type().index(x,y) (see class PMType for an explanation) 
  inline int   channelindex(double x,double y) const { 
    return pmtype->index(x,y); }          

///  print the contents of the class to an ASCII-File. See also page: 
  void print(FILE *fp=stdout);

/// constructor reading information from an ASCII-File produced using 'print'
/** constructor reading information from an ASCII-File produced using 'print'. 
 To set the PMT-Type information a List containg all available PMT-Types 
 have to be supplied in the second argument.
  */  
  PM(ScriptFile& f,List<PMType>& PMTypeList,List<LensSystem>& LensSystemList);

/// constructor using a record in the CPMD-ARTE-Table. See also page:
/** constructor using a record in the CPMD-ARTE-Table.
 To set the PMT-Type information a List containg all available PMT-Types 
 have to be supplied in the second argument.
  */  
  PM(const ArtePointer<CPMD>& a,List<PMType>& PMTypeList);

/// return the index in table CEFF (special purpose)
  inline int CEFFindex(int i) const { 
    return a->efv0+i;
    }    

  };

// -------------------------------------------------------------------------
/// description of a cell of a supermodule 
/**
  The set up in one cell of a supermodule is describe using the 
  class PhotonDetectorCell. The describtion of the lense system,
  which is at the moment assumed to be the same for all cells,
  is done in the "dodetect" member function. Up to now the implementation
  of the lense system is only a 2:1 mapping of the x,y-coordinates.
  Work has to be done here.
 */
 
class PhotonDetectorCell {
  PM *ppm;
  int *id;
  inline int ID(int channel) const { return id[channel]; }
  
public:
/// detection of a photon, which hits this cell of the supermodule. See also page:  
  int dodetect(const Photon& p,int *ChannelID);
/**
  The function "dodetect" describes the imaging of the lens system 
  as well as losses, which occur during the transmission. 
  After doing this it is checked, if the pmt detects the photon.
  The return value indicated, if the photon is detected (return value=1)
  or not (return value=0). In the first case the channelid is return
  with the last argument.
 */  
/// best guess position on the front of the lense-system / supermodule
  Vec3 pos(int channel,double costhein=0,double phiin=0);

/// return a reference to the pmt plugged into this cell
  inline PM& pm()            const { return *ppm; }
/// short cut for pm().SuperModule() (see class PM for an explanation)
  inline int SuperModule()   const { return ppm->SuperModule(); }
/// short cut for pm().Row() (see class PM for an explanation)
  inline int Row()           const { return ppm->Row(); }
/// short cut for pm().Col() (see class PM for an explanation)
  inline int Col()           const { return ppm->Col(); }

// constructor have to be supplied with the information about the pmt in the cell
  PhotonDetectorCell(PM* pm);
// destructor to clean up private member functions.
  ~PhotonDetectorCell();
  };


// ------------------------------------------------------------------------

/// description of a supermodule
/**
  A supermodule is described by the class PhotonDetectorSuperModule.
  The front surface of the module is described as an rectangle. 
  The rectangular surface is subdivided into a certain number of rows 
  and columns (32 x 12) and has coordinate system with the x-direction 
  parallel to the short edge, the y-axis is parallel to the long edge 
  of the module.
  For the upper supermodules the y-axis points upwards and the x-axis
  is (nearly) parallel to the HERA-B x-axis. The origin is the lower 
  left corner. For the lower ones the y-axis points downwards and the
  x-axis is (nearly) antiparallel to the HERA-B x-axis. The origin is
  is the upper right corner. In this coordinate systems the index of the
  row and columns are always positiv (incl. zero) and the symmetrie between
  lower and upper half is respected. The momentum vector of the infalling
  photons has it largest component parallel to the z-axis of this coordinate
  system.
*/
class PhotonDetectorSuperModule {
PhotonDetectorCell **pdcell;


protected:
  Rectangle r;
  int numcell,iid;
  inline PhotonDetectorSuperModule() {;};

public:
/// detection of a photon, which hits this cell of the supermodule. See also page:  
/**
  The function "dodetect" is called to detect a photon. It search the hit
  cell, through a determination of the hit row and column. After that
  it ask the cell, if the photon is detected, by calling the "dodetect"
  member function of the class PhotonDetectorCell.
  The return value indicated, if the photon is detected (return value=1)
  or not (return value=0). In the first case the channelid is return
  with the last argument.
*/
  int dodetect(Photon* p,int *daqid);
/// returns the MC-position on the front of the supermodule/lense system of the last hit
  inline Vec3 hitpos() const { return r.TransformToOld(r.hit()); }
/// returns the best-guess position of a hit, if supplied with the row,column of the cell and the pmt-channel
  Vec3 pos(int row,int col,int channel) const;
/// returns a pointer to the cell information and a {\bf zero} pointer, if the cell is empty (!).
  inline  PhotonDetectorCell* Cell(int row,int col) const {
    return pdcell[r.index(row,col)];
    }
/// returns a pointer to the cell information and a {\bf zero} pointer, if the cell is empty (!).
  inline  PhotonDetectorCell* Cell(double x,double y) const {
    return pdcell[r.index(x,y)];
    }
/// Geometrical properties of the front surface of the supermodule/lense system.
  inline const Rectangle& Area() const { return r;}
/// determine intersection point.
  inline int intersection(const Photon& p,double *l) {
    return r.intersection(p,l);
    }
///  returns the number of rows
  inline int numrow()  const { return r.ny(); }
///  returns the number of columns
  inline int numcol()  const { return r.nx(); }
///  returns the size in x
  inline double width()  const { return r.xsize(); }
///  returns the size in y
  inline double height()  const { return r.ysize(); }
///  returns the local coordinate of the last hit
  inline Vec3 hit()  const { return r.hit(); }

/// returns the identifier of the supermodule
  inline int id() const { return iid; }

///  print the contents of the class to an ASCII-File. 
  void print(FILE* fp);
  
/// constructor reading information from an ASCII-File produced using 'print'. See also page:
/** constructor reading information from an ASCII-File produced using 'print'
 To set the PM information a List containg all available PMT's 
 have to be supplied in the second argument.
  */   
  PhotonDetectorSuperModule(ScriptFile& f,
      List<PM>& PMList);
      
/// constructor using a record in the CSUP-ARTE-Table. See also page:
/** constructor using a record in the CSUP-ARTE-Table.
 To set the PM information a List containg all available PMT's 
 have to be supplied in the second argument.
  */   
  PhotonDetectorSuperModule( const ArtePointer<CSUP>& a,
      List<PM>& PMList);
/// destructor
  ~PhotonDetectorSuperModule();  
  };

// ------------------------------------------------------------------------
//
//   the description of the detector: everything is accessable from here
//
//@Man: Using the ChannelID
//@{
/// return the unique channel-ID=1000000*sup+10000*row+100*col+pmchan;
  inline int ChannelID(int sup,int row,int col,int pmchan) {
    return 1000000*sup+10000*row+100*col+pmchan;
    }
/// determine the supermodule-ID using the ChannelID 
  inline int smFromChannelID(int i)     { return i/1000000; }
/// determine the supermodule row using the ChannelID 
  inline int rowFromChannelID(int i)    { return (i%1000000)/10000; }
/// determine the supermodule column  using the ChannelID 
  inline int colFromChannelID(int i)    { return (i%10000)/100; }
/// determine the PMT channel number using the ChannelID 
  inline int pmchanFromChannelID(int i) { return i%100; }
//@}
/// root for the photondetector description
/**
 The Photondetector is described by the class RICHDetector.
 The detector is described as a tree structure, where the first level
 consists out of the supermodule, the second is the description of the
 detector cells, especially which kind of pmt is sitting in the cell.
 The last level describe the pmt.
 The detector consists out of many supermodules, which are described
 using the class PhotonDetectorSupermodule. The number of supermodules
 you can get using the function NumSuperModule() and the information
 of the i'th supermodule you can access using the call "SuperModule(i)".
 */
 

class RICHDetector {
List<PMType>        pmtype_list;
List<PM>            pm_list;
List<LensSystem>    lens_list;

int numsm,lasthit;
PhotonDetectorSuperModule **sm;

int numchannel,firstCEFFindex;
int s,r,c,p,maxp;
public:

/// simulate the detection process of a photon. See also page:
/**
 The function "dodetect" is called to detect a photon. The function
 goes through the tree and check if the photon is detected.
 For this it first searches the hit supermodule. After this it takes
 this supermodule and call the "dodetect"-funtion of the class
 PhotonDetectorSupermodule, which handle the further search through the 
 tree. 
 The return value indicated, if the photon is detected (return value=1)
 or not (return value=0). In the first case the channelid is return
 with the last argument.
 */
   
  int  dodetect(Photon *p,int *ChannelID);
  
/// determine intersection point with the detector plane
  void intersection(const Photon &p,double *path);
  int  hit_in_acceptance(const Photon &p);

///  return the position of the last intersection
  inline Vec3 hitpos() const {
    Vec3 h;
    if (lasthit>=0) h=SuperModule(lasthit).hitpos();
    return h;
    }
/// return the number of supermodules
  inline       int        NumSuperModule() const { return numsm; }
/// returns the i'th supermodule
  inline const PhotonDetectorSuperModule& SuperModule(int i) const { 
    return *sm[i]; 
    }  

/// check if cell is equiped
  inline int ChannelExist(int id) {
    s=smFromChannelID(id);
    r=rowFromChannelID(id);
    c=colFromChannelID(id);
    p=pmchanFromChannelID(id);
//    printf("channelexist %d %2d %2d %2d %2d\n",id,s,r,c,p);
    return (SuperModule(s).Cell(r,c)!=0&&
            p<SuperModule(s).Cell(r,c)->pm().numchannel());
    }
/// check if cell is equiped
  inline int ChannelExist(int ss,int rr,int cc,int pp) {
    s=ss;r=rr;c=cc;p=pp;
    int ret=(SuperModule(s).Cell(r,c)!=0);
    if (ret) ret= (p<SuperModule(s).Cell(r,c)->pm().numchannel());
    return ret;
    }
/// access to the supermodule (to be used after call of CellExist,ChannelLoop)
  inline const PhotonDetectorSuperModule& SuperModule() const {
    return SuperModule(s);
    }
/// access to the cell (to be used after call of CellExist,ChannelLoop)
  inline PhotonDetectorCell& Cell() {
    return *SuperModule(s).Cell(r,c);
    }
/// access to the pm (to be used after call of CellExist,ChannelLoop)
  inline  PM& pm()  {
    return SuperModule(s).Cell(r,c)->pm();
    }
  inline const PM& pm() const {
    return SuperModule(s).Cell(r,c)->pm();
    }
/// returns the pmt-channel (to be used after call of CellExist,ChannelLoop)
  inline int pmChannel() const { return p; }
/// best-guess position as function of the channel ID. See also page:
/**   
 The function "pos" inverses the detection process. From a given channel number
 it determines the best-guess position of the hit on the front of the
 supermodule/lense system (to be used after call of CellExist,ChannelLoop)
 in the HERA-B coordinate system.
 */
  inline Vec3 pos() const { return SuperModule(s).pos(r,c,p);}
/// return pos() transformed to the supermodule coordinates
  inline Vec3 pos_smcoord() const {
    return SuperModule(s).Area().TransformFromOld(pos());
    }
/// index of each channel in an array, all elements are used. (to be used after call of CellExist,ChannelLoop)
  inline int ChannelIndex()  {
    return pm().CEFFindex(p)-firstCEFFindex;
    }

/// access to an existing cell
  inline const PhotonDetectorCell& ExistingCell(int sup,int row,int col) const {
    PhotonDetectorCell* c=SuperModule(sup).Cell(row,col);
    if (!c) {
      fprintf(stderr,"Access to not existing cell: %d %d %d\n",s,r,c);
      exit(1);
      }
    return *c;
    }

/// access to an existing pmt
  inline const PM& ExistingPM(int sup,int row,int col) const {
    return ExistingCell(sup,row,col).pm();
    }

/// total number of channels of the detector
  inline int NumChannel() { return numchannel; }

/// index of each channel in an array, all elements are used.
  inline int ChannelIndex(int s,int r,int c,int p) {
    return ExistingPM(s,r,c).CEFFindex(p)-firstCEFFindex;
    }
/// initialize loop over all existing channels
  void NewChannelLoop();

/// loop over all existing channels
  int ChannelLoop();
/// loop over all existing channels returning the specification of the channel.
  int ChannelLoop(int *sup,int *row,int *col,int *pmtchannel);
/// return if channel is dead
  inline int ChannelIsDead() const { return pm().efficiency(p)<0; }
/// return if channel is dead
  inline int ChannelIsHot() const { return pm().efficiency(p)>1; }
/// return if the channel is working
  inline int ChannelWorks() const { return !(ChannelIsHot()||ChannelIsDead()); }
/// return the channel dependend efficiency factor
  inline double efficiency() const { return pm().efficiency(p); }
/// set the channel dependend efficiency factor
  inline float& efficiency()  { return pm().efficiency(p); }
/// returns the maximum efficiency (for gean simulation cut)
  double MaxEfficiency(double lmin,double lmax);

/// returns the maximum Quantum efficiency (for gean simulation cut)
  double MaxQEfficiency(double lambda);
  
///  print the complete detector description to a set of ASCII-Files. 
  void print();

/// constructor reading information from an ASCII-File produced using 'print'. 
  RICHDetector(char* path=0);

/// destructor
  ~RICHDetector();

  };

extern RICHDetector* RICH_Detector;
#endif
