#ifndef __VESSEL_H__
#define __VESSEL_H__

#include "rich/ricutil.h"
#include "rich/dielectric.h"

// ------------------------------------------------------------------------
/// description of a single spherical mirror
/**
  A single spherical mirror have to have a polygon as boundary.
  The number of corners is limited to six points. The mirror
  have SingleSphericalMirror's as neighbors on the left,right, top
  and bottom side. All top/bottom neighbors defining a column, the
  left/right neighbors a row. 
  */
//
//   All mirror types have the following three corners:
//
//           x
//
//    1  x       x
//
//    2  x       x
//
//         3 x
//
//    They allow to define the geometry of the mirror uniquely
//
//
class SphericalMirror;

class SingleSphericalMirror {

friend class SphericalMirror;
Table  refl;     // mirror reflectifity
int    loss;     // flag showing where photon is lossed

ArtePointer<CSSM> a;

char   seno[21];
char   mytype[10];

void init();

protected:
CoordinateSystem A;
Vec3   coc;      // center of curvature (=center of sphere);
double rr;       // radius of the mirror curvature
double dd;       // radius of the smallest circle surounding the mirror
double xdim;     // horizontal dimension
int    r,c;      // row and column

Vec3   corner[6];
Vec3   nbound[6];
double dbound[6];
int    numcorner;

/// definition of the neighborhood. (not for public use)
SingleSphericalMirror  *top,*bottom,*left,*right; // neighbouring mirrors

/// set the serial number
  inline void serialno(const char* s) {
    int l=strlen(s);
    if (l>sizeof(seno)-1)l=sizeof(seno)-1;
    memcpy(seno,s,l);
    seno[l]='\0';
    }
/// set the mirror type 
  inline void type(const char* s) {
    int l=strlen(s);
    if (l>sizeof(mytype)-1)l=sizeof(mytype)-1;
    memcpy(mytype,s,l);
    mytype[l]='\0';
    }
// constructor (only for special purposes) 
  inline SingleSphericalMirror() { serialno("undefined");}

/// return information about loss type (not for public use);
  inline int  Loss()            const { return loss; }
//  inline int  IsReflected()     const { return (loss==1); }

public:
//@Man:  coordinate system
//@{
/** 
  See also class CoordinateSystem
  */
/// returns center of the mirror 
  inline  const Vec3& x0() const { return A.x0(); } 
/// normal vector: pointing from center to center of curvature
  inline  const Vec3& ez() const { return A.ez(); }
/// returns the horinzontal orientation
  inline  const Vec3& ex() const { return A.ex(); }
/// returns the vertical direction: ey=cross(ez.ey);
  inline  const Vec3& ey() const { return A.ey(); }
/// transform into the mirror coordinate system
  inline  Vec3 TransformFromOld(const Vec3& v)const { 
    return A.TransformFromOld(v);
    }
//@}
/// returns the center of curvature
  inline  const Vec3& CoC() const { return coc; } 
/// returns the row, where the mirror is placed
  inline       int     row()    const { return r; }
/// returns the column, where the mirror is placed
  inline       int     col()    const { return c; }
/// returns the type of the mirror (HEX,FIVE,...)
  inline       char*   type()               { return mytype;};
/// returns the radius of curvature
  inline double radius()      const  { return rr; }
/// returns half of the diameter of the mirror
  inline double halfdiameter() const { return dd; }

//@Man: reflection on the mirror
//@{
/// returns 1, if the photon hits the mirror
  int  inside(const Vec3& v);
/// returns 1, if the photon is reflected on the mirror (0=losses)
  int  reflect(Photon* p);
/// returns 1, if reflexion loss occured
  inline int  ReflexionLoss()   const { return (loss==0); }
/// returns 1, if the photon didn't hit the mirror
  inline int  GeometricalLoss() const { return (loss==2); }
/// returns the reflectivity-curve of the mirror
  inline const Table&  reflectivity() const { return refl; }
/// set the reflectivity curve
  inline       Table&  reflectivity()       { return refl; }
//@}

/// produce paw command to plot the mirror boundary
  void xyview(FILE* fp);
/// print the contents of the class to an ASCII file.
  void print(FILE *fp=stdout);
/// constructor reading the information from an ASCII-FILE. See also page:
/**
 constructor reading the information from an ASCII-FILE produce by'print'.
  In addition a Table providing the reflectivity curve has to be
  supplied using the second argument.
*/
  SingleSphericalMirror(ScriptFile& f,Table& reflectivy);
/// constructor using the ARTE-Table CSSM
  SingleSphericalMirror(const ArtePointer<CSSM>& a);
  };


// ------------------------------------------------------------------------
/// description of the spherical mirror
/**

*/
class SphericalMirror {

protected:

double r,alphah,alphah0,alphav,alphav0;
double radiusprotonhole,radiuselectronhole;
int nummir,splitted;
int speed;
int loss;

SingleSphericalMirror **usm,**lsm;
Table  refl;
double dx,dy;

inline int index(int i,int j) { return i*10+j;}

/// constructor (only for special purposes) 
  inline SphericalMirror() {}

/// reflexion especially for tracing the track direction to the detector  
  int reflectonly(Photon *pp,const Vec3& coc);

public:
/// returns the radius of the hole around the proton beam shroud
  inline double  RadiusProtonHole()   const { return radiusprotonhole; }
/// returns the radius of the hole around the electron beam shroud
  inline double  RadiusElectronHole() const { return radiuselectronhole; }
/// return the width of the mirror (dimension in x-direction)
  inline double  xdim() const { return dx; }
/// return the height of the mirror (dimension in y-direction)
  inline double  ydim() const { return dy; }
/// return the reflectivity curve (at the moment common for a single mirrors)
  inline const Table&  reflectivity() const { return refl; }
/// returns the SingleSphericalMirror at position (row,column)
  inline SingleSphericalMirror* SingleMirror(int row,int col) {
    SingleSphericalMirror *sm;
    int k=index(row,col);
    if (k<0||k>=nummir) return 0;
    if (k<nummir/2) sm=lsm[index(3-row,col)];
    else sm=usm[k-nummir/2];
    return sm;
    }

//@Man: reflection
//@{
/// reflexion of a photon. Return 1, if there are no losses.
/**
  Reflexion of a photon. If no losses happen the photon
  position is changed to the intersection point on the mirror
  and the flight direction is turned to the reflected one.
  In this case the function returns 1. If losses occure the
  function returns 0 and one can analyse the reason using
  the functions ReflexionLoss and GeometricalLoss.
  */
  int reflect(Photon* pp);
/// fast reflexion  
  int fullmirror_reflect(Photon *pp);
/// reflexion especially for tracing the track direction to the detector  
  inline int reflectonly_upper(Photon *pp) { reflectonly(pp,usm[0]->CoC()); }
  inline int reflectonly_lower(Photon *pp) { reflectonly(pp,lsm[0]->CoC()); }
  int geometry_trace(Photon *pp) ;
/// no losses at last reflection
  inline int  IsReflected()     const { return (loss==1); }
/// reflexion loss at last reflexion
  inline int  ReflexionLoss()   const { return (loss==0); }
/// geometrical loss at last reflexion
  inline int  GeometricalLoss() const { return (loss==2); }
/// photon outside of spheres 
  inline int  OutOfSphereLoss() const { return (loss==3); }
/// out of mirror boundary 
  inline int  OutOfBoundaryLoss() const { return (loss==9); }
/// hit proton beam shroud 
  inline int  ProtonShroudLoss() const { return (loss==4); }
/// hit electron beam shrouf
  inline int  ElectronShroudLoss() const { return (loss==5); }
/// aborption in gas 
  inline int  RadiatorAbsorptionLoss() const { return (loss==6); }
/// switch to full simulation in function reflect (default)
  inline void FullSimulation() { speed=0; }
/// switch to fast simulation in function reflect : no subdivision of the mirrors
  inline void FastSimulation() { speed=1; }
//@}      

/// provide a kumac-file, which can be used to plot the mirror in paw.
  void xyview();

/// print the contents of the class to an ASCII file.
  void print(FILE *fp);
/// constructor reading the information from an ASCII-FILE. 
  SphericalMirror(ScriptFile& f);
/// constructor reading the information from thew CSMI ARTE-table. (not ready)
  SphericalMirror(ArtePointer<CSMI> a);
/// destructor
  ~SphericalMirror();
  };


// ------------------------------------------------------------------------
/// description of a single planar mirror
/**
  The class SinglePlaneMirror describe a single planar mirror. Such
  a mirror is a rectangle, described using the class Rectangle. See there
  for additional information.
*/
class PlaneMirror;

class SinglePlaneMirror {
friend class PlaneMirror;

Table refl;
Rectangle r;
int loss;

int    ro,co;
char   ttype[21];
char   seno[21];


  inline int  Loss()            const { return loss; }
  inline int  IsReflected()     const { return (loss==1); }
  inline SinglePlaneMirror() {;}

protected:
  inline SinglePlaneMirror(const Rectangle& rr,Table& rrefl,int rro,int c) {
    r=rr;refl=rrefl;ro=rro;co=c;loss=0;strcpy(seno,"undefined");
    strcpy(ttype,"undefined");
    }

/// sets  Rectangle::xsize() 
  inline  double&  xsize()       { return r.xsize(); }
/// sets  Rectangle::ysize() 
  inline  double&  ysize()       { return r.ysize(); }

public:
//@Man: reflexion on the mirror
//@{
/// returns 1, if the photon is reflected on the mirror (0=losses)
  int  reflect(Photon* p);
  void reflectonly(Photon* p);
  int  geometry_trace(Photon* pp);
  
  inline int intersection(const Photon& p,double *l) { 
    return r.intersection(p,l); 
    }
/// returns 1, if reflexion loss occured
  inline int  ReflexionLoss()   const { return (loss==0); }
/// returns 1, if the photon didn't hit the mirror  int reflect(Photon* pp);
  inline int  GeometricalLoss() const { return (loss==2); }
//@}
/// returns the rectangular surface of the mirror
  inline Rectangle  surface()   const  { return r; }
/// returns  Rectangle::x0() 
  inline  const Vec3& x0() const { return r.x0(); } 
/// returns  Rectangle::ez() 
  inline  const Vec3& ez() const { return r.ez(); } 
/// returns  Rectangle::ex() 
  inline  const Vec3& ex() const { return r.ex(); }
/// returns  Rectangle::ey() 
  inline  const Vec3& ey() const { return r.ey(); }
/// returns  Rectangle::xsize() 
  inline  double   xsize() const { return r.xsize(); }
/// returns  Rectangle::ysize() 
  inline  double   ysize() const { return r.ysize(); }
/// returns the reflectivity curve of the mirror
  inline const Table& reflectivity()  const  { return refl; }
/// returns the row, where the mirror is placed
  inline int  row() const { return ro; }
/// return the column, where the mirror is placed
  inline int  col() const { return co;}
/// return the serial number of the mirror
  inline const char*  serialno() const { return &seno[0];}
/// return the type of the mirror
  inline const char *type() const { return &ttype[0];}

/// assignement: SingleMirror a,b; a=b;
  inline void operator=(const SinglePlaneMirror& sm) {
    refl=sm.refl;
    r=sm.r;
    loss=sm.loss;
    }
/// print the contents of the class to an ASCII-FILE
  void print(FILE* fp=stdout);
/// constructor reading an ASCII-FILE produced by 'print'
  SinglePlaneMirror(ScriptFile& f,Table& reflectivy);
/// constructor using a record in the ARTE--Table CSPM
  SinglePlaneMirror(ArtePointer<CSPM> a);
/// copy constructor
  SinglePlaneMirror(const SinglePlaneMirror &);
  };

// ------------------------------------------------------------------------
/// description of the planar mirror
/**
  At the moment the mirror consist out of 6 flat SinglePlaneMirror's
  for the upper half, and six for the lower half.
  */
class PlaneMirror {

protected:
int       nummirror,splitted;
int       speed;
SinglePlaneMirror **upm,**lpm;
SinglePlaneMirror fullupper,fulllower;

double    radiuselectronhole;
int       loss;
Table refl;

inline PlaneMirror(){;}

int index(int row,int col) {
  return row*nummirror/2+col;
  }
  
public:
/// returns the radius of the hole around the electron beam shroud
  inline double  RadiusElectronHole() const { return radiuselectronhole; }
/// sets the radius of the hole around the electron beam shroud
  inline double& RadiusElectronHole()       { return radiuselectronhole; }
/// returns the SinglePlaneMirror at position (row,column)
  inline SinglePlaneMirror* SingleMirror(int row,int col) {
    SinglePlaneMirror *pm;
    if (col>=nummirror/2) return 0;
    if (row==0) pm=lpm[col];
    else pm=upm[col];
    return pm;
    }
/// return the reflectivity curve (at the moment common for a single mirrors)
  inline const Table&  reflectivity() const { return refl; }
//@Man: reflection
//@{
/// reflexion of a photon. Return 1, if there are no losses.
/**
  Reflexion of a photon. If no losses happen the photon
  position is changed to the intersection point on the mirror
  and the flight direction is turned to the reflected one.
  In this case the function returns 1. If losses occure the
  function returns 0 and one can analyse the reason using
  the functions ReflexionLoss and GeometricalLoss.
  */
  int reflect(Photon* pp);
/// no losses at last reflection
  inline int  IsReflected()     const { return (loss==1); }
/// reflexion loss at last reflexion
  inline int  ReflexionLoss()   const { return (loss==0); }
/// geometrical loss at last reflexion
  inline int  GeometricalLoss() const { return (loss==2); }
/// switch to full simulation in function reflect (default)
  inline void FullSimulation() { speed=0; }
/// switch to fast simulation in function reflect : no subdivision of the mirrors
  inline void FastSimulation() { speed=1; }
/// fast reflection 
  int fullmirror_reflect(Photon* pp);
/// reflexion especially for tracing the track direction to the detector  
  inline void reflectonly_upper(Photon* pp) { fullupper.reflectonly(pp); }
  inline void reflectonly_lower(Photon* pp) { fulllower.reflectonly(pp); }
  int  geometry_trace(Photon* pp);
//@}

/// print the contents of the class to an ASCII file.
  void print(FILE* fp);
/// constructor reading from an ASCII-FILE produced using 'print'.  PlaneMirror(ScriptFile& f);
  PlaneMirror(ScriptFile&);
/// destructor
  ~PlaneMirror();
  };

// ------------------------------------------------------------------------
/// description of the photon tracing in the vessel
/**
 Description of the photon tracing in the vessel
*/
class Vessel {
Tube eshroud,pshroud;
//double xdim,ydim,zdim;
DiElectric* radiator;

public:
/// photon tracing through the radiator over a the distance 'l'
  int trace(Photon* p,const double l,int *loss);   
/// return the Tube describing the electron beam shroud (See class Tube).
  inline const Tube& ElectronBeamShroud() const { return eshroud; }
/// return the Tube describing the proton beam shroud (See class Tube).
  inline const Tube& ProtonBeamShroud()   const { return pshroud; }
/// set the Tube describing the electron beam shroud (See class Tube).
  inline       Tube& ElectronBeamShroud()       { return eshroud; }
/// set the Tube describing the proton beam shroud (See class Tube).
  inline       Tube& ProtonBeamShroud()         { return pshroud; }
//  inline const DiElectric& Radiator()     const { return *radiator; }
/// returns the dielectric material describing the radiator
  inline       DiElectric& Radiator()           { return *radiator; }
/// set the dielectric material describing the radiator
  inline       void  Radiator(DiElectric* r)    { radiator=r; }
/// print the contents of the class to an ASCII file.
  void print(FILE *fp);
/// constructor reading from an ASCII-FILE produced using 'print'.
  Vessel(ScriptFile& f);
// constructor only for the special purposes
  Vessel();
  };
  
// ------------------------------------------------------------------------
/// description of the photon tracing through the photon detector entry window
/**
  Description of the photon tracing through the photon detector entry window
  using a rectangle for the description of the surface inside the vessel.
  A thickness is assigned to the window and a dielectric material (Plexi).
 */
class PhotonDetectorEntryWindow {
int splitted;
DiElectric *mat;
Rectangle *uw,*lw;
double    uwthickness,lwthickness;

public:
/// returns 1, if the photon has an intersection point with the window
  int intersection(Photon* pp);
/// returns 1, if no losses occure during tracing the photon to the window
  int trace(Photon* pp);
/// tracing the photon to the window
  void traceonly_upper(Photon* pp);
  void traceonly_lower(Photon* pp);
/// returns the dielectric material, describing the window material (see also class DiElectric)
  inline const DiElectric& Material()             const { return *mat; }
/// set the dielectric material, describing the window material
  inline       void        Material(DiElectric* m)      { mat=m; }
// returns the rectangle, describing the surface of the upper window
  inline const Rectangle&  UpperWindow()          const { return *uw; }
// set the rectangle describing the surface off the upper window
  inline       void        UpperWindow(Rectangle* r)    { uw=r; }
// returns the rectangle, describing the surface of the lower window 
  inline const Rectangle&  LowerWindow()          const { return *lw; }
// set the rectangle describing the surface off the lower window
  inline       void        LowerWindow(Rectangle* r)    { lw=r; }
/// returns the thickness of the upper window
  inline       double      UpperWindowThickness() const { return uwthickness; }
/// set the thickness of the upper window
  inline       double&     UpperWindowThickness()       { return uwthickness; }
/// returns the thickness of the lower window
  inline       double      LowerWindowThickness() const { return lwthickness; }
/// set the thickness of the lower window
  inline       double&     LowerWindowThickness()       { return lwthickness; }
/// print the contents of the class to an ASCII file.
  void print(FILE* fp=stdout);
/// constructor reading from an ASCII-FILE produced using 'print'.
  PhotonDetectorEntryWindow(ScriptFile&);
/// destructor
  ~PhotonDetectorEntryWindow();
// constructor only for special purposes
  PhotonDetectorEntryWindow();
  };


// ------------------------------------------------------------------------


class RICHVessel {

public:


SphericalMirror*           spherical_mirror;
PlaneMirror*              plane_mirror;
PhotonDetectorEntryWindow* photondetector_window;
Vessel*                    vessel;
inline  DiElectric& Radiator() { return vessel->Radiator(); }
inline  Tube& ProtonBeamShroud()  { return vessel->ProtonBeamShroud();}
inline  Tube& ElectronBeamShroud()  { return vessel->ElectronBeamShroud(); }

/// return 1, if the photon could be successfully trace until it leaves the photon detector entry window, else a zero is returned
int raytrace(Photon *p);
/// transformation of v, which removes the planar mirror and the rotation of the spherical mirror.
void track_upper(Photon* p,double *l);
void track_lower(Photon* p,double *l);
int  geometry_trace(Photon* p);
/// print the configuration of the vessel to an ASCII file (vessel.dat) .
void print();
/// constructor reading from an ASCII-FILE produced using 'print'.
RICHVessel(char* path=0);
/// destructor
~RICHVessel();
};

// ---------------------------------------------------------------------------

extern RICHVessel   *RICH_Vessel;

  
// ---------------------------------------------------------------------------

#endif
