#include "rich/vessel.h"

RICHVessel *RICH_Vessel=0;
extern int RICHdebug;
// ---------------------------------------------------------------------------

inline int reflexion(Photon* p,const Vec3& ns,double reflectivity) {
  p->n()-=double(2)*(ns*p->n())*ns;  // short hand formula for the reflection
  return rnd_normal()<reflectivity;
  }

inline int vesseltrace(Photon* p,const double l,int *loss) {
  return RICH_Vessel->vessel->trace(p,l,loss);
  }

// ===========================================================================

int Vessel::trace(Photon* p,const double l,int *loss) {
  double ll;
  *loss=4;
  if (ProtonBeamShroud().intersection(*p,&ll)&&ll<=l) return 0;
  *loss=5;
  if (ElectronBeamShroud().intersection(*p,&ll)&&ll<l) return 0;
  *loss=6;
  return radiator->trace(p,l);
  }


// -------------------------------------------------------------------------
//
//

void Vessel::print(FILE *fp) {
  fprintf(fp,"#\nVessel\n");
  fprintf(fp,"#\n ProtonBeamShroud\n");
  ProtonBeamShroud().print(fp,0);
  fprintf(fp,"#\n ElectronBeamShroud\n");
  ElectronBeamShroud().print(fp,0);
  fprintf(fp,"#\n# Radiator\n");
  radiator->print(fp);
  fprintf(fp,"  end  # Vessel\n"); 
  }

// -------------------------------------------------------------------------
//
//

Vessel::Vessel(ScriptFile& f) {
  char n[256],tag[256];
  while (f.getline(n)) {
    char *s=n;
    sscanf(s,"%s",tag);s+=f.itemlength(s,1);
    if (!strcmp(tag,"ProtonBeamShroud")) {
      pshroud=Tube(f);
      }
    else if (!strcmp(tag,"ElectronBeamShroud")) {
      eshroud=Tube(f);
      }
    else if (!strcmp(tag,"dielectric")) {
      radiator=new DiElectric(f);  
      }
    else if (!strcmp(tag,"end")) break;
    else {
      fprintf(stderr,
         "Vessel(ScriptFile):  Unknown tag: %s\n",tag);
      exit(1);
      }
    }   

  CVES a;
  a.pori[0]=ProtonBeamShroud().x0().x();
  a.pori[1]=ProtonBeamShroud().x0().y();
  a.pori[2]=ProtonBeamShroud().x0().z();
  a.pez [0]=ProtonBeamShroud().ez().x();
  a.pez [1]=ProtonBeamShroud().ez().y();
  a.pez [2]=ProtonBeamShroud().ez().z();
  a.prad   =ProtonBeamShroud().radius();
  a.plen   =ProtonBeamShroud().length();
  a.eori[0]=ElectronBeamShroud().x0().x();
  a.eori[1]=ElectronBeamShroud().x0().y();
  a.eori[2]=ElectronBeamShroud().x0().z();
  a.eez [0]=ElectronBeamShroud().ez().x();
  a.eez [1]=ElectronBeamShroud().ez().y();
  a.eez [2]=ElectronBeamShroud().ez().z();
  a.erad   =ElectronBeamShroud().radius();
  a.elen   =ElectronBeamShroud().length();
  a.irad   =radiator->ArteTableIndex();
  ArtePointer<CVES>::append(a);
  }

Vessel::Vessel() {
  if (ArtePointer<CVES>::nrow()) {
    ArtePointer<CVES> a;
    CoordinateSystem A;
    A.setrotation(Vec3(a->pez));
    ProtonBeamShroud().x0()=Vec3(a->pori);
    ProtonBeamShroud().ez()=A.ez();
    ProtonBeamShroud().ez()=A.ex();
    ProtonBeamShroud().ey()=A.ey();
    ProtonBeamShroud().radius()=a->prad;
    ProtonBeamShroud().length()=a->plen;
    A.setrotation(Vec3(a->eez));
    ElectronBeamShroud().x0()=Vec3(a->eori);
    ElectronBeamShroud().ez()=A.ez();
    ElectronBeamShroud().ez()=A.ex();
    ElectronBeamShroud().ey()=A.ey();
    ElectronBeamShroud().radius()=a->erad;
    ElectronBeamShroud().length()=a->elen; 
    radiator=new DiElectric(ArtePointer<CDIM>(a->irad));   
    }
  }

// ===========================================================================
//
//   reflection algorithm used for every spherical mirror
//

int SingleSphericalMirror::reflect(Photon* pp) {
  Photon& p=*pp;
  loss=3;
//
//  get intersection point
//
  Vec3   v =p.x()-coc;
  double v2=v*v;
  if (v2>radius()*radius()) return 0; // out of sphere
  double vn=v*p.n();
  double l=-vn+sqrt(radius()*radius()-v2+vn*vn);

//
//  trace through gas
//
  if (!vesseltrace(&p,l,&loss)) return 0;
//
//  throw photon away, if no mirror is hit
//
  loss=2;
  if (!inside(A.TransformFromOld(p.x()))) return 0;
// 
//
// photon hit mirror: now do reflection
//
// normal of surface on hit position:   normalize(p.x()-coc)
//
  return (loss=
            reflexion(&p,normalize(p.x()-coc),reflectivity().f(p.lambda()))); 
  }

// -------------------------------------------------------------------------
//
//  check if distance to plane has always the same sign <=> point is inside 
//  v has to be given in local coordinates
//

int SingleSphericalMirror::inside(const Vec3& v) {
  if (fabs(v.x())>xdim      ) return 0;
  if (fabs(v.y())>halfdiameter()) return 0;
  int i;
  int is,side=0;
  for (i=0;i<numcorner;i++) {
    float s = nbound[i]*v+dbound[i];
    is= (s<0) ? -1 : 1;
    if (side==0) side=is;
    else if (!(is==side)) break;
    }
  return (is==side);
  }


// -------------------------------------------------------------------------
//
//  set up geometry of the boundary
//

void SingleSphericalMirror::init() {
  A.ey()=cross(ez(),ex());
  coc=x0()+radius()*ez();  

  double s=halfdiameter();
  double r=radius();

  if (RICHdebug) {printf("mytype %s\n",mytype);fflush(stdout);}
  if (strncmp(mytype,"HEX",3)==0) {
    corner[0]=Vec3(    0         ,+s   ,r-sqrt(r*r-s*s));
    corner[1]=Vec3(-sqrt(3./4.)*s,+s/2.,r-sqrt(r*r-s*s)); 
    corner[2]=Vec3(-sqrt(3./4.)*s,-s/2.,r-sqrt(r*r-s*s));
    corner[3]=Vec3(    0         ,-s   ,r-sqrt(r*r-s*s));
    corner[4]=Vec3(+sqrt(3./4.)*s,-s/2.,r-sqrt(r*r-s*s)); 
    corner[5]=Vec3(+sqrt(3./4.)*s,+s/2.,r-sqrt(r*r-s*s));
    numcorner=6;
    }  
  else if (strncmp(mytype,"FIVE",4)==0) {
    corner[0]=Vec3(-sqrt(3./4.)*s,+s/2.,r-sqrt(r*r-s*s)); 
    corner[1]=Vec3(-sqrt(3./4.)*s,-s/2.,r-sqrt(r*r-s*s));
    corner[2]=Vec3(    0         ,-s   ,r-sqrt(r*r-s*s));
    corner[3]=Vec3(+sqrt(3./4.)*s,-s/2.,r-sqrt(r*r-s*s)); 
    corner[4]=Vec3(+sqrt(3./4.)*s,+s/2.,r-sqrt(r*r-s*s));
    numcorner=5;
    }  
  else if (strncmp(mytype,"HALF",4)==0) {
    corner[0]=Vec3(    0         ,-s   ,r-sqrt(r*r-s*s));
    corner[1]=Vec3(+sqrt(3./4.)*s,-s/2.,r-sqrt(r*r-s*s)); 
    corner[2]=Vec3(+sqrt(3./4.)*s,+s/2.,r-sqrt(r*r-s*s));
    corner[3]=Vec3(    0         ,+s   ,r-sqrt(r*r-s*s));
    numcorner=4;
    }  
  else if (strncmp(mytype,"FOUR",4)==0) {
    corner[0]=Vec3(-sqrt(3./4.)*s,+s/2.,r-sqrt(r*r-s*s)); 
    corner[1]=Vec3(-sqrt(3./4.)*s,-s/2.,r-sqrt(r*r-s*s));
    corner[2]=Vec3(    0         ,-s   ,r-sqrt(r*r-s*s));
    corner[3]=Vec3(    0         ,+s/2 ,r-sqrt(r*r-s*s)); 
    numcorner=4;
    }
  int i;
  if (x0().y()<0) {
    for (i=0;i<numcorner;i++) {
      corner[i][1]=-corner[i][1];
      }
    }
  for (i=0;i<numcorner;i++) {
    int j= (i==0) ? numcorner-1 : i-1;
    nbound[i][0]= (corner[i][1]-corner[j][1]);
    nbound[i][1]=-(corner[i][0]-corner[j][0]);
    nbound[i][2]= 0;
    dbound[i]=-(corner[i]*nbound[i]);
    }

  }

// -------------------------------------------------------------------------
//
//    initialisation from input file  
//

SingleSphericalMirror::SingleSphericalMirror(ScriptFile& f,Table& rrefl) :
  a(ArtePointer<CSSM>(ArtePointer<CSSM>::nrow()+1)){
  char n[256],tag[256];

  refl=rrefl;
  loss=-1;
  
  strcpy(mytype,"\0");
  while (f.getline(n)) {
    char *s=n;
    sscanf(s,"%s",tag);s+=f.itemlength(s,1);
    if (RICHdebug) {printf("%sXXX %s",tag,n);fflush(stdout);}
    int err=0;
    if      (!strcmp(tag,"index")) { 
      err=(sscanf(s,"%d %d",&r,&c)!=2);
      }  
    else if (!strcmp(tag,"type")) {
      err=(sscanf(s,"%s",&mytype[0])!=1);  
      }
    else if (!strcmp(tag,"mirrorname"))   
      err=(sscanf(s,"%s",&seno[0])!=1);      
    else if (!strcmp(tag,"center")) 
      err=(sscanf(s,"%lf %lf %lf",&A.x0().x(),&A.x0().y(),&A.x0().z())!=3);
    else if (!strcmp(tag,"normal")) 
      err=(sscanf(s,"%lf %lf %lf",&A.ez().x(),&A.ez().y(),&A.ez().z())!=3);
    else if (!strcmp(tag,"horizontal")) 
      err=(sscanf(s,"%lf %lf %lf",&A.ex().x(),&A.ex().y(),&A.ex().z())!=3);
    else if (!strcmp(tag,"radiusofcurvature")) 
      err=(sscanf(s,"%lf",&rr)!=1);
//    else if (!strcmp(tag,"horizontal_dimension")) 
//      err=(sscanf(s,"%lf",&xdim)!=1);
    else if (!strcmp(tag,"halfdiameter")) 
      err=(sscanf(s,"%lf",&dd)!=1);
    else if (!strcmp(tag,"end"))  break;
    else {
      fprintf(stderr,
         "SingleSphericalMirror(ScriptFile):  Unknown tag: %s\n",tag);
      exit(0);
      }
/*
    if (err) {
      fprintf(stderr,
         "SingleSphericalMirror(ScriptFile):  Error during scan: %d\n",err);
      exit(0);
      }
*/
    }   
  if (RICHdebug) {print();fflush(stdout);}

  xdim=sqrt(3./4.)*dd;
  CSSM ssm;
  ssm.x0[0]=x0().x();
  ssm.x0[1]=x0().y();
  ssm.x0[2]=x0().z();
  ssm.norm[0]=radius()*ez().x();
  ssm.norm[1]=radius()*ez().y();
  ssm.norm[2]=radius()*ez().z();
  ssm.xvec[0]=xdim*ex().x();
  ssm.xvec[1]=xdim*ex().y();
  ssm.xvec[2]=xdim*ex().z();
  ssm.hdia=halfdiameter();
  ssm.iref=reflectivity().firstpoint();
  ssm.nref=reflectivity().numpoints();
  ArtePointer<CSSM>::append(ssm);

  
  if (RICHdebug) {printf("call init\n");fflush(stdout);}
  init();
  if (RICHdebug) {printf("init\n");fflush(stdout);}
  }
//
//    print data  
//

void SingleSphericalMirror::print(FILE *fp) {
  fprintf(fp,"#\nSingleSphericalMirror\n");
  fprintf(fp,"  index                 %d %d\n",row(),col());
  fprintf(fp,"  type                  %s\n",type());
  fprintf(fp,"  mirrorname            %s\n",seno);
  fprintf(fp,"  center                %12lf %12lf %12lf\n",x0().x(),x0().y(),x0().z());
  fprintf(fp,"  normal                %12lf %12lf %12lf\n",ez().x(),ez().y(),ez().z());
  fprintf(fp,"  horizontal            %12lf %12lf %12lf\n",ex().x(),ex().y(),ex().z());
  fprintf(fp,"  radiusofcurvature     %12lf\n",radius()); 
//  fprintf(fp,"  horizontal_dimension  %12lf\n",xdim); 
  fprintf(fp,"  halfdiameter          %12lf\n",halfdiameter()); 
  fprintf(fp,"  end\n"); 
/*
int i;
fprintf(fp,"# ");ey().print("%f %f %f\n",fp);
  for (i=0;i<numcorner;i++) {
fprintf(fp,"# ");corner[i].print("%f %f %f\n",fp);
  }
*/
  }


//
//   initialisation using an ARTE-pointer  
//

SingleSphericalMirror::SingleSphericalMirror(const ArtePointer<CSSM>& aa) 
        : a(aa),refl(a->iref,a->nref) {
  loss=-1;
  A.x0()=Vec3(a->x0);
  A.ez()=Vec3(a->norm);
  A.ex()=Vec3(a->xvec);
  xdim=norm(A.ex());
  rr  =norm(A.ez());
  dd  =a->hdia;
  A.ex()/=xdim;
  A.ez()/=rr;
  memcpy(mytype,a->type,4);
  int i;
  for (i=4;i>=0;i--) if (mytype[i]!=' ') break;
  mytype[i+1]='\0';
  }

//
//    to have a look at the mirror with paw  
//

void SingleSphericalMirror::xyview(FILE* fp) {
  int i,j=numcorner-1;
  fprintf(fp,"* numcorner %d\n",numcorner);
  for (i=0;i<numcorner;i++) {
    Vec3 c1=A.TransformToOld(corner[j]);
    Vec3 c2=A.TransformToOld(corner[i]);
    fprintf(fp,"line %f %f %f %f\n",c1.x(),c1.y(),c2.x(),c2.y());
    j=i;
    }
  }


// ===========================================================================
//
//    Spherical Mirror

  
int SphericalMirror::reflectonly(Photon* pp,const Vec3& coc) {
  Photon& p=*pp;
//
//  get intersection point and decide if lower or upper mirror is hit.
//
  Vec3 v=p.x()-coc;
  double l,v2=v*v;
  if (v2<=r*r) {
    double vn=v*p.n();
    l=-vn+sqrt(r*r-v2+vn*vn);
    p.x()+=l*p.n();
    Vec3 ns=normalize(p.x()-coc);
    p.n()-=double(2)*(ns*p.n())*ns;  // short hand formula for the reflection
    return 1;
    }
  return 0;
  }
  
int SphericalMirror::geometry_trace(Photon* pp) {
  SingleSphericalMirror **m=0,*sm;
  Photon& p=*pp;
  loss=3;
  //
//  get intersection point and decide if lower or upper mirror is hit.
//
  Vec3 coc=usm[0]->CoC();
  Vec3 v=p.x()-coc;
  double v2=v*v;
  double l;
  if (v2<=r*r) {
    double vn=v*p.n();
    l=-vn+sqrt(r*r-v2+vn*vn);
    if (l<0) return 0;
    v=p.x()+l*p.n();  // intersection point in global coordinates
    }
  else if (v.y()>=0) return 0;
  if (v.y()<0) {
    coc=lsm[0]->CoC();
    v=p.x()-coc;
    double v2=v*v;
    if (v2>r*r) return 0;    
    double vn=v*p.n();
    l=-vn+sqrt(r*r-v2+vn*vn);
    v=p.x()+l*p.n();  // intersection point in global coordinates
    }
//
//  trace through gas
//

  double ll;
  loss=4;
  if (RICH_Vessel->ProtonBeamShroud().intersection(p,&ll)&&ll<l) return 0;
  loss=5;
  if (RICH_Vessel->ElectronBeamShroud().intersection(p,&ll)&&ll<l) return 0;
  loss=1;
 
  p.x()=v;
  Vec3 ns=normalize(p.x()-coc);
  p.n()-=double(2)*(ns*p.n())*ns;  // short hand formula for the reflection
  return 1; 
  }

  
int SphericalMirror::fullmirror_reflect(Photon* pp) {
  SingleSphericalMirror **m=0,*sm;
  Photon& p=*pp;
  loss=3;

//
//  get intersection point and decide if lower or upper mirror is hit.
//
  Vec3 coc=usm[0]->CoC();
  Vec3 v=p.x()-coc;
  double l,v2=v*v;
  if (v2<=r*r) {
    double vn=v*p.n();
    l=-vn+sqrt(r*r-v2+vn*vn);
    if (l<0) return 0;
    }
  if (v.y()<0) {
    coc=lsm[0]->CoC();
    v=p.x()-coc;
    double v2=v*v;
    if (v2>r*r) return 0;    
    double vn=v*p.n();
    l=-vn+sqrt(r*r-v2+vn*vn);
    }

  Vec3 vv=v-coc;
  int col=(int)((atan(vv.x()/vv.z())-alphah0)/alphah);
  int row;
  if (v.y()<0) row=(int)((+alphav0-asin(vv.y()/r))/alphav);
  else         row=(int)((+alphav0+asin(vv.y()/r))/alphav);
  int k=index(row,col);
  if (k<0||k>=40) return 0;

//
//  trace through gas
//
  if (!vesseltrace(&p,l,&loss)) return 0;
  loss=2;
  if (fabs(p.x().x())<xdim()||fabs(p.x().y())>ydim()) return 0;
  loss=4;
  if (norm(p.x()-RICH_Vessel->ProtonBeamShroud().x0())  <RadiusProtonHole())
  return 0;
  loss=5;
  if (norm(p.x()-RICH_Vessel->ElectronBeamShroud().x0())<RadiusElectronHole())
  return 0;

  return (loss=
     reflexion(&p,normalize(p.x()-coc),reflectivity().f(p.lambda()))); 
  }
  
 
//
//  reflect: only search of the hit single mirror, no physics here
//

int SphericalMirror::reflect(Photon* pp) {
  if (speed==1) return fullmirror_reflect(pp);
  
  SingleSphericalMirror **m=0,*sm;
  Photon& p=*pp;
  loss=3;
  //
//  get intersection point and decide if lower or upper mirror is hit.
//
  Vec3 v=p.x()-usm[0]->CoC();
  double v2=v*v;
  if (v2<=r*r) {
    double vn=v*p.n();
    double l=-vn+sqrt(r*r-v2+vn*vn);
    if (l<0) return 0;
    v=p.x()+l*p.n();  // intersection point in global coordinates
    if (v.y()>=0) m=usm;
    }
  else if (v.y()>=0) return 0;
  if (v.y()<0) {
    v=p.x()-lsm[0]->CoC();
    double v2=v*v;
    if (v2>r*r) return 0;    
    double vn=v*p.n();
    double l=-vn+sqrt(r*r-v2+vn*vn);
    v=p.x()+l*p.n();  // intersection point in global coordinates
    m=lsm;
    }
  loss=9;
  Vec3 vv=v-m[0]->CoC();
  int col=(int)((atan(vv.x()/vv.z())-alphah0)/alphah);
  int row;
  if (v.y()<0) row=(int)((+alphav0-asin(vv.y()/r))/alphav);
  else         row=(int)((+alphav0+asin(vv.y()/r))/alphav);
  int k=index(row,col);
  if (RICHdebug) {
    if(k<0||k>=40) {
      p.x().print();p.n().print();
      vv.print();printf("row %d %d %d\n",row,col,k);
      return 0;
      }
    }
  if (k<0||k>=40) return 0;

  loss=10;
  sm=m[k];
// check special cases
  if (!sm->inside(sm->TransformFromOld(v))) {
    if (row%2==0) {
      if (!(sm=sm->top)) return 0;
      if (!sm->inside(sm->TransformFromOld(v))) {
        if (!(sm=sm->left)) return 0;
        }
      }
    else {
      if (!sm->left) {
        if (!(sm=sm->top)) return 0;
        }
      else {
        sm=sm->left;
        if (!sm->inside(sm->TransformFromOld(v))) {
          if (!(sm=sm->right->top)) return 0;
          }
        }
      }
    }
  sm->reflect(pp);
  loss=sm->Loss();
  if (!IsReflected()) return 0;
  return 1;
  }
  /*
  loss=4;
  double dx=p.x().x()-RICH_Vessel->ProtonBeamShroud().x0().x();
  double dy=p.x().y()-RICH_Vessel->ProtonBeamShroud().x0().y();
  if (norm(p.x()-RICH_Vessel->ProtonBeamShroud().x0())  <RadiusProtonHole()*RadiusProtonHole()) return 0;
  loss=5;
  if (norm(p.x()-RICH_Vessel->ElectronBeamShroud().x0())<RadiusElectronHole()) return 0;
  loss=0;
  return 1;
  }

*/
// -------------------------------------------------------------------------
//
//  print data
//

void SphericalMirror::print(FILE *fp) {
  fprintf(fp,"#\nSphericalMirror\n");
  fprintf(fp,"  horizontal_tiling    %lf %lf\n",alphah,alphah0);
  fprintf(fp,"  vertical_tiling      %lf %lf\n",alphav,alphav0);
  fprintf(fp,"  radiusofcurvature    %lf\n",r);
  fprintf(fp,"  radius_proton_hole   %lf\n",radiusprotonhole);
  fprintf(fp,"  radius_electron_hole %lf\n",radiuselectronhole);
  fprintf(fp,"  numberofmirrors      %d\n",nummir);
//  fprintf(fp,"  splitted          %d\n",split);
  fprintf(fp,"#\n reflectivity\n");
  usm[0]->reflectivity().print(fp,0);
  fprintf(fp,"  end\n");
  if (splitted) {
    int i;
    for (i=0;i<nummir/2;i++) lsm[i]->print(fp);
    for (i=0;i<nummir/2;i++) usm[i]->print(fp);
    }
  }


//
//  initialisation from input file
//

SphericalMirror::SphericalMirror(ScriptFile& f) {
  char n[256],tag[256];

  FullSimulation();
  loss=-1;
  splitted=1;
  while (f.getline(n)) {
    char *s=n;
    sscanf(s,"%s",tag);s+=f.itemlength(s,1);
    if (RICHdebug) { printf("%s %s",tag,n);fflush(stdout);}
    int err=0;
    if      (!strcmp(tag,"horizontal_tiling"))   
      err=(sscanf(s,"%lf %lf",&alphah,&alphah0)!=2);  
    else if (!strcmp(tag,"vertical_tiling"))   
      err=(sscanf(s,"%lf %lf",&alphav,&alphav0)!=2);  
    else if (!strcmp(tag,"radiusofcurvature"))   
      err=(sscanf(s,"%lf",&r)!=1);  
    else if (!strcmp(tag,"radius_proton_hole")) {
      err=(sscanf(s,"%lf",&radiusprotonhole)!=1); 
      } 
    else if (!strcmp(tag,"radius_electron_hole")) {
      err=(sscanf(s,"%lf",&radiuselectronhole)!=1); 
      } 
    else if (!strcmp(tag,"numberofmirrors"))  
      err=(sscanf(s,"%d",&nummir)!=1); 
    else if (!strcmp(tag,"splitted"))   
      err=(sscanf(s,"%d",&splitted)!=1);  
    else if (!strcmp(tag,"reflectivity")) refl=Table(f);
    else if (!strcmp(tag,"end"))  break;
    else {
      fprintf(stderr,
         "SphericalMirror(ScriptFile):  Unknown tag: %s\n",tag);
      exit(0);
      }
    }   

  dx=r*sin(alphah0);
  dy=2*r*sin((4.*alphav)/2.)*cos(M_PI-alphav0-(M_PI/2-(4.*alphav)/2.));
  
  CSMI sm;
  sm.ah =alphah;
  sm.ah0=alphah0;
  sm.av =alphav;
  sm.av0=alphav0;
  sm.r  =r;
  sm.rpc=radiusprotonhole;
  sm.rec=radiuselectronhole;
  ArtePointer<CSMI>::append(sm);

  if (splitted) {
    usm=new SingleSphericalMirror*[nummir/2];
    lsm=new SingleSphericalMirror*[nummir/2];
    }
  else {
    usm=new SingleSphericalMirror*[nummir];
    }

  int i;
  SingleSphericalMirror* m;
  for (i=0;i<nummir;i++) {
    f.getline(n);sscanf(n,"%s",tag);
    if (RICHdebug) { printf("%s %s",tag,n);fflush(stdout);}
    if (strcmp(tag,"SingleSphericalMirror")) {
      fprintf(stderr,"SphericalMirror(ScriptFile): Not enough single spherical "
              "mirrors are defined in the input file. Abort\n");
      exit(1);
      }
    m=new SingleSphericalMirror(f,refl);
    if (RICHdebug) { printf(" %d %d\n",m->row(),m->col());fflush(stdout); }
    int k=index(m->row(),m->col());
    if (splitted) {
      if (k<nummir/2) { k=index(3-m->row(),m->col());lsm[k]=m;}
      else usm[k-nummir/2]=m;
      }
    else usm[k]=m;
    }
  if (!splitted) return;

//
//  define neighbourhood in local coordinate:
//    pay attention: for the lower mirror the coordinatesystem is rotated
//
  int j;
  for (j=0;j<10;j++) {
    usm[index(0,j)]->bottom=0;
    usm[index(0,j)]->top   =usm[index(1,j)];  
    usm[index(1,j)]->bottom=usm[index(0,j)];
    usm[index(1,j)]->top   =usm[index(2,j)];  
    usm[index(2,j)]->bottom=usm[index(1,j)];
    usm[index(2,j)]->top   =usm[index(3,j)];  
    usm[index(3,j)]->bottom=usm[index(2,j)];
    usm[index(3,j)]->top   =0;  
    for (i=0;i<4;i++) {
      usm[index(i,j)]->left=0;
      if (j==0) {
        usm[index(i,j)]->right=0;
        }
      else {
        usm[index(i,j-1)]->left =usm[index(i,j)];
        usm[index(i,j)  ]->right=usm[index(i,j-1)];
        }
      }

    lsm[index(0,j)]->bottom=0;
    lsm[index(0,j)]->top   =lsm[index(1,j)];  
    lsm[index(1,j)]->bottom=lsm[index(0,j)];
    lsm[index(1,j)]->top   =lsm[index(2,j)];  
    lsm[index(2,j)]->bottom=lsm[index(1,j)];
    lsm[index(2,j)]->top   =lsm[index(3,j)];  
    lsm[index(3,j)]->bottom=lsm[index(2,j)];
    lsm[index(3,j)]->top   =0;  
    for (i=0;i<4;i++) {
      lsm[index(i,j)]->left=0;
      if (j==0) {
        lsm[index(i,j)]->right=0;
        }
      else {
        lsm[index(i,j-1)]->left =lsm[index(i,j)];
        lsm[index(i,j)  ]->right=lsm[index(i,j-1)];
        }
      }
    }
/*
  int r,c;
  for (r=0;r<4;r++) {
    for (c=0;c<10;c++) {
      SingleSphericalMirror* s=usm[index(r,c)];
      printf("mirror %d %d %d %d\n",r,c,s->row(),s->col());
      if (s->top)    printf("top %d %d\n",s->top->row(),s->top->col());
      if (s->bottom) printf("bot %d %d\n",s->bottom->row(),s->bottom->col());
      if (s->left)   printf("lef %d %d\n",s->left->row(),s->left->col());
      if (s->right)  printf("rig %d %d\n",s->right->row(),s->right->col());
      }
    }
*/
  if (RICHdebug) { xyview();printf("SphericalMirror end\n");fflush(stdout);  }
  }


SphericalMirror::SphericalMirror(ArtePointer<CSMI> a) {
  FullSimulation();
  loss=-1;
  splitted=1;

  r      =a->r;
  alphah =a->ah;
  alphah0=a->ah0;
  alphav =a->av;
  alphav0=a->av0;

  if (!(nummir=ArtePointer<CSSM>::nrow())) return;
  if (splitted) {
    usm=new SingleSphericalMirror*[nummir/2];
    lsm=new SingleSphericalMirror*[nummir/2];
    }
  else {
    usm=new SingleSphericalMirror*[nummir];
    }
  int i;
  for (i=0;i<nummir;i++) {
    if (splitted) {
      if (i<nummir/2) usm[i]=new SingleSphericalMirror(ArtePointer<CSSM>(i+1));
      else lsm[i-nummir/2]=new SingleSphericalMirror(ArtePointer<CSSM>(i+1));
      }
    else usm[i]=new SingleSphericalMirror(ArtePointer<CSSM>(i+1));
    }
  refl=usm[0]->reflectivity();
  }

SphericalMirror::~SphericalMirror() {
  int i;
  for (i=0;i<nummir;i++) {
    if (splitted) {
      if (i<nummir/2) delete usm[i];
      else delete  lsm[i-nummir/2];
      }
    else delete usm[i];
    }
  delete usm;
  delete lsm;
  }



void SphericalMirror::xyview() {
  FILE* fp=fopen("mirror.kumac","w");
  int i,j;
  for (i=0;i<4;i++) {
    for (j=0;j<10;j++) {
      usm[index(i,j)]->xyview(fp);
      lsm[index(i,j)]->xyview(fp);
      }
    }
  fclose(fp);
  }

// ========================================================================
//
//    reflect: reflexion on a single planar mirror
//

int SinglePlaneMirror::reflect(Photon* pp) {
  Photon& p=*pp;    // synonym
  loss=2;
  double l;
  if (r.intersection(p,&l)) {
    if (!vesseltrace(&p,l,&loss)) return 0;
    return (loss=reflexion(&p,r.ez(),reflectivity().f(p.lambda())));   
    }  
  return 0;  
  }

void SinglePlaneMirror::reflectonly(Photon* pp) {
  Photon& p=*pp;    // synonym
  double l;
  r.intersection(p,&l);
  p.x()+=l*p.n();
  p.n()-=double(2)*(r.ez()*p.n())*r.ez();  // short hand formula for the reflection
  }

int SinglePlaneMirror::geometry_trace(Photon* pp) {
  Photon& p=*pp;    // synonym
  double l,ll;
  if (!r.intersection(p,&l)) return 0; 
  loss=4;
  if (RICH_Vessel->ProtonBeamShroud().intersection(p,&ll)&&ll<=l) return 0;
  loss=5;
  if (RICH_Vessel->ElectronBeamShroud().intersection(p,&ll)&&ll<=l) return 0;
  loss=1;
  p.x()+=l*p.n();
  p.n()-=double(2)*(r.ez()*p.n())*r.ez();  // short hand formula for the reflection
  return 1;
  }

// -----------------------------------------------------------------------
//
//  print data
//

void SinglePlaneMirror::print(FILE *fp) {
  fprintf(fp,"#\nSinglePlaneMirror\n");
//  fprintf(fp,"  reflectivity   %d\n",reflectivity());
  fprintf(fp,"  index       %d %d\n",row(),col());
  fprintf(fp,"  type        %s\n",type());
  fprintf(fp,"  mirrorname  %s\n",serialno());
//  fprintf(fp,"  end\n");
  surface().print(fp,(char*)0);
  }

// -----------------------------------------------------------------------
//
//  initialisation from file
//

SinglePlaneMirror::SinglePlaneMirror(ScriptFile& f,Table& rrefl) {
  char n[256],tag[256];
  loss=-1;
  refl=rrefl;
  
  Vec3 x0,vx,vy;
  
  while (f.getline(n)) {
    char *s=n;
    sscanf(s,"%s",tag);s+=f.itemlength(s,1);
    int err=0;
    if (!strcmp(tag,"index"))   
      err=(sscanf(s,"%d %d",&ro,&co)!=1);  
    else if (!strcmp(tag,"type"))   
      err=(sscanf(s,"%s",&ttype)!=1);  
    else if (!strcmp(tag,"mirrorname"))   
      err=(sscanf(s,"%s",&seno)!=1);  
    else if      (!strcmp(tag,"origin"))   
      err=(sscanf(s,"%lf %lf %lf",&x0[0],&x0[1],&x0[2])!=3);
    else if (!strcmp(tag,"x_vector")) 
      err=(sscanf(s,"%lf %lf %lf",&vx[0],&vx[1],&vx[2])!=3);
    else if (!strcmp(tag,"y_vector"))   
      err=(sscanf(s,"%lf %lf %lf",&vy[0],&vy[1],&vy[2])!=3);
    else if (!strcmp(tag,"end"))  break;
    else {
      fprintf(stderr,
         "SinglePlaneMirror(ScriptFile):  Unknown tag: %s\n",tag);
      exit(0);
      }
    }   
  r=Rectangle(x0,vx,vy);
  
  CSPM pm;
  pm.orig[0]=surface().x0().x();
  pm.orig[1]=surface().x0().y();
  pm.orig[2]=surface().x0().z();
  pm.xvec[0]=surface().xsize()*surface().ex().x();
  pm.xvec[1]=surface().xsize()*surface().ex().y();
  pm.xvec[2]=surface().xsize()*surface().ex().z();
  pm.yvec[0]=surface().ysize()*surface().ey().x();
  pm.yvec[1]=surface().ysize()*surface().ey().y();
  pm.yvec[2]=surface().ysize()*surface().ey().z();
  pm.iref=reflectivity().firstpoint();
  pm.nref=reflectivity().numpoints();
  ArtePointer<CSPM>::append(pm);

  }


SinglePlaneMirror::SinglePlaneMirror(ArtePointer<CSPM> a)
   : refl(a->iref,a->nref),
     r(Vec3(a->orig),Vec3(a->xvec),Vec3(a->yvec)) {
  }

SinglePlaneMirror::SinglePlaneMirror(const SinglePlaneMirror& s) {
  memcpy(this,&s,sizeof(SinglePlaneMirror));
  }
// ===========================================================================
//
//   only geometry: search the right single planar mirror
//

int PlaneMirror::geometry_trace(Photon* pp) {
  if (pp->x().y()>0) return (loss=fullupper.geometry_trace(pp));
  else return (loss=fulllower.geometry_trace(pp));
  }

int PlaneMirror::fullmirror_reflect(Photon* pp) {
  SinglePlaneMirror *m;
  loss=2;
  if (pp->x().y()>0) {
    return (loss=fullupper.reflect(pp));
    }
  else {
    if (loss=fulllower.reflect(pp)) {
      if (norm(pp->x()-RICH_Vessel->ElectronBeamShroud().x0())<RadiusElectronHole()) {
        return 0;
        }
      return 1;      
      }
    else return 0;
    }
  return 0;
  }


int PlaneMirror::reflect(Photon* pp) {
  if (speed==1) return fullmirror_reflect(pp);
  
  SinglePlaneMirror **m;
  double l;
  loss=2;
  if (pp->x().y()>0) {
    if (!fullupper.intersection(*pp,&l)) return 0;
    m=upm;
    }
  else {
    if (!fulllower.intersection(*pp,&l)) return 0;
    m=lpm;
    }
  pp->x()+=0.999*l*pp->n(); // stop some mm in front off the mirror
  double xbin=m[1]->x0().x()-m[0]->x0().x();
  int k=int((pp->x().x()-m[0]->x0().x())/xbin);
  if (m[k]->reflect(pp)) {
    if (norm(pp->x()-RICH_Vessel->ElectronBeamShroud().x0())<RadiusElectronHole()) {
      return 0;
      }
    loss=m[k]->Loss();
    return 1;
    } 
  return 0;  
  }

// -----------------------------------------------------------------------

void PlaneMirror::print(FILE *fp) {
  fprintf(fp,"#\nPlaneMirror\n");
  fprintf(fp,"  numberofmirrors   %d\n",nummirror);
  fprintf(fp,"  radius_electron_hole %lf\n",radiuselectronhole);
//  fprintf(fp,"  splitted          %d\n",split);
  fprintf(fp,"#\n reflectivity\n");
  upm[0]->reflectivity().print(fp,0);
  fprintf(fp,"  end  # plane mirror \n");
  if (splitted) {
    int i;
    fprintf(fp,"#\n# Lower PlaneMirror\n");
    for (i=0;i<nummirror/2;i++) lpm[i]->print(fp);
    fprintf(fp,"#\n# Upper PlaneMirror\n");
    for (i=0;i<nummirror/2;i++) upm[i]->print(fp);
    }
  }

// -----------------------------------------------------------------------
//
//  initialisation from file
//

PlaneMirror::PlaneMirror(ScriptFile& f) {
  char n[256],tag[256];
  
  FullSimulation();
  splitted=1;
  while (f.getline(n)) {
    char *s=n;
    sscanf(s,"%s",tag);s+=f.itemlength(s,1);
    int err=0;
    if (!strcmp(tag,"numberofmirrors"))   
      err=(sscanf(s,"%d",&nummirror)!=1);  
    else if (!strcmp(tag,"radius_electron_hole")) {
      err=(sscanf(s,"%lf",&radiuselectronhole)!=1); 
      } 
    else if (!strcmp(tag,"splitted"))   
      err=(sscanf(s,"%d",&splitted)!=1);  
    else if (!strcmp(tag,"reflectivity"))   
      refl=Table(f);  
    else if (!strcmp(tag,"end"))  break;
    else {
      fprintf(stderr,
         "PlaneMirror(ScriptFile):  Unknown tag: %s\n",tag);
      exit(0);
      }
    }   
  if (splitted) {
    upm=new SinglePlaneMirror*[nummirror/2];
    lpm=new SinglePlaneMirror*[nummirror/2];
    }
  else {
    upm=new SinglePlaneMirror*[nummirror];
    }

  int i;
  SinglePlaneMirror* m;
  for (i=0;i<nummirror;i++) {
    f.getline(n);sscanf(n,"%s",tag);
    if (strcmp(tag,"SinglePlaneMirror")) {
      fprintf(stderr,"PlaneMirror(ScriptFile): Not enough single planar mirrors are"
              " defined in the input file. Abort\n");
      exit(1);
      }
    m=new SinglePlaneMirror(f,refl);
    int k=index(m->row(),m->col());
    if (splitted) {
      if (k<nummirror/2) lpm[k]=m;
      else upm[k-nummirror/2]=m;
      }
    else upm[k]=m;
    }

  fullupper=SinglePlaneMirror(*upm[0]);
  fullupper.xsize()=upm[nummirror/2-1]->x0().x()-upm[0]->x0().x()+upm[0]->xsize();
  fulllower=SinglePlaneMirror(*lpm[0]);
  fulllower.xsize()=lpm[nummirror/2-1]->x0().x()-lpm[0]->x0().x()+lpm[0]->xsize();  

  }
/*

PlaneMirror::PlaneMirror() {
  FullSimulation();
  splitted=1;
  if (!(nummirror=ArtePointer<CSPM>::nrow())) return;
  
  if (splitted) {
    upm=new SinglePlaneMirror*[nummirror/2];
    lpm=new SinglePlaneMirror*[nummirror/2];
    }
  else {
    upm=new SinglePlaneMirror*[nummirror];
    }

  int i;
  for (i=0;i<nummirror;i++) {
    if (splitted) {
      if (i<nummirror/2) upm[i]=new SinglePlaneMirror(ArtePointer<CSPM>(i+1));
      else lpm[i-nummirror/2]=new SinglePlaneMirror(ArtePointer<CSPM>(i+1));
      }
    else upm[i]=new SinglePlaneMirror(ArtePointer<CSPM>(i+1));
    }

  fullupper=SinglePlaneMirror(*upm[0]);
  fullupper.xsize()=upm[nummirror/2-1]->x0().x()-upm[0]->x0().x()+upm[0]->xsize();
  fulllower=SinglePlaneMirror(*lpm[0]);
  fulllower.xsize()=lpm[nummirror/2-1]->x0().x()-lpm[0]->x0().x()+lpm[0]->xsize();  

  }

*/

PlaneMirror::~PlaneMirror() {
  int i;
  for (i=0;i<nummirror;i++) {
    if (splitted) {
      if (i<nummirror/2) delete upm[i];
      else delete  lpm[i-nummirror/2];
      }
    else delete upm[i];
    }
  delete upm;
  delete lpm;
  }

// ===========================================================================
//

int PhotonDetectorEntryWindow::trace(Photon* p) {
  Rectangle *m;
  double t;
  if (p->x().y()>0) {
    m=uw;
    t=uwthickness;
    }
  else {
    m=lw;
    t=lwthickness;
    }
  double l;
  if (!m->intersection(*p,&l)) return 0;
  int loss;
  if (!vesseltrace(p,l,&loss)) return 0;
  return mat->trace(p,fabs(t/(m->ez()*p->n())));
  }

void PhotonDetectorEntryWindow::traceonly_upper(Photon* p) {
  double l;
  uw->intersection(*p,&l);
  p->x()+=(l+uwthickness)*p->n();
  }
  
void PhotonDetectorEntryWindow::traceonly_lower(Photon* p) {
  double l;
  lw->intersection(*p,&l);
  p->x()+=(l+lwthickness)*p->n();
  }

 
void PhotonDetectorEntryWindow::print(FILE *fp) {
  fprintf(fp,"#\nPhotonDetectorEntryWindow\n");
  fprintf(fp,"#\n upper_window\n");
  fprintf(fp,"  thickness  %lf\n",UpperWindowThickness());
  UpperWindow().print(fp,0);
  fprintf(fp,"#\n lower_window\n");
  fprintf(fp,"  thickness  %lf\n",LowerWindowThickness());
  LowerWindow().print(fp,0);
  fprintf(fp,"#\n# material\n");
  Material().print(fp);
  fprintf(fp,"  end\n");
  }

PhotonDetectorEntryWindow::PhotonDetectorEntryWindow(ScriptFile& f) {
  char n[256],tag[256];

  splitted=1;
  while (f.getline(n)) {
    char *s=n;
    sscanf(s,"%s",tag);s+=f.itemlength(s,1);
    int err=0;
    if      (!strcmp(tag,"upper_window")) {
      f.getline(n);
      err=(sscanf(n,"%s %lf",tag,&uwthickness)!=1);  
      uw=new Rectangle(f);
      }
    else if (!strcmp(tag,"lower_window")) {
      f.getline(n);
      err=(sscanf(n,"%s %lf",tag,&lwthickness)!=1);  
      lw=new Rectangle(f);
      }
    else if (!strcmp(tag,"dielectric")) {
      mat=new DiElectric(f);
      }
    else if (!strcmp(tag,"end"))  break;
    else {
      fprintf(stderr,
         "PhotonDetectorEntryWindow(ScriptFile):  Unknown tag: %s %d\n",tag,err);
      exit(1);
      }
    }  
  if (ArtePointer<CPDW>::nrow()==0) {
    CPDW win;
    ArtePointer<CPDW>::append(win);
    ArtePointer<CPDW>::append(win);
    }

  ArtePointer<CPDW> uwin(1);
  ArtePointer<CPDW> lwin(2);

  uwin->orig[0]=UpperWindow().x0().x();
  uwin->orig[1]=UpperWindow().x0().y();
  uwin->orig[2]=UpperWindow().x0().y();
  uwin->xvec[0]=UpperWindow().xsize()*UpperWindow().ex().x();
  uwin->xvec[1]=UpperWindow().xsize()*UpperWindow().ex().y();
  uwin->xvec[2]=UpperWindow().xsize()*UpperWindow().ex().z();
  uwin->yvec[0]=UpperWindow().ysize()*UpperWindow().ey().x();
  uwin->yvec[1]=UpperWindow().ysize()*UpperWindow().ey().y();
  uwin->yvec[2]=UpperWindow().ysize()*UpperWindow().ey().z();
  uwin->t      =UpperWindowThickness();
  uwin->imat   =Material().ArteTableIndex();

  lwin->orig[0]=LowerWindow().x0().x();
  lwin->orig[1]=LowerWindow().x0().y();
  lwin->orig[2]=LowerWindow().x0().y();
  lwin->xvec[0]=LowerWindow().xsize()*LowerWindow().ex().x();
  lwin->xvec[1]=LowerWindow().xsize()*LowerWindow().ex().y();
  lwin->xvec[2]=LowerWindow().xsize()*LowerWindow().ex().z();
  lwin->yvec[0]=LowerWindow().ysize()*LowerWindow().ey().x();
  lwin->yvec[1]=LowerWindow().ysize()*LowerWindow().ey().y();
  lwin->yvec[2]=LowerWindow().ysize()*LowerWindow().ey().z();
  lwin->t      =LowerWindowThickness();
  lwin->imat   =Material().ArteTableIndex();

  }

PhotonDetectorEntryWindow::PhotonDetectorEntryWindow() {
  if (ArtePointer<CPDW>::nrow()!=2) {
    fprintf(stderr,
      "RICH: PhotonDetectorEntryWindow: Arte Table CPDW is not filled.\n");
    return;
    }
  ArtePointer<CPDW> uwin(1);
  ArtePointer<CPDW> lwin(2);
  uw=new Rectangle(Vec3(uwin->orig),Vec3(uwin->xvec),Vec3(uwin->yvec));
  lw=new Rectangle(Vec3(lwin->orig),Vec3(lwin->xvec),Vec3(lwin->yvec));
  UpperWindowThickness()=uwin->t;
  LowerWindowThickness()=lwin->t;
  mat=new DiElectric(ArtePointer<CDIM>(uwin->imat));
  }

PhotonDetectorEntryWindow::~PhotonDetectorEntryWindow() {
  delete mat;
  delete uw;
  delete lw;
  }



// ==========================================================================

int RICHVessel::raytrace(Photon *p) {
  if (p->n().z()>0) {
    if(!spherical_mirror->reflect(p)) return 0;
    }
  if (!plane_mirror->reflect(p)) return 0;
  return photondetector_window->trace(p);
  }

void RICHVessel::track_upper(Photon *p,double *l) {
  Vec3 x=p->x();
  spherical_mirror->reflectonly_upper(p);
  x=p->x();
  plane_mirror->reflectonly_upper(p);
  *l=norm(p->x()-x);
  x=p->x();
  photondetector_window->traceonly_upper(p);
  *l+=norm(p->x()-x);
  }

void RICHVessel::track_lower(Photon *p,double *l) {
  Vec3 x=p->x();
  spherical_mirror->reflectonly_lower(p);
  x=p->x();
  plane_mirror->reflectonly_lower(p);
  *l=norm(p->x()-x);
  x=p->x();
  photondetector_window->traceonly_lower(p);
  *l+=norm(p->x()-x);
  }

int RICHVessel::geometry_trace(Photon *p) {
  if (!spherical_mirror->geometry_trace(p)) return 1;
  if (!plane_mirror->geometry_trace(p)) return 2;
  if (p->x().y()<0) photondetector_window->traceonly_lower(p);
  else             photondetector_window->traceonly_upper(p);
  return 0;
  }


// ---------------------------------------------------------------------------
//
//  initialisation from ascii files
//

RICHVessel::RICHVessel(char *path) {
  if (path) {
    char n[256],s[256];
    strcpy(n,path);
    if (n[strlen(n)-1]!='/') strcat(n,"/");
    strcat(n,"vessel.dat");
    ScriptFile fm(n);
    while (fm.gettag(n)) {
      fm.skipline(s);
      if (!strcmp(n,"Vessel")) {
        if (RICHdebug){ printf("scanning vessel\n");fflush(stdout);}
        vessel=new Vessel(fm);
        }
      else if (!strcmp(n,"PhotonDetectorEntryWindow")) {
        if (RICHdebug){ printf("scanning window\n");fflush(stdout);}
        photondetector_window=new PhotonDetectorEntryWindow(fm);
        }
      else if (!strcmp(n,"SphericalMirror")) {
        if (RICHdebug){ printf("scanning spherical mirror\n");fflush(stdout);}
        spherical_mirror=new SphericalMirror(fm);
        }
      else if (!strcmp(n,"PlaneMirror")) {
        if (RICHdebug){  printf("scanning plane mirror\n");fflush(stdout);}
        plane_mirror=new PlaneMirror(fm);
        }
      else {
        fprintf(stderr,"vessel.dat: Unknown tag: %s\n",n);
        exit(1);
        }
      }
    }
  else {  // init from Arte Tables
  /*
    spherical_mirror     = new SphericalMirror();
    plane_mirror        = new PlaneMirror();
    photondetector_window= new PhotonDetectorEntryWindow();
    vessel              = new Vessel();
 */
    }
  } 

RICHVessel::~RICHVessel() {  
  delete spherical_mirror;
  delete plane_mirror;
  delete photondetector_window;
  delete vessel;

  }

// ---------------------------------------------------------------------------
//
//  print to ascii files
//

void RICHVessel::print() {
  FILE *fp;  

  if (!(fp=fopen("vessel.dat","w"))) {
    fprintf(stderr,"RICHDetector::print: Can't open file: vessel.dat\n");
    exit(1);
    }
  if (vessel) vessel->print(fp);
  else {
    fprintf(stderr,"RICHDetector::print: vessel not initialized\n");
    }
  if (photondetector_window) photondetector_window->print(fp);
  else {
    fprintf(stderr,"RICHDetector::print: photondetector_window not
    initialized\n");
    }
  if (spherical_mirror) spherical_mirror->print(fp);
  else {
    fprintf(stderr,"RICHDetector::print: spherical_mirror not initialized\n");
    }
  if (plane_mirror) plane_mirror->print(fp);
  else {
    fprintf(stderr,"RICHDetector::print: plane_mirror not initialized\n");
    }
  fclose(fp); 
  } 

